#!/usr/local/bin/ruby
# $Id: test_cgi.rb,v 1.3.2.1 2002/02/27 14:16:46 toki Exp $
# -
# Test of WWWsrv::CGIdocument class

require 'comm'
require 'pseudo_io'
require 'pseudo_doc'
require 'wwwsrv/logger'
require 'wwwsrv/request'
require 'wwwsrv/response'
require 'wwwsrv/cgi'
require 'runit/testcase'
require 'runit/cui/testrunner'

module TestWWWsrv
  class TestCGIdocument < RUNIT::TestCase
    include Communicate

    def setup
      @cgi_path = 'test.cgi'
      @cgi_doc = WWWsrv::CGIdocument::new(Dir::getwd, false, 'ruby')

      @log_iobuf = PseudoIO::new
      @srv_opt = WWWsrv::ServerOption::new
      @srv_opt.root = @cgi_doc
      @srv_opt.log = WWWsrv::Logger::new(@log_iobuf, WWWsrv::Logger::DEBUG)

      @request = WWWsrv::Request::new
      @request.version = WWWsrv::HTTP_1_1
      @request.header = WWWsrv::Header::new
      @request.header.user_agent = 'Test/0.9 (with RUNIT)'
      @request.header.host = 'localhost'
      @request.content = PseudoIO::new
      @request.host = WWWsrv::Host::new('localhost')
      @request.client = WWWsrv::Client::new('localhost', '127.0.0.1')
    end

    def test_cgi_var
      request_content = "Hello world.\n"
      @request.method = 'GET'
      @request.path = '/test.cgi?1+2+3+%7E'
      @request.content.write(request_content)
      @request.header.content_type = 'text/html'
      @request.header.content_length =  request_content.length.to_s
      cgi_var = WWWsrv::CGIdocument::cgi_var(@request, @srv_opt)
      assert_equal('CGI/1.1', cgi_var['GATEWAY_INTERFACE'])
      assert_equal('GET', cgi_var['REQUEST_METHOD'])
      assert_equal('1+2+3+%7E', cgi_var['QUERY_STRING'])
      assert_equal('127.0.0.1', cgi_var['REMOTE_ADDR'])
      assert_equal('localhost', cgi_var['REMOTE_HOST'])
      assert_equal('/test.cgi', cgi_var['SCRIPT_NAME'])
      assert_equal('localhost', cgi_var['SERVER_NAME'])
      assert_equal('80', cgi_var['SERVER_PORT'])
      assert_equal('HTTP/1.1', cgi_var['SERVER_PROTOCOL'])
      assert_equal(WWWsrv::PRODUCT, cgi_var['SERVER_SOFTWARE'])
      assert_equal('text/html', cgi_var['CONTENT_TYPE'])
      assert_equal(request_content.length.to_s, cgi_var['CONTENT_LENGTH'])
      assert_equal('Test/0.9 (with RUNIT)', cgi_var['HTTP_USER_AGENT'])
    end

    def test_new
      assert_instance_of(WWWsrv::CGIdocument, @cgi_doc)
    end

    def test_content
      content = "Hello world.\n"
      open(@cgi_path, 'w') do |output|
	output.binmode
	output.write <<-EOF
#!/usr/local/bin/ruby
content = #{content.dump}
print "Content-Type: text/plain\\r\\n"
print "Content-Length: \#{content.length}\\r\\n"
print "\\r\\n"
print content
	EOF
      end

      begin
	@request.method = 'GET'
	@request.path = '/test.cgi'
	response = WWWsrv::Response::new

	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@cgi_doc, '1-1', @srv_opt, '/', @request, response)

	assert_equal(:success, state)
	assert_instance_of(WWWsrv::Response, doc_response)
	assert_equal(200, doc_response.status)
	assert_equal('OK', doc_response.reason)
	assert_equal('text/plain', doc_response.header.content_type)
	assert_equal(content.length.to_s, doc_response.header.content_length)
	assert_equal(content, doc_content)
      ensure
	File::unlink(@cgi_path)
      end
    end

    def test_content_with_status
      content = "Hello world.\n"
      open(@cgi_path, 'w') do |output|
	output.binmode
	output.write <<-EOF
#!/usr/local/bin/ruby
content = #{content.dump}
print "Status: 402 Forbidden\\r\\n"
print "Content-Type: text/plain\\r\\n"
print "Content-Length: \#{content.length}\\r\\n"
print "\\r\\n"
print content
	EOF
      end

      begin
	@request.method = 'GET'
	@request.path = '/test.cgi'
	response = WWWsrv::Response::new

	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@cgi_doc, '1-1', @srv_opt, '/', @request, response)

	assert_equal(:success, state)
	assert_instance_of(WWWsrv::Response, doc_response)
	assert_equal(402, doc_response.status)
	assert_equal('Forbidden', doc_response.reason)
	assert_equal('text/plain', doc_response.header.content_type)
	assert_equal(content.length.to_s, doc_response.header.content_length)
	assert_equal(content, doc_content)
      ensure
	File::unlink(@cgi_path)
      end
    end

    def test_content_with_location_curr_path
      content = "Hello world.\n"
      open(@cgi_path, 'w') do |output|
	output.binmode
	output.write <<-EOF
#!/usr/local/bin/ruby
content = #{content.dump}
print "Location: index.html\\r\\n"
print "Content-Type: text/plain\\r\\n"
print "Content-Length: \#{content.length}\\r\\n"
print "\\r\\n"
print content
	EOF
      end

      begin
	@request.method = 'GET'
	@request.path = '/test.cgi'
	response = WWWsrv::Response::new

	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@cgi_doc, '1-1', @srv_opt, '/', @request, response)

	assert_equal(:success, state)
	assert_instance_of(WWWsrv::Response, doc_response)
	assert_equal(302, doc_response.status)
	assert_equal('http://localhost:80/index.html', doc_response.header.location)
	assert_equal('Found', doc_response.reason)
	assert_equal(content.length.to_s, doc_response.header.content_length)
	assert_equal('text/plain', doc_response.header.content_type)
	assert_equal(content, doc_content)
      ensure
	File::unlink(@cgi_path)
      end
    end

    def test_content_with_location_rel_path
      content = "Hello world.\n"
      open(@cgi_path, 'w') do |output|
	output.binmode
	output.write <<-EOF
#!/usr/local/bin/ruby
content = #{content.dump}
print "Location: /index.html\\r\\n"
print "Content-Type: text/plain\\r\\n"
print "Content-Length: \#{content.length}\\r\\n"
print "\\r\\n"
print content
	EOF
      end

      begin
	@request.method = 'GET'
	@request.path = '/test.cgi'
	response = WWWsrv::Response::new

	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@cgi_doc, '1-1', @srv_opt, '/', @request, response)

	assert_equal(:success, state)
	assert_instance_of(WWWsrv::Response, doc_response)
	assert_equal(302, doc_response.status)
	assert_equal('http://localhost:80/index.html', doc_response.header.location)
	assert_equal('Found', doc_response.reason)
	assert_equal(content.length.to_s, doc_response.header.content_length)
	assert_equal('text/plain', doc_response.header.content_type)
	assert_equal(content, doc_content)
      ensure
	File::unlink(@cgi_path)
      end
    end

    def test_content_with_location_abs_path
      content = "Hello world.\n"
      open(@cgi_path, 'w') do |output|
	output.binmode
	output.write <<-EOF
#!/usr/local/bin/ruby
content = #{content.dump}
print "Location: http://localhost/index.html\\r\\n"
print "Content-Type: text/plain\\r\\n"
print "Content-Length: \#{content.length}\\r\\n"
print "\\r\\n"
print content
	EOF
      end

      begin
	@request.method = 'GET'
	@request.path = '/test.cgi'
	response = WWWsrv::Response::new

	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@cgi_doc, '1-1', @srv_opt, '/', @request, response)

	assert_equal(:success, state)
	assert_instance_of(WWWsrv::Response, doc_response)
	assert_equal(302, doc_response.status)
	assert_equal('http://localhost/index.html', doc_response.header.location)
	assert_equal('Found', doc_response.reason)
	assert_equal(content.length.to_s, doc_response.header.content_length)
	assert_equal('text/plain', doc_response.header.content_type)
	assert_equal(content, doc_content)
      ensure
	File::unlink(@cgi_path)
      end
    end

    def test_content_none_parsed_header
      @cgi_doc = WWWsrv::CGIdocument::new(Dir::getwd, true, 'ruby')
      content = "Hello world.\n"
      open(@cgi_path, 'w') do |output|
	output.binmode
	output.write <<-EOF
#!/usr/local/bin/ruby
content = #{content.dump}
print "HTTP/1.0 200 OK\\r\\n"
print "Content-Type: text/plain\\r\\n"
print "Content-Length: #{content.length}\\r\\n"
print "\\r\\n"
print content
	EOF
      end

      begin
	@request.method = 'GET'
	@request.path = '/test.cgi'
	response = WWWsrv::Response::new

	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@cgi_doc, '1-1', @srv_opt, '/', @request, response)

	assert_equal(:success, state)
	assert_instance_of(WWWsrv::Response, doc_response)
	assert_equal(WWWsrv::HTTP_1_0, doc_response.version)
	assert_equal(200, doc_response.status)
	assert_equal('OK', doc_response.reason)
	assert_equal('text/plain', doc_response.header.content_type)
	assert_equal(content.length.to_s, doc_response.header.content_length)
	assert_equal(content, doc_content)
      ensure
	File::unlink(@cgi_path)
      end
    end

    def test_empty_response
      open(@cgi_path, 'w') do |output|
	output.binmode
	output.write <<-EOF
#!/usr/local/bin/ruby
# nothing to do.
	EOF
      end

      begin
	@request.method = 'GET'
	@request.path = '/test.cgi'
	response = WWWsrv::Response::new

	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@cgi_doc, '1-1', @srv_opt, '/', @request, response)

	assert_equal(:error, state)
	status, message, header, version = err_args
	assert_equal(500, status)
      ensure
	File::unlink(@cgi_path)
      end
    end

    def test_override_header_field
      open(@cgi_path, 'w') do |output|
	output.binmode
	output.write <<-EOF
#!/usr/local/bin/ruby
print "Test: override\r\n"
print "Test2: not override\r\n"
print "Test2: not override2\r\n"
print "\r\n"
	EOF
      end

      begin
	@request.method = 'GET'
	@request.path = '/test.cgi'
	response = WWWsrv::Response::new
	response.header.test = 'initialized'

	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@cgi_doc, '1-1', @srv_opt, '/', @request, response)

	assert_equal(:success, state)
	assert_equal('override', doc_response.header.test)
	assert_equal('not override, not override2', doc_response.header.test2)
      ensure
	File::unlink(@cgi_path)
      end
    end
  end
end

if ($0 == __FILE__) then
  TestWWWsrv::constants.map{ |name|
    TestWWWsrv::const_get(name)
  }.select{ |const|
    (const.kind_of? Class) && (const < RUNIT::TestCase)
  }.each do |test_case|
    RUNIT::CUI::TestRunner::run(test_case::suite)
  end
end
