#!/usr/local/bin/ruby
# $Id: test_content.rb,v 1.3.2.1 2002/02/27 14:16:49 toki Exp $
# -
# Test of WWWsrv::LengthLimitedContentReader class
# Test of WWWsrv::ChunkedContentReader class
# Test of WWWsrv::EmptyContentReader class
# Test of WWWsrv::LineReader class

require 'pseudo_io'
require 'wwwsrv/logger'
require 'wwwsrv/content'
require 'runit/testcase'
require 'runit/cui/testrunner'

module TestWWWsrv
  class TestContentReader < RUNIT::TestCase
    def setup
      @content = "Hello world.\n"
      @input = PseudoIO::new
      @input.write(@content)
      @watcher = WWWsrv::RequestWatcher::new
      @timeout = 300
      @content_reader = WWWsrv::ContentReader::new('A.1-1', @input, @watcher, @timeout)
    end

    def test_new
      assert_instance_of(WWWsrv::ContentReader, @content_reader)
    end

    def test_sysread
      assert_equal('', @content_reader.sysread(0))

      content_spool = ''
      while (fragment = @content_reader.sysread(4))
	content_spool.concat(fragment)
      end
      assert_equal(@content, content_spool)

      assert_nil(@content_reader.sysread(4))
    end

    def test_read
      assert_equal('', @content_reader.read(0))

      content_spool = ''
      while (fragment = @content_reader.read(4))
	content_spool.concat(fragment)
      end
      assert_equal(@content, content_spool)

      assert_nil(@content_reader.read(4))
    end

    def test_read_all
      assert_equal(@content, @content_reader.read)
      assert_nil(@content_reader.read)
    end

    def test_to_io
      assert_equal(@input, @content_reader.to_io)
    end
  end

  class TestLengthLimitedContentReader < RUNIT::TestCase
    def setup
      @content = "Hello world.\n"
      @content_tail = "EOF\n"
      @input = PseudoIO::new
      @input.write(@content)
      @input.write(@content_tail)
      @watcher = WWWsrv::RequestWatcher::new
      @timeout = 300
    end

    def test_new
      assert_instance_of(WWWsrv::LengthLimitedContentReader,
			 WWWsrv::LengthLimitedContentReader::new('1-1', @input, @watcher, @timeout, @content.length))
    end

    def test_read_all
      content_reader =
	WWWsrv::LengthLimitedContentReader::new('1-1', @input, @watcher, @timeout, @content.length)
      assert_equal(@content, content_reader.read)
      assert_nil(content_reader.read)
    end

    def test_read_all2
      content_reader =
	WWWsrv::LengthLimitedContentReader::new('1-1', @input, @watcher, @timeout, @content.length)
      assert_equal(@content, content_reader.read(1024))
      assert_nil(content_reader.read(1024))
    end

    def test_read_fragment
      content_reader =
	WWWsrv::LengthLimitedContentReader::new('1-1', @input, @watcher, @timeout, @content.length)
      assert_equal('', content_reader.read(0))

      content_spool = ''
      while (fragment = content_reader.read(4))
	content_spool.concat(fragment)
      end

      assert_equal(@content, content_spool)
      assert_nil(content_reader.read(4))
      assert_nil(content_reader.read(0))
    end

    def test_read_all_no_content_length
      content_reader =
	WWWsrv::LengthLimitedContentReader::new('1-1', @input, @watcher, @timeout)
      assert_equal(@content, content_reader.read)
      assert_nil(content_reader.read)
    end

    def test_read_fragment_no_content_length
      content_reader =
	WWWsrv::LengthLimitedContentReader::new('1-1', @input, @watcher, @timeout)
      assert_equal('', content_reader.read(0))

      content_spool = ''
      while (fragment = content_reader.read(4))
	content_spool.concat(fragment)
      end

      assert_equal(@content, content_spool)
      assert_nil(content_reader.read(4))
      assert_nil(content_reader.read(0))
    end
  end

  class TestChunkedContentReader < RUNIT::TestCase
    def setup
      @chunk = [
	[ "Hello world.\n", [ 'debug', 'true' ] ],
	[ "Bye.\n", [] ]
      ]
      @trailer = [
	[ 'First', 'a' ],
	[ 'Second', 'b' ]
      ]

      @chunked_content = ''
      for chunk_data, chunk_extension in @chunk
	@chunked_content.concat(format('%x', chunk_data.length))
	for chunk_ext_name, chunk_ext_value in chunk_extension
	  @chunked_content.concat('; ')
	  @chunked_content.concat(chunk_ext_name)
	  if (chunk_ext_value) then
	    @chunked_content.concat('=')
	    @chunked_content.concat(chunk_ext_value)
	  end
	end
	@chunked_content.concat("\r\n")
	@chunked_content.concat(chunk_data)
	@chunked_content.concat("\r\n")
      end
      @chunked_content.concat("0\r\n")
      for field_name, field_value in @trailer
	@chunked_content.concat(field_name)
	@chunked_content.concat(': ')
	@chunked_content.concat(field_value)
	@chunked_content.concat("\r\n")
      end
      @chunked_content.concat("\r\n")

      @content = @chunk.map{ |chunk_data, chunk_extension|
	chunk_data
      }.join

      @input = PseudoIO::new
      @input.write(@chunked_content)
      @watcher = WWWsrv::RequestWatcher::new
      @timeout = 300
      @content_reader = WWWsrv::ChunkedContentReader::new(
        '1-1', @input, @watcher, @timeout)
    end

    def test_new
      assert_instance_of(WWWsrv::ChunkedContentReader, @content_reader)
    end

    def test_read_all
      assert_equal(@content, @content_reader.read)
      assert_nil(@content_reader.read)
    end

    def test_read_all2
      assert_equal(@content, @content_reader.read(1024))
      assert_nil(@content_reader.read(1024))
    end

    def test_read_fragment
      assert_equal('', @content_reader.read(0))

      content_spool = ''
      while (fragment = @content_reader.read(4))
	content_spool.concat(fragment)
      end

      assert_equal(@content, content_spool)
      assert_nil(@content_reader.read(4))
      assert_nil(@content_reader.read(0))
    end

    def test_read_chunk
      assert(@content_reader.trailer.empty?)

      while (true)
	chunk_data, chunk_extension = @content_reader.read_chunk
	chunk_data or break
	chunk = @chunk.shift
	assert_equal(chunk[0], chunk_data)
	chunk[1].each_with_index do |(chunk_ext_name, chunk_ext_value), i|
	  assert_equal(chunk_ext_name, chunk_extension[i][0])
	  assert_equal(chunk_ext_value, chunk_extension[i][1])
	end
      end
      assert_nil(@content_reader.read_chunk)

      @content_reader.trailer.each_with_index do |(field_name, field_value), i|
	assert_equal(@trailer[i][0].upcase, field_name.upcase)
	assert_equal(@trailer[i][1], field_value)
      end
    end
  end

  class TestEmptyContentReader < RUNIT::TestCase
    def setup
      @content_reader = WWWsrv::EmptyContentReader::instance
    end

    def test_new
      assert_instance_of(WWWsrv::EmptyContentReader, @content_reader)
    end

    def test_read_all
      assert_nil(@content_reader.read)
    end

    def test_read_all2
      assert_nil(@content_reader.read(1024))
    end

    def test_each_fragment
      @content_reader.each_fragment do |content_framgent|
	assert_fail("#{@content_reader.type} has contents!")
      end
    end
  end

  class TestLineReader < RUNIT::TestCase
    def setup
      @input = PseudoIO::new
      @watcher = WWWsrv::RequestWatcher::new
      @timeout = 300
      @content_reader = WWWsrv::ContentReader::new('A.1-1', @input, @watcher, @timeout)
    end

    def test_new
      assert_instance_of(WWWsrv::LineReader, WWWsrv::LineReader::new(@content_reader))
    end

    def test_gets
      l = '*'
      line_buf = [ l + "\n" ]
      18.times do
	l *= 2
	line_buf.push(l + "\n")
      end
      @input.write(line_buf.join)
      line_reader = WWWsrv::LineReader::new(@content_reader)

      while (line = line_reader.gets)
	assert_equal(line_buf.shift, line)
      end
    end

    def test_read
      @input.write('*' * 37)
      line_reader = WWWsrv::LineReader::new(@content_reader)
      assert_equal('*' * 37, line_reader.read)

      read_size = 23
      for len in 1..256
	dat = '*' * len
	@input.clear
	@input.write(dat.dup)
	line_reader = WWWsrv::LineReader::new(@content_reader)
	0.step(len - 1, read_size) do |i|
	  assert_equal(dat[i, read_size], line_reader.read(read_size))
	end
      end
    end

    def test_to_io
      line_reader = WWWsrv::LineReader::new(@content_reader)
      assert_equal(@input, line_reader.to_io)
    end
  end
end

if ($0 == __FILE__) then
  TestWWWsrv::constants.map{ |name|
    TestWWWsrv::const_get(name)
  }.select{ |const|
    (const.kind_of? Class) && (const < RUNIT::TestCase)
  }.each do |test_case|
    RUNIT::CUI::TestRunner::run(test_case::suite)
  end
end
