#!/usr/local/bin/ruby
# $Id: test_document.rb,v 1.4.2.1 2002/02/27 14:16:50 toki Exp $
# -
# Test of WWWsrv::Document class
# Test of WWWsrv::Page class
# Test of WWWsrv::DocumentFolder class
# Test of WWWsrv::DocumentPart class
# Test of WWWsrv::DocumentDispatcher class
# Test of WWWsrv::Redirect class
# Test of WWWsrv::AccessControl class
# Test of WWWsrv::PartialContent class

require 'comm'
require 'pseudo_io'
require 'pseudo_doc'
require 'wwwsrv/logger'
require 'wwwsrv/request'
require 'wwwsrv/response'
require 'wwwsrv/document'
require 'runit/testcase'
require 'runit/cui/testrunner'

module TestWWWsrv
  class TestDocument < RUNIT::TestCase
    include Communicate

    def setup
      @document = WWWsrv::Document::new
      @log_iobuf = PseudoIO::new
      @srv_opt = WWWsrv::ServerOption::new
      @srv_opt.root = PseudoDocument::new('text/plain', [ "Hello world\n" ])
      @srv_opt.log = WWWsrv::Logger::new(@log_iobuf, WWWsrv::Logger::DEBUG)
      @request = WWWsrv::Request::new('GET', '/', WWWsrv::HTTP_1_0)
      @request.header = WWWsrv::Header::new
      @request.header.host = 'localhost'
      @request.header.user_agent = 'Test/0.9 (with RUNIT)'
      @request.content = PseudoIO::new
      @request.host = WWWsrv::Host::new('localhost')
      @request.client = WWWsrv::Client::new('localhost', '127.0.0.1')
      @response = WWWsrv::Response::new(WWWsrv::HTTP_1_0)
      @response.header.date = 'Mon, 19 Nov 2001 12:12:12 GMT'
      @response.header.server = WWWsrv::PRODUCT
    end

    def test_http_date
      time = Time::gm(2001, 11, 19, 12, 12, 12)
      assert_equal('Mon, 19 Nov 2001 12:12:12 GMT', WWWsrv::Document::http_date(time))
    end

    def test_quoteHTML
      assert_equal('&quot;&lt;&gt;&amp;', WWWsrv::Document::quoteHTML('"<>&'))
    end

    def test_unquoteHTML
      assert_equal('"<>&', WWWsrv::Document::unquoteHTML('&quot;&lt;&gt;&amp;'))
    end

    def test_quoteSHELL
      assert_equal('"\"\$\`\\\\"', WWWsrv::Document::quoteSHELL('"$`\\'))
    end

    def test_encodeURL
      for c in 0x00..0xFF
	char = c.chr
	encode_url = format('%%%02X', c)
	if (char !~ /[ _0-9A-Za-z.-]/nn) then
	  if (char == ' ') then
	    assert_equal('+', WWWsrv::Document::encodeURL(char))
	  else
	    assert_equal(encode_url, WWWsrv::Document::encodeURL(char))
	  end
	end
      end
    end

    def test_decodeURL
      for c in 0x00..0xFF
	char = c.chr
	encode_url = format('%%%02X', c)
	assert_equal(char, WWWsrv::Document::decodeURL(encode_url))
	encode_url = format('%%%02x', c)
	assert_equal(char, WWWsrv::Document::decodeURL(encode_url))
      end
    end

    def test_new
      assert_instance_of(WWWsrv::Document, @document)
    end

    def test_content
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@document, '1-1', @srv_opt, '/', @request, @response)
      assert_equal(:error, state)
      err_status, err_message = err_args
      assert_equal(404, err_status)
      assert_not_nil(err_message)
      assert(! err_message.empty?)
      assert_nil(doc_content)
      assert(doc_content_alist.empty?)
    end
  end

  class TestPage < RUNIT::TestCase
    include Communicate

    def setup
      @content = "Hello world.\n"
      @type = 'text/html'
      @mtime = Time::now
      @page = WWWsrv::Page::new(@content, @type, @mtime)

      @log_iobuf = PseudoIO::new
      @srv_opt = WWWsrv::ServerOption::new
      @srv_opt.root = @page
      @srv_opt.log = WWWsrv::Logger::new(@log_iobuf, WWWsrv::Logger::DEBUG)

      @content_reader = PseudoIO::new
      @request = WWWsrv::Request::new('GET', '/', WWWsrv::HTTP_1_1)
      @request.header = WWWsrv::Header::new
      @request.header.host = 'localhost'
      @request.header.user_agent = "Test/#{WWWsrv::VERSION} (with RUNIT)"
      @request.host = WWWsrv::Host::new('localhost')
      @request.content = @content_reader
      @request.client = WWWsrv::Client::new('localhost', '127.0.0.1')
    end

    def test_head
      @request.method = 'HEAD'
      response = WWWsrv::Response::new
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@page, '1-1', @srv_opt, '', @request, response)
      assert_equal(:success, state)
      assert_equal(200, doc_response.status)
      assert_match(/^".+"$/n, doc_response.header.etag)
      assert_equal(@type, doc_response.header.content_type)
      assert_equal(@content.length.to_s, doc_response.header.content_length)
      assert_nil(doc_content)
      assert(doc_content_alist.empty?)

      pattern = Regexp::compile('^' + Regexp::quote(WWWsrv::HTMLconv::http_date(@mtime)) + '$',
				false, 'n')
      assert_match(pattern, doc_response.header.last_modified)

      md5 = MD5::new
      md5.update(@content)
      assert_equal(md5.digest, doc_response.header.content_md5.unpack('m').first)
    end

    def test_get
      @request.method = 'GET'
      response = WWWsrv::Response::new
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@page, '1-1', @srv_opt, '', @request, response)
      assert_equal(:success, state)
      assert_equal(200, doc_response.status)
      assert_match(/^".+"$/n, doc_response.header.etag)
      assert_equal(@type, doc_response.header.content_type)
      assert_equal(@content.length.to_s, doc_response.header.content_length)
      assert_equal(@content, doc_content)
      assert(! doc_content_alist.empty?)

      pattern = Regexp::compile('^' + Regexp::quote(WWWsrv::HTMLconv::http_date(@mtime)) + '$',
				false, 'n')
      assert_match(pattern, doc_response.header.last_modified)

      md5 = MD5::new
      md5.update(@content)
      assert_equal(md5.digest, doc_response.header.content_md5.unpack('m').first)
    end
  end

  class TestDocumentFolder < RUNIT::TestCase
    def setup
      @folder = WWWsrv::DocumentFolder::new
      @mount_alist = [
	[ '/',                [ '/',     WWWsrv::Document::new, nil          ] ],
	[ '/test',            [ '/test', WWWsrv::Document::new, nil          ] ],
	[ '/test/index.html', [ '/test', WWWsrv::Document::new, 'index.html' ] ],
	[ '/data',            [ '/data', WWWsrv::Document::new, nil          ] ],
	[ '/data/index.html', [ '/data', WWWsrv::Document::new, 'index.html' ] ],
	[ '/cgi/index.cgi',   [ '/cgi',  WWWsrv::Document::new, /\.cgi$/n    ] ],
	[ '/ssi/index.shtml', [ '/ssi',  WWWsrv::Document::new, /\.shtml/n   ] ]
      ]
    end

    def test_new
      assert_instance_of(WWWsrv::DocumentFolder, @folder)
    end

    def test_mount
      for path, (mount_point, document, page) in @mount_alist
	if (page) then
	  assert_equal(document, @folder.mount(document, mount_point, page))
	else
	  assert_equal(document, @folder.mount(document, mount_point))
	end
      end

      for path, (mount_point, document, page) in @mount_alist
	mount_pair = @folder.find(path)
	assert_equal(document, mount_pair.first)
	assert_equal(mount_point, mount_pair.last)
      end
    end

    def test_umount
      for path, (mount_point, document, page) in @mount_alist
	if (page) then
	  assert_equal(document, @folder.mount(document, mount_point, page))
	else
	  assert_equal(document, @folder.mount(document, mount_point))
	end
      end

      assert_nil(@folder.umount('/detarame'))

      for path, (mount_point, document, page) in @mount_alist
	if (page) then
	  assert_equal(document, @folder.umount(mount_point, page))
	else
	  assert_equal(document, @folder.umount(mount_point))
	end
      end
    end

    def test_find
      for path, (mount_point, document, page) in @mount_alist
	if (page) then
	  assert_equal(document, @folder.mount(document, mount_point, page))
	else
	  assert_equal(document, @folder.mount(document, mount_point))
	end
      end

      [ %w[ /                /                ],
	%w[ /                /index.html      ],
	%w[ /test            /test            ],
	%w[ /test            /test/test.html  ],
	%w[ /test/index.html /test/index.html ],
	%w[ /data            /data            ],
	%w[ /data            /data/test.html  ],
	%w[ /data/index.html /data/index.html ],
	%w[ /cgi/index.cgi   /cgi/test.cgi    ],
	%w[ /cgi/index.cgi   /cgi/index.cgi   ],
	%w[ /                /cgi/index.html  ],
	%w[ /ssi/index.shtml /ssi/test.shtml  ],
	%w[ /ssi/index.shtml /ssi/index.shtml ],
	%w[ /                /ssi/index.html  ]
      ].each do |path, page_path|
	assert_not_nil(mount_pair = @folder.find(page_path))
	assert_equal(@mount_alist.assoc(path)[1][1], mount_pair.first)
	assert_equal(@mount_alist.assoc(path)[1][0], mount_pair.last)
      end
    end

    def test_setup
      folder = WWWsrv::DocumentFolder::new
      document = WWWsrv::Document::new
      class << document
	attr_reader :count

	def warm_up(log)
	  @count = 0 unless @count
	  @count += 1
	end

	def cool_down(log)
	  @count -= 1
	end
      end

      folder.mount(document, '/')
      folder.mount(document, '/index.html')
      folder.mount(folder, '/recursive')
      folder.mount(folder.subdocument('/subpath'), '/subfolder')

      log = WWWsrv::Logger::new(PseudoIO::new, WWWsrv::Logger::DEBUG)
      folder.setup(log)
      assert_equal(1, document.count)
      folder.cleanup(log)
      assert_equal(0, document.count)
    end
  end

  class TestDocumentPart < RUNIT::TestCase
    include Communicate

    def setup
      @path = 'test'
      @name = 'index.html'
      @content = <<-EOF
<html>
<head>
<title>Test</title>
</head>
<body>
<h1>Test</h1>
<p>Hello world.</p>
</body>
</html>
      EOF

      @root = WWWsrv::DocumentFolder::new
      @root.mount(WWWsrv::Page::new(@content), '/' + @path, @name)
      @document = WWWsrv::DocumentPart::new(@root, '/' + @path)

      @log_iobuf = PseudoIO::new
      @srv_opt = WWWsrv::ServerOption::new
      @srv_opt.root = @root
      @srv_opt.log = WWWsrv::Logger::new(@log_iobuf, WWWsrv::Logger::DEBUG)

      @request = WWWsrv::Request::new
      @request.version = WWWsrv::HTTP_1_1
      @request.header = WWWsrv::Header::new
      @request.header.host = 'localhost'
      @request.header.user_agent = 'Test/0.9 (with RUNIT)'
      @request.host = WWWsrv::Host::new('localhost')
      @request.content = PseudoIO::new
      @request.client = WWWsrv::Client::new('localhost', '127.0.0.1')
    end

    def test_new
      assert_instance_of(WWWsrv::DocumentPart, @document)
    end

    def test_content
      prefix = '/' + @path
      @request.method = 'GET'
      @request.path = [ prefix, @name ].join('/')
      response = WWWsrv::Response::new
      response.header.date = WWWsrv::HTMLconv::http_date(Time::now)
      response.header.server = WWWsrv::PRODUCT

      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@document, '1-1', @srv_opt, prefix, @request, response)

      assert_equal(:success, state)
      assert_instance_of(WWWsrv::Response, doc_response)
      assert_equal(WWWsrv::HTTP_1_1, doc_response.version)
      assert_equal(200, doc_response.status)
      assert_equal('OK', doc_response.reason)
      assert_equal(@content, doc_content)
      assert(! doc_content_alist.empty?)

      prefix = '/detarame'
      @request.path = [ prefix, @name ].join('/')
      response = WWWsrv::Response::new
      response.header.date = WWWsrv::HTMLconv::http_date(Time::now)
      response.header.server = WWWsrv::PRODUCT

      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@document, '1-1', @srv_opt, prefix, @request, response)

      assert_equal(:success, state)
      assert_instance_of(WWWsrv::Response, doc_response)
      assert_equal(WWWsrv::HTTP_1_1, doc_response.version)
      assert_equal(200, doc_response.status)
      assert_equal('OK', doc_response.reason)
      assert_equal(@content, doc_content)
      assert(! doc_content_alist.empty?)
    end
  end

  class TestDocumentDispatcher < RUNIT::TestCase
    include Communicate

    def setup
      @dispatcher = WWWsrv::DocumentDispatcher::new(WWWsrv::Page::new('default', 'text/plain'))
      @dispatcher.add_cond(WWWsrv::Page::new('foo', 'text/plain')) { |srv_opt, request|
	request.header.x_select == 'foo'
      }
      @dispatcher.add_cond(WWWsrv::Page::new('bar', 'text/plain')) { |srv_opt, request|
	request.header.x_select == 'bar'
      }

      @sid = 'A.1-1'
      @log_iobuf = PseudoIO::new
      @srv_opt = WWWsrv::ServerOption::new
      @srv_opt.root = @document
      @srv_opt.log = WWWsrv::Logger::new(@log_iobuf)
      @request = WWWsrv::Request::new('GET', '/', WWWsrv::HTTP_1_1)
      @request.header = WWWsrv::Header::new
      @request.header.host = 'localhost'
      @request.header.user_agent = "Test/#{WWWsrv::VERSION} (with RUNIT)"
      @request.host = WWWsrv::Host::new('localhost')
      @request.content = PseudoIO::new
      @request.client = WWWsrv::Client::new('localhost', '127.0.0.1')
      @response = WWWsrv::Response::new
      @response.header.server = WWWsrv::PRODUCT
      @response.header.date = WWWsrv::HTMLconv::http_date(Time::now)
    end

    def test_dispatch_default
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@dispatcher, @sid, @srv_opt, '', @request, @response)
      assert_equal('default', doc_content)
    end

    def test_dispatch_foo
      @request.header.x_select = 'foo'
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@dispatcher, @sid, @srv_opt, '', @request, @response)
      assert_equal('foo', doc_content)
    end

    def test_dispatch_bar
      @request.header.x_select = 'bar'
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@dispatcher, @sid, @srv_opt, '', @request, @response)
      assert_equal('bar', doc_content)
    end

    def test_not_found
      @dispatcher.default_document = nil
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@dispatcher, @sid, @srv_opt, '', @request, @response)
      assert_equal(:error, state)
      err_status, err_message = err_args
      assert_equal(404, err_status)
      assert_not_nil(err_message)
    end
  end

  class TestRedirect < RUNIT::TestCase
    include Communicate

    def setup
      @sid = 'A.1-1'
      @log_iobuf = PseudoIO::new
      @srv_opt = WWWsrv::ServerOption::new
      @srv_opt.root = @document
      @srv_opt.log = WWWsrv::Logger::new(@log_iobuf)
      @request = WWWsrv::Request::new('GET', '/', WWWsrv::HTTP_1_1)
      @request.header = WWWsrv::Header::new
      @request.header.host = 'localhost'
      @request.header.user_agent = "Test/#{WWWsrv::VERSION} (with RUNIT)"
      @request.host = WWWsrv::Host::new('localhost')
      @request.content = PseudoIO::new
      @request.client = WWWsrv::Client::new('localhost', '127.0.0.1')
      @response = WWWsrv::Response::new
      @response.header.server = WWWsrv::PRODUCT
      @response.header.date = WWWsrv::HTMLconv::http_date(Time::now)
    end

    def test_redirect_301
      redirect = WWWsrv::Redirect::new{ |srv_opt, request| '/foo' }
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(redirect, @sid, @srv_opt, '', @request, @response)
      assert_equal(301, doc_response.status)
      assert_equal('http://localhost:80/foo', doc_response.header.location)
    end

    def test_redirect_307
      redirect = WWWsrv::Redirect::new(307) { |srv_opt, request| '/bar' }
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(redirect, @sid, @srv_opt, '', @request, @response)
      assert_equal(307, doc_response.status)
      assert_equal('http://localhost:80/bar', doc_response.header.location)
    end
  end

  class TestAccessControl < RUNIT::TestCase
    include Communicate

    def setup
      @root = WWWsrv::DocumentFolder::new
      @root.mount(WWWsrv::Page::new("Hello world.\n", 'text/plain'), '/', '')
      @access = WWWsrv::AccessControl::new(@root)

      @log_iobuf = PseudoIO::new
      @srv_opt = WWWsrv::ServerOption::new
      @srv_opt.root = @root
      @srv_opt.log = WWWsrv::Logger::new(@log_iobuf, WWWsrv::Logger::DISCREET)

      request_line = WWWsrv::Request::new('GET', '/', WWWsrv::HTTPversion::new(1, 0))
      header = WWWsrv::Header::new
      header.host = 'localhost'
      header.user_agent = 'Test/0.9 (with RUNIT)'

      @request1 = WWWsrv::Request::new('GET', '/', WWWsrv::HTTP_1_0, header)
      @request1.content = PseudoIO::new
      @request1.client = WWWsrv::Client::new('localhost', '127.0.0.1')
      @request1.host = WWWsrv::Host::new('localhost')

      @request2 = WWWsrv::Request::new('GET', '/', WWWsrv::HTTP_1_0, header)
      @request2.content = PseudoIO::new
      @request2.client = WWWsrv::Client::new('private', '192.168.0.1')
      @request2.host = WWWsrv::Host::new('localhost')
    end

    def test_new
      assert_instance_of(WWWsrv::AccessControl, @access)
    end

    def test_allow
      @access.add_access_control(:allow) { |srv_opt, request|
	request.client.address == '127.0.0.1'
      }
      @access.add_access_control(:allow, &proc{ |srv_opt, request|
        request.client.address == '192.168.0.1'
      })

      response = WWWsrv::Response::new
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@access, @sid, @srv_opt, '', @request1, response)
      assert_equal(:success, state)
      assert_equal(200, doc_response.status)

      response = WWWsrv::Response::new
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@access, @sid, @srv_opt, '', @request2, response)
      assert_equal(:success, state)
      assert_equal(200, doc_response.status)

      assert_nil(@log_iobuf.data)
    end

    def test_allow_with_logging
      @access.add_access_control(:allow, true) { |srv_opt, request|
	request.client.address == '127.0.0.1'
      }
      @access.add_access_control(:allow, true, &proc{ |srv_opt, request|
        request.client.address == '192.168.0.1'
      })

      response = WWWsrv::Response::new
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@access, @sid, @srv_opt, '', @request1, response)
      assert_equal(:success, state)
      assert_equal(200, doc_response.status)

      response = WWWsrv::Response::new
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@access, @sid, @srv_opt, '', @request2, response)
      assert_equal(:success, state)
      assert_equal(200, doc_response.status)

      assert_not_nil(@log_iobuf.data)
    end

    def test_deny
      @access.add_access_control(:deny) { |srv_opt, request|
	request.client.address == '127.0.0.1'
      }
      @access.add_access_control(:deny, &proc{ |srv_opt, request|
        request.client.address == '192.168.0.1'
      })

      response = WWWsrv::Response::new
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@access, @sid, @srv_opt, '', @request1, response)
      assert_equal(:error, state)
      err_status, message = err_args
      assert_equal(403, err_status)

      response = WWWsrv::Response::new
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@access, @sid, @srv_opt, '', @request2, response)
      assert_equal(:error, state)
      err_status, message = err_args
      assert_equal(403, err_status)

      assert_nil(@log_iobuf.data)
    end

    def test_deny_with_logging
      @access.add_access_control(:deny, true) { |srv_opt, request|
	request.client.address == '127.0.0.1'
      }
      @access.add_access_control(:deny, true, &proc{ |srv_opt, request|
        request.client.address == '192.168.0.1'
      })

      response = WWWsrv::Response::new
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@access, @sid, @srv_opt, '', @request1, response)
      assert_equal(:error, state)
      err_status, err_message = err_args
      assert_equal(403, err_status)
      assert_nil(doc_response)
      assert_nil(doc_content)
      assert(doc_content_alist.empty?)

      response = WWWsrv::Response::new
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@access, @sid, @srv_opt, '', @request2, response)
      assert_equal(:error, state)
      err_status, err_message = err_args
      assert_equal(403, err_status)
      assert_nil(doc_response)
      assert_nil(doc_content)
      assert(doc_content_alist.empty?)

      assert_not_nil(@log_iobuf.data)
    end

    def test_allow_deny
      @access.add_access_control(:allow) { |srv_opt, request|
        request.client.address == '127.0.0.1'
      }
      @access.add_access_control(:deny, &proc{ |srv_opt, request|
        request.client.address == '192.168.0.1'
      })

      response = WWWsrv::Response::new
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@access, @sid, @srv_opt, '', @request1, response)
      assert_equal(:success, state)
      assert_equal(200, doc_response.status)

      response = WWWsrv::Response::new
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@access, @sid, @srv_opt, '', @request2, response)
      assert_equal(:error, state)
      err_status, message = err_args
      assert_equal(403, err_status)
      assert_nil(doc_response)
      assert_nil(doc_content)
      assert(doc_content_alist.empty?)

      assert_nil(@log_iobuf.data)
    end
  end

  class TestPartialContent < RUNIT::TestCase
    include Communicate

    def setup
      @content = %w[
        01234
        56789
        ABCDE
      ]
      @content_length = @content.join.length
      @source = PseudoDocument::new('text/plain', @content)
      @partial = WWWsrv::PartialContent::new(@source)
      @request = WWWsrv::Request::new('GET', '/', WWWsrv::HTTP_1_1)
      @request.header = WWWsrv::Header::new
      @request.header.host = 'localhost'
      @request.header.user_agent = "Test/#{WWWsrv::VERSION} (with RUNIT)"
      @request.content = PseudoIO::new
      @request.host = WWWsrv::Host::new('localhost')
      @request.client = WWWsrv::Client::new('localhost', '127.0.0.1')
      @log_iobuf = PseudoIO::new
      @srv_opt = WWWsrv::ServerOption::new
      @srv_opt.root = @document
      @srv_opt.log = WWWsrv::Logger::new(@log_iobuf, WWWsrv::Logger::DEBUG)
    end

    def test_new
      assert_instance_of(WWWsrv::PartialContent, @partial)
    end

    def test_without_range
      response = WWWsrv::Response::new
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@partial, 'A.1-1', @srv_opt, '', @request, response)
      assert_equal(:success, state)
      assert_equal(200, doc_response.status)
      assert_nil(doc_response.header.content_range)
      assert_equal(@content.join, doc_content)
    end

    def test_range_not_200
      response = WWWsrv::Response::new(WWWsrv::HTTP_1_1, 201)
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@partial, 'A.1-1', @srv_opt, '', @request, response)
      assert_equal(:success, state)
      assert_equal(201, doc_response.status)
      assert_nil(doc_response.header.content_range)
      assert_equal(@content.join, doc_content)
    end

    def check_range
      # first fragment
      @request.header.range = 'bytes=0-4'
      response = WWWsrv::Response::new
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@partial, 'A.1-1', @srv_opt, '', @request, response)
      assert_equal(:success, state)
      assert_equal(206, doc_response.status)
      assert_equal("bytes 0-4/#{@content_length}", doc_response.header.content_range)
      assert_equal('5', doc_response.header.content_length)
      assert_equal(@content[0], doc_content)

      # second fragment
      @request.header.range = 'bytes=5-9'
      response = WWWsrv::Response::new
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@partial, 'A.1-1', @srv_opt, '', @request, response)
      assert_equal(:success, state)
      assert_equal(206, doc_response.status)
      assert_equal("bytes 5-9/#{@content_length}", doc_response.header.content_range)
      assert_equal('5', doc_response.header.content_length)
      assert_equal(@content[1], doc_content)

      # third fragment
      @request.header.range = 'bytes=10-14'
      response = WWWsrv::Response::new
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@partial, 'A.1-1', @srv_opt, '', @request, response)
      assert_equal(:success, state)
      assert_equal(206, doc_response.status)
      assert_equal("bytes 10-14/#{@content_length}", doc_response.header.content_range)
      assert_equal('5', doc_response.header.content_length)
      assert_equal(@content[2], doc_content)

      # fragments over first to second
      @request.header.range = 'bytes=3-7'
      response = WWWsrv::Response::new
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@partial, 'A.1-1', @srv_opt, '', @request, response)
      assert_equal(:success, state)
      assert_equal(206, doc_response.status)
      assert_equal("bytes 3-7/#{@content_length}", doc_response.header.content_range)
      assert_equal(@content.join[3..7], doc_content)

      # fragments over second to third
      @request.header.range = 'bytes=7-'
      response = WWWsrv::Response::new
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@partial, 'A.1-1', @srv_opt, '', @request, response)
      assert_equal(:success, state)
      assert_equal(206, doc_response.status)
      assert_equal("bytes 7-14/#{@content_length}", doc_response.header.content_range)
      assert_equal('8', doc_response.header.content_length)
      assert_equal(@content.join[7..14], doc_content)

      # part of last fragment
      @request.header.range = 'bytes=-3'
      response = WWWsrv::Response::new
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@partial, 'A.1-1', @srv_opt, '', @request, response)
      assert_equal(:success, state)
      assert_equal(206, doc_response.status)
      assert_equal("bytes 12-14/#{@content_length}", doc_response.header.content_range)
      assert_equal('3', doc_response.header.content_length)
      assert_equal(@content.join[12..14], doc_content)

      # overall of last part
      @request.header.range = 'bytes=-1024'
      response = WWWsrv::Response::new
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@partial, 'A.1-1', @srv_opt, '', @request, response)
      assert_equal(:success, state)
      assert_equal(206, doc_response.status)
      assert_equal("bytes 0-14/#{@content_length}", doc_response.header.content_range)
      assert_equal('15', doc_response.header.content_length)
      assert_equal(@content.join, doc_content)
    end
    private :check_range

    def test_range
      check_range
    end

    def test_range_with_content_length
      @partial.set_content_length
      check_range
    end

    def test_out_of_range
      @request.header.range = 'bytes=1024-'
      response = WWWsrv::Response::new
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@partial, 'A.1-1', @srv_opt, '', @request, response)
      assert_equal(:error, state)
      err_status, err_message = err_args
      assert_equal(416, err_status)

      @request.header.range = 'bytes=10-3'
      response = WWWsrv::Response::new
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@partial, 'A.1-1', @srv_opt, '', @request, response)
      assert_equal(:error, state)
      err_status, err_message = err_args
      assert_equal(416, err_status)
    end
  end
end

if ($0 == __FILE__) then
  TestWWWsrv::constants.map{ |name|
    TestWWWsrv::const_get(name)
  }.select{ |const|
    (const.kind_of? Class) && (const < RUNIT::TestCase)
  }.each do |test_case|
    RUNIT::CUI::TestRunner::run(test_case::suite)
  end
end
