#!/usr/local/bin/ruby
# $Id: test_erbdoc.rb,v 1.7 2002/01/09 23:57:43 toki Exp $
# -
# Test of eRuby document for WWW server

require 'wwwsrv'
require 'wwwsrv/erbdoc'
require 'runit/testcase'
require 'runit/cui/testrunner'
require 'comm'
require 'pseudo_io'

module TestWWWsrv
  class TestERubyDocument < RUNIT::TestCase
    include Communicate

    def setup
      @path = '/'
      @name = 'index.html'

      @local_dir = 'test'
      @local_path = File::join(@local_dir, @name)
      Dir::mkdir(@local_dir)
      File::open(@local_path, 'w') do |output|
	output.write('')
      end
      @document = WWWsrv::ERubyDocument::new(@local_dir)

      @log_iobuf = PseudoIO::new
      @srv_opt = WWWsrv::ServerOption::new
      @srv_opt.root = @document
      @srv_opt.log = WWWsrv::Logger::new(@log_iobuf, WWWsrv::Logger::DEBUG)

      @request = WWWsrv::Request::new('GET', '/', WWWsrv::HTTP_1_1)
      @request.header = WWWsrv::Header::new
      @request.header.host = 'localhost'
      @request.header.user_agent = "Test/#{WWWsrv::VERSION} (with RUNIT)"
      @request.content = WWWsrv::EmptyContentReader::instance
      @request.host = WWWsrv::Host::new('localhost')
      @request.client = WWWsrv::Client::new('localhost', '127.0.0.1')

      @response = WWWsrv::Response::new(WWWsrv::HTTP_1_1)
      @response.header.date = WWWsrv::Document::format_date(Time::now)
      @response.header.server = WWWsrv::PRODUCT
    end

    def teardown
      File::unlink(@local_path)
      Dir::rmdir(@local_dir)
    end

    def test_new
      assert_instance_of(WWWsrv::ERubyDocument, @document)
    end

    def test_content
      File::open(@local_path, 'w') do |output|
	output.write(<<-EOF)
<% extend RUNIT::Assert
%><% assert_equal('1-1', SID)
%><% assert_equal('1-1', __sid__)
%><% assert_equal(SID, __sid__)
%><% assert_instance_of(WWWsrv::ServerOption, SRV_OPT)
%><% assert_instance_of(WWWsrv::ServerOption, __srv_opt__)
%><% assert_equal(SRV_OPT, __srv_opt__)
%><% assert_instance_of(WWWsrv::Request, REQUEST)
%><% assert_instance_of(WWWsrv::Request, __request__)
%><% assert_equal(REQUEST, __request__)
%><% assert_instance_of(WWWsrv::Response, RESPONSE)
%><% assert_instance_of(WWWsrv::Response, __response__)
%><% assert_equal(RESPONSE, __response__)
%><% assert_equal('text/html', RESPONSE.header.content_type)
%><% RESPONSE.header.content_type = 'text/plain'
%><% __response__.header.x_sid = __sid__
%><%= 'SnVzdCBhbm90aGVyIHJ1YnkgaGFja2VyLg=='.unpack('m').first %>
	EOF
      end


      @request.method = 'GET'
      @request.path = @path + @name
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@document, '1-1', @srv_opt, '', @request, @response)

      assert_equal(:success, state)
      assert_instance_of(WWWsrv::Response, doc_response)
      assert_equal(200, doc_response.status)
      assert_equal('1-1', doc_response.header.x_sid)
      assert_equal('text/plain', doc_response.header.content_type)
      assert_equal(doc_content.length.to_s, doc_response.header.content_length)
      assert_equal("Just another ruby hacker.\n", doc_content)
      assert(! doc_content_alist.empty?)
    end

    def test_content_with_not_allowed_method
      File::open(@local_path, 'w') do |output|
	output.write(<<-EOF)
<%= 'SnVzdCBhbm90aGVyIHJ1YnkgaGFja2VyLg=='.unpack('m').first %>
	EOF
      end

      @request.method = 'PUT'
      @request.path = @path + @name
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@document, '1-1', @srv_opt, '', @request, @response)

      assert_equal(:error, state)
      err_status, err_message, err_header, err_version = err_args
      assert_equal(405, err_status)
      assert(! (err_header['Allow'].include? 'PUT'))
    end
  end

  class TestERubyFilter < RUNIT::TestCase
    include Communicate

    def setup
      @path = '/'
      @name = 'index.html'
      @root = WWWsrv::DocumentFolder::new
      @document = WWWsrv::ERubyFilter::new(@root)

      @log_iobuf = PseudoIO::new
      @srv_opt = WWWsrv::ServerOption::new
      @srv_opt.root = @root
      @srv_opt.log = WWWsrv::Logger::new(@log_iobuf, WWWsrv::Logger::DEBUG)

      @request = WWWsrv::Request::new('GET', '/', WWWsrv::HTTP_1_1)
      @request.header = WWWsrv::Header::new
      @request.header.host = 'localhost'
      @request.header.user_agent = "Test/#{WWWsrv::VERSION} (with RUNIT)"
      @request.content = WWWsrv::EmptyContentReader::instance
      @request.host = WWWsrv::Host::new('localhost')
      @request.client = WWWsrv::Client::new('localhost', '127.0.0.1')

      @response = WWWsrv::Response::new(WWWsrv::HTTP_1_1)
      @response.header.date = WWWsrv::Document::format_date(Time::now)
      @response.header.server = WWWsrv::PRODUCT
    end

    def test_new
      assert_instance_of(WWWsrv::ERubyFilter, @document)
    end

    def test_content
      @root.mount(WWWsrv::Page::new(<<-EOF), @path, @name)
<% extend RUNIT::Assert
%><% assert_equal('1-1', SID)
%><% assert_equal('1-1', __sid__)
%><% assert_equal(SID, __sid__)
%><% assert_instance_of(WWWsrv::ServerOption, SRV_OPT)
%><% assert_instance_of(WWWsrv::ServerOption, __srv_opt__)
%><% assert_equal(SRV_OPT, __srv_opt__)
%><% assert_instance_of(WWWsrv::Request, REQUEST)
%><% assert_instance_of(WWWsrv::Request, __request__)
%><% assert_equal(REQUEST, __request__)
%><% assert_instance_of(WWWsrv::Response, RESPONSE)
%><% assert_instance_of(WWWsrv::Response, __response__)
%><% assert_equal(RESPONSE, __response__)
%><% assert_equal('text/html', RESPONSE.header.content_type)
%><% RESPONSE.header.etag = 'detarame'           # disabled header field
%><% RESPONSE.header.last_modified = 'detarame'  # disabled header field
%><% RESPONSE.header.cntent_md5 = 'detarame'     # disabled header field
%><% RESPONSE.header.content_type = 'text/plain'
%><% __response__.header.x_sid = __sid__
%><%= 'SnVzdCBhbm90aGVyIHJ1YnkgaGFja2VyLg=='.unpack('m').first %>
      EOF

      @request.method = 'GET'
      @request.path = @path + @name
      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@document, '1-1', @srv_opt, '', @request, @response)

      assert_equal(:success, state)
      assert_instance_of(WWWsrv::Response, doc_response)
      assert_equal(200, doc_response.status)
      assert_nil(doc_response.header.etag)
      assert_nil(doc_response.header.last_modified)
      assert_nil(doc_response.header.content_md5)
      assert_equal('1-1', doc_response.header.x_sid)
      assert_equal('text/plain', doc_response.header.content_type)
      assert_equal(doc_content.length.to_s, doc_response.header.content_length)
      assert_equal("Just another ruby hacker.\n", doc_content)
      assert(! doc_content_alist.empty?)
    end
  end
end

if ($0 == __FILE__) then
  RUNIT::CUI::TestRunner::run(TestWWWsrv::TestERubyDocument::suite)
  RUNIT::CUI::TestRunner::run(TestWWWsrv::TestERubyFilter::suite)
end
