#!/usr/local/bin/ruby
# $Id: test_header.rb,v 1.3 2002/01/01 15:39:37 toki Exp $
# -
# Test of WWWsrv::Header class

require 'pseudo_io'
require 'wwwsrv/header'
require 'runit/testcase'
require 'runit/cui/testrunner'

module TestWWWsrv
  class TestHeader < RUNIT::TestCase
    def setup
      @header = WWWsrv::Header::new
    end

    def test_normalize
      [ %w[ Date           Date ],
	%w[ Server         SERVER ],
	%w[ Content-Type   content-type ],
	%w[ Content-Length CONTENT-length ],
	%w[ WWW-ETag-MD5-TE www-etag-md5-te ]
      ].each do |regular_name, field_name|
	assert_equal(regular_name, WWWsrv::Header::normalize(field_name))
      end
    end

    def test_new
      assert_instance_of(WWWsrv::Header, @header)
    end

    def test_push
      assert_nil(@header['Test-Head'])
      @header.push('Test-Head', 'Hello world')
      assert_equal('Hello world', @header['Test-Head'])
      @header.push('Test-Head', 'Hello Ruby')
      assert_equal('Hello world, Hello Ruby', @header['Test-Head'])
    end

    def test_pop
      assert_nil(@header.pop('Test-Head'))
      @header.push('Test-Head', 'Hello world')
      assert_equal('Hello world', @header.pop('Test-Head'))
    end

    def test_accessor
      assert_nil(@header['test-head'])
      assert_nil(@header['TEST-HEAD'])
      assert_nil(@header['Test-Head'])
      assert_nil(@header.test_head)
      @header['Test-Head'] = 'hello'
      assert_equal('hello', @header['test-head'])
      assert_equal('hello', @header['TEST-HEAD'])
      assert_equal('hello', @header['Test-Head'])
      assert_equal('hello', @header.test_head)
      @header.test_head = 'world'
      assert_equal('world', @header['test-head'])
      assert_equal('world', @header['TEST-HEAD'])
      assert_equal('world', @header['Test-Head'])
      assert_equal('world', @header.test_head)
      @header.test_end = 'bye'
      assert_equal('bye', @header['test-end'])
      assert_equal('bye', @header['TEST-END'])
      assert_equal('bye', @header['Test-End'])
      assert_equal('bye', @header.test_end)
    end

    def test_split_token
      assert_equal([], @header.split_token('Test'))
      @header['Test'] = 'one, two, three'
      first, second, third = @header.split_token('Test')
      assert_equal('one', first)
      assert_equal('two', second)
      assert_equal('three', third)

      @header['Test'] = 'one; q=0, two; q=0.1, three; q=0.1, four; q=1, five'
      tokens = @header.split_token('Test', true, true)

      assert_equal(2, tokens[0].length)
      assert_equal('four', tokens[0][0])
      assert_equal('q', tokens[0][1][0])
      assert_equal('1', tokens[0][1][1])

      assert_equal(1, tokens[1].length)
      assert_equal('five', tokens[1][0])
      assert_nil(tokens[1][1])

      assert_equal(2, tokens[2].length)
      assert_equal('two', tokens[2][0])
      assert_equal('q', tokens[2][1][0])
      assert_equal('0.1', tokens[2][1][1])

      assert_equal(2, tokens[3].length)
      assert_equal('three', tokens[3][0])
      assert_equal('q', tokens[3][1][0])
      assert_equal('0.1', tokens[3][1][1])

      assert_equal(2, tokens[4].length)
      assert_equal('one', tokens[4][0])
      assert_equal('q', tokens[4][1][0])
      assert_equal('0', tokens[4][1][1])
    end

    def test_each
      header_alist = [
	[ 'Date', 'Date: Mon, 19 Nov 1975 00:00:00 GMT' ],
	[ 'Server', 'Test (with RUNIT)' ],
	[ 'Content-Type', 'text/html' ],
	[ 'Content-Length', '256' ]
      ]

      for field_name, field_value in header_alist
	@header[field_name] = field_value
      end

      @header.each_with_index do |(field_name, field_value), i|
	assert_equal(header_alist[i].first, field_name)
	assert_equal(header_alist[i].last, field_value)
      end
    end

    def test_multi_field
      @header.push('test', 'first')
      assert_equal('first', @header.test)
      @header.push('test', 'second')
      assert_equal('first, second', @header.test)
      @header.push('test', 'third')
      assert_equal('first, second, third', @header.test)

      for field_name, field_value in @header
	if (field_name == 'Test') then
	  assert_equal('first, second, third', field_value)
	end
      end

      @header.set_multi_field('test')
      assert_equal('first, second, third', @header.test)
      field_values = %w[ first second third ]
      for field_name, field_value in @header
	if (field_name == 'Test') then
	  assert_equal(field_values.shift, field_value)
	end
      end

      @header.unset_multi_field('test')
      assert_equal('first, second, third', @header.test)
      for field_name, field_value in @header
	if (field_name == 'Test') then
	  assert_equal('first, second, third', field_value)
	end
      end
    end

    def test_fields
      @header.push('test', 'first')
      assert_equal(%w[ first ], @header.fields('test'))
      @header.push('test', 'second')
      assert_equal(%w[ first second ], @header.fields('test'))
      @header.push('test', 'third')
      assert_equal(%w[ first second third ], @header.fields('test'))
      @header.test = 'Hello_World'
      assert_equal(%w[ Hello_World ], @header.fields('test'))
    end
  end

  class TestHeaderParse < RUNIT::TestCase
    def setup
      @input = PseudoIO::new
    end

    def test_parse_header
      @input.write("Host: localhost\r\n")
      @input.write("User-Agent: Test/0.9 (with RUNIT)\r\n")
      @input.write("\r\n")
      @input.write("Content body...")
      header = WWWsrv::Header::parse(@input)
      assert_equal(2, header.length)
      assert_equal('localhost', header.host)
      assert_equal('Test/0.9 (with RUNIT)', header.user_agent)
    end

    def test_parse_header_eol_spaces
      @input.write("Host: localhost \r\n")
      @input.write("User-Agent: Test/0.9 (with RUNIT) \r\n")
      @input.write("\r\n")
      @input.write("Content body...")
      header = WWWsrv::Header::parse(@input)
      assert_equal(2, header.length)
      assert_equal('localhost', header.host)
      assert_equal('Test/0.9 (with RUNIT)', header.user_agent)
    end

    def test_parse_header_continue
      @input.write("Host:\r\n")
      @input.write(" localhost\r\n")
      @input.write("User-Agent: \r\n")
      @input.write(" Test/0.9 (with RUNIT)\r\n")
      @input.write("\r\n")
      @input.write("Content body...")
      header = WWWsrv::Header::parse(@input)
      assert_equal(2, header.length)
      assert_equal('localhost', header.host)
      assert_equal('Test/0.9 (with RUNIT)', header.user_agent)
    end

    def test_parse_header_new_line_sep
      @input.write("Host: localhost\n")
      @input.write("User-Agent: Test/0.9 (with RUNIT)\n")
      @input.write("\n")
      @input.write("Content body...")
      header = WWWsrv::Header::parse(@input)
      assert_equal(2, header.length)
      assert_equal('localhost', header.host)
      assert_equal('Test/0.9 (with RUNIT)', header.user_agent)
    end

    def test_parse_header_invalid
      @input.write("Host; localhost\r\n")
      @input.write("User-Agent; Test/0.9 (with RUNIT)\r\n")
      @input.write("\r\n")
      @input.write("Content body...")
      assert_exception(WWWsrv::HeaderParseError) { WWWsrv::Header::parse(@input) }
    end

    def test_parse_header_invalid_continue
      @input.write(" localhost\r\n")
      @input.write("Host:\r\n")
      @input.write("User-Agent: Test/0.9 (with RUNIT)\r\n")
      @input.write("\r\n")
      @input.write("Content body...")
      assert_exception(WWWsrv::HeaderParseError) { WWWsrv::Header::parse(@input) }
    end

    def test_parse_header_empty
      @input.write("\r\n")
      @input.write("Content body...")
      header = WWWsrv::Header::parse(@input)
      assert(header.empty?)
    end
  end
end

if ($0 == __FILE__) then
  TestWWWsrv::constants.map{ |name|
    TestWWWsrv::const_get(name)
  }.select{ |const|
    (const.kind_of? Class) && (const < RUNIT::TestCase)
  }.each do |test_case|
    RUNIT::CUI::TestRunner::run(test_case::suite)
  end
end
