#!/usr/local/bin/ruby
# $Id: test_local.rb,v 1.3 2002/01/01 13:21:12 toki Exp $
# -
# Test of WWWsrv::MIMEresolver class
# Test of WWWsrv::LocalDirIndex class
# Test of WWWsrv::LocalFileDocument class

require 'comm'
require 'pseudo_io'
require 'pseudo_doc'
require 'wwwsrv/logger'
require 'wwwsrv/request'
require 'wwwsrv/response'
require 'wwwsrv/local'
require 'runit/testcase'
require 'runit/cui/testrunner'

module TestWWWsrv
  class TestMIMEresolver < RUNIT::TestCase
    def setup
      @map = {
	'.htm'     => 'text/html',
	'.html'    => 'text/html',
	'_ja.html' => 'text/html; charset=EUC-JP',
	'.rb'      => 'text/plain',
	'.txt'     => 'text/plain'
      }

      @mime = WWWsrv::MIMEresolver::new
      for ext, mime_type in @map
	@mime.add_mime_type(ext, mime_type)
      end
    end

    def test_resolve
      assert_equal('text/html', @mime.resolve('/index.htm'))
      assert_equal('text/html', @mime.resolve('/index.html'))
      assert_equal('text/html; charset=EUC-JP', @mime.resolve('/index_ja.html'))
      assert_equal('text/plain', @mime.resolve('/index.rb'))
      assert_equal('text/plain', @mime.resolve('/index.txt'))
      assert_equal('application/octet-stream', @mime.resolve('/index.aaa'))
      assert_equal('application/octet-stream', @mime.resolve('/index.gtm'))
      assert_equal('application/octet-stream', @mime.resolve('/index.itm'))
      assert_equal('application/octet-stream', @mime.resolve('/index.zzz'))
    end

    def test_clone
      new_mime = @mime.clone
      new_mime.add_mime_type('_ja.html', 'text/html; charset=Shift_JIS')
      assert_equal('text/html; charset=EUC-JP', @mime.resolve('/index_ja.html'))
      assert_equal('text/html; charset=Shift_JIS', new_mime.resolve('/index_ja.html'))
    end

    def test_each
      for ext, mime_type in @mime
	assert_equal(@map[ext], mime_type)
	@map.delete(ext)
      end
      assert(@map.empty?)
    end
  end

  class TestLocalDirIndex < RUNIT::TestCase
    def setup
      @root_path = File::join(Dir::pwd, 'test')
      @first = 'first'
      @second = 'second'
      @third = 'third'
      @dot_test = '.test'
      @content = "Hello world.\n"

      Dir::mkdir(@root_path)
      [ @first,
	@second,
	@third,
	@dot_test
      ].each do |name|
	File::open(File::join(@root_path, name), 'w') do |output|
	  output.write(@content)
	end
      end

      @log_iobuf = PseudoIO::new
      @srv_opt = WWWsrv::ServerOption::new
      @srv_opt.root = @document
      @srv_opt.log = WWWsrv::Logger::new(@log_iobuf)
      @request = WWWsrv::Request::new
      @request.version = WWWsrv::HTTP_1_1
      @request.header = WWWsrv::Header::new
      @request.header.host = 'localhost'
      @request.host = WWWsrv::Host::new('localhost')
      @request.content = PseudoIO::new
      @request.client = WWWsrv::Client::new('localhost', '127.0.0.1')

      @local_dir = WWWsrv::LocalDirIndex::new
    end

    def teardown
      [ @first,
	@second,
	@third,
	@dot_test
      ].each do |name|
	File::unlink(File::join(@root_path, name))
      end
      Dir::rmdir(@root_path)
    end


    def test_new
      assert_instance_of(WWWsrv::LocalDirIndex, @local_dir)
    end

    def test_browse
      @request.method = 'GET'
      @request.path = '/'

      doc_response = nil
      content_spool = ''

      @local_dir.browse('1-1', @srv_opt, '', @request, WWWsrv::Response::new, @root_path) do |arg, *opts|
	unless (doc_response) then
	  doc_response = arg
	else
	  content_spool.concat(arg)
	end
      end

      assert_instance_of(WWWsrv::Response, doc_response)
      assert_equal(200, doc_response.status)
      assert_not_nil(doc_response.header.content_type)
      assert_match(/#{Regexp::quote(@first)}/n, content_spool)
      assert_match(/#{Regexp::quote(@second)}/n, content_spool)
      assert_match(/#{Regexp::quote(@third)}/n, content_spool)
      assert_not_match(/#{Regexp::quote(@dot_test)}/n, content_spool)
    end
  end

  class TestLocalFileDocument < RUNIT::TestCase
    include ParseDate
    include Communicate

    def setup
      @root_local_dir = File::expand_path(File::join(Dir::pwd, 'test'))
      @doc_local_dir = File::join(@root_local_dir, 'doc')
      @content = <<-EOF
<html>
<head>
<title>Test</title>
</head>
<body>
<h1>Test</h1>
<p>Hello world.</p>
</body>
</html>
      EOF
      @root_index_local_path = File::join(@root_local_dir, 'index.html')
      @doc_index_local_path = File::join(@doc_local_dir, 'index.htm')
      @last_modified = Time::gm(2000, 11, 19, 12, 12, 12)
      @document = WWWsrv::LocalFileDocument::new(@root_local_dir)

      [ @root_local_dir,
	@doc_local_dir
      ].each do |path|
	Dir::mkdir(path)
      end

      [ @root_index_local_path,
	@doc_index_local_path,
      ].each do |path|
	File::open(path, 'w') do |index|
	  index.binmode
	  index.write(@content)
	end
	File::utime(@last_modified, @last_modified, path)
      end

      @log_iobuf = PseudoIO::new
      @srv_opt = WWWsrv::ServerOption::new
      @srv_opt.root = @document
      @srv_opt.log = WWWsrv::Logger::new(@log_iobuf)
      @request = WWWsrv::Request::new
      @request.version = WWWsrv::HTTP_1_1
      @request.header = WWWsrv::Header::new
      @request.header.host = 'localhost'
      @request.header.user_agent = 'Test/0.9 (with RUNIT)'
      @request.host = WWWsrv::Host::new('localhost')
      @request.content = PseudoIO::new
      @request.client = WWWsrv::Client::new('localhost', '127.0.0.1')
    end

    def teardown
      [ @doc_index_local_path,
	@root_index_local_path
      ].each do |path|
	File::unlink(path)
      end

      [ @doc_local_dir,
	@root_local_dir
      ].each do |path|
	Dir::rmdir(path)
      end
    end

    def test_new
      assert_instance_of(WWWsrv::LocalFileDocument, @document)
    end

    def test_head
      @request.method = 'HEAD'
      @request.path = '/index.html'
      response = WWWsrv::Response::new
      response.header.date = WWWsrv::HTMLconv::http_date(Time::now)
      response.header.server = WWWsrv::PRODUCT

      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@document, '1-1', @srv_opt, '', @request, response)

      assert_equal(:success, state)
      assert_equal(@root_index_local_path, doc_response.option.local_path)
      assert_equal(WWWsrv::HTTP_1_1, doc_response.version)
      assert_equal(200, doc_response.status)
      assert_equal('OK', doc_response.reason)
      assert_equal('text/html', doc_response.header.content_type)
      assert_not_nil(doc_response.header.content_length)
      assert_not_nil(doc_response.header.etag)
      assert_equal(@last_modified, Time::gm(*parsedate(doc_response.header.last_modified)[0, 6]))
      assert_nil(doc_content)
      assert(doc_content_alist.empty?)
    end

    def test_get
      @request.method = 'GET'
      @request.path = '/index.html'
      response = WWWsrv::Response::new
      response.header.date = WWWsrv::HTMLconv::http_date(Time::now)
      response.header.server = WWWsrv::PRODUCT

      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@document, '1-1', @srv_opt, '', @request, response)

      assert_equal(:success, state)
      assert_equal(@root_index_local_path, doc_response.option.local_path)
      assert_equal(WWWsrv::HTTP_1_1, doc_response.version)
      assert_equal(200, doc_response.status)
      assert_equal('OK', doc_response.reason)
      assert_equal('text/html', doc_response.header.content_type)
      assert_equal(doc_content.length.to_s, doc_response.header.content_length)
      assert_not_nil(doc_response.header.etag)
      assert_equal(@last_modified, Time::gm(*parsedate(doc_response.header.last_modified)[0, 6]))
      assert_equal(@content, doc_content)
    end

    def test_get_root
      @request.method = 'GET'
      @request.path = '/'
      response = WWWsrv::Response::new
      response.header.date = WWWsrv::HTMLconv::http_date(Time::now)
      response.header.server = WWWsrv::PRODUCT

      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@document, '1-1', @srv_opt, '', @request, response)

      assert_equal(:success, state)
      assert_equal(@root_index_local_path, doc_response.option.local_path)
      assert_equal(WWWsrv::HTTP_1_1, doc_response.version)
      assert_equal(200, doc_response.status)
      assert_equal('OK', doc_response.reason)
      assert_equal('text/html', doc_response.header.content_type)
      assert_equal(doc_content.length.to_s, doc_response.header.content_length)
      assert_not_nil(doc_response.header.etag)
      assert_equal(@last_modified, Time::gm(*parsedate(doc_response.header.last_modified)[0, 6]))
      assert_equal(@content, doc_content)

      response = WWWsrv::Response::new
      response.header.date = WWWsrv::HTMLconv::http_date(Time::now)
      response.header.server = WWWsrv::PRODUCT

      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@document, '1-1', @srv_opt, '', @request, response)

      assert_equal(:success, state)
      assert_equal(@root_index_local_path, doc_response.option.local_path)
      assert_equal(WWWsrv::HTTP_1_1, doc_response.version)
      assert_equal(200, doc_response.status)
      assert_equal('OK', doc_response.reason)
      assert_equal('text/html', doc_response.header.content_type)
      assert_equal(doc_content.length.to_s, doc_response.header.content_length)
      assert_not_nil(doc_response.header.etag)
      assert_equal(@last_modified, Time::gm(*parsedate(doc_response.header.last_modified)[0, 6]))
      assert_equal(@content, doc_content)
    end

    def test_dir
      @request.method = 'GET'
      @request.path = '/doc/'
      response = WWWsrv::Response::new
      response.header.date = WWWsrv::HTMLconv::http_date(Time::now)
      response.header.server = WWWsrv::PRODUCT

      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@document, '1-1', @srv_opt, '', @request, response)

      assert_equal(:success, state)
      assert_equal(@doc_index_local_path, doc_response.option.local_path)
      assert_equal(WWWsrv::HTTP_1_1, doc_response.version)
      assert_equal(200, doc_response.status)
      assert_equal('OK', doc_response.reason)
      assert_equal('text/html', doc_response.header.content_type)
      assert_equal(doc_content.length.to_s, doc_response.header.content_length)
      assert_not_nil(doc_response.header.etag)
      assert_equal(@last_modified, Time::gm(*parsedate(doc_response.header.last_modified)[0, 6]))

      @request.path = '/test/doc/'
      response = WWWsrv::Response::new
      response.header.date = WWWsrv::HTMLconv::http_date(Time::now)
      response.header.server = WWWsrv::PRODUCT

      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@document, '1-1', @srv_opt, '/test', @request, response)

      assert_equal(:success, state)
      assert_equal(@doc_index_local_path, doc_response.option.local_path)
      assert_equal(WWWsrv::HTTP_1_1, doc_response.version)
      assert_equal(200, doc_response.status)
      assert_equal('OK', doc_response.reason)
      assert_equal('text/html', doc_response.header.content_type)
      assert_equal(doc_content.length.to_s, doc_response.header.content_length)
      assert_not_nil(doc_response.header.etag)
      assert_equal(@last_modified, Time::gm(*parsedate(doc_response.header.last_modified)[0, 6]))

      response = WWWsrv::Response::new
      response.header.date = WWWsrv::HTMLconv::http_date(Time::now)
      response.header.server = WWWsrv::PRODUCT

      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@document, '1-1', @srv_opt, '/test/', @request, response)

      assert_equal(:success, state)
      assert_equal(@doc_index_local_path, doc_response.option.local_path)
      assert_equal(WWWsrv::HTTP_1_1, doc_response.version)
      assert_equal(200, doc_response.status)
      assert_equal('OK', doc_response.reason)
      assert_equal('text/html', doc_response.header.content_type)
      assert_equal(doc_content.length.to_s, doc_response.header.content_length)
      assert_not_nil(doc_response.header.etag)
      assert_equal(@last_modified, Time::gm(*parsedate(doc_response.header.last_modified)[0, 6]))
    end

    def test_dir_move
      @request.method = 'GET'
      @request.path = '/doc'
      response = WWWsrv::Response::new
      response.header.date = WWWsrv::HTMLconv::http_date(Time::now)
      response.header.server = WWWsrv::PRODUCT

      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@document, '1-1', @srv_opt, '', @request, response)

      assert_equal(:success, state)
      assert_nil(doc_response.option.local_path)
      assert_equal(WWWsrv::HTTP_1_1, doc_response.version)
      assert_equal(301, doc_response.status)
      assert_equal('Moved Permanently', doc_response.reason)
      assert_equal('http://localhost:80/doc/', doc_response.header.location)
      assert_nil(doc_response.header.content_length)

      @request.method = 'GET'
      @request.path = '/test'
      response = WWWsrv::Response::new
      response.header.date = WWWsrv::HTMLconv::http_date(Time::now)
      response.header.server = WWWsrv::PRODUCT

      state, err_args, doc_response, doc_content, doc_content_alist =
	communicate(@document, '1-1', @srv_opt, '/test', @request, response)

      assert_equal(:success, state)
      assert_nil(doc_response.option.local_path)
      assert_equal(WWWsrv::HTTP_1_1, doc_response.version)
      assert_equal(301, doc_response.status)
      assert_equal('Moved Permanently', doc_response.reason)
      assert_equal('http://localhost:80/test/', doc_response.header.location)
      assert_nil(doc_response.header.content_length)
    end
  end
end

if ($0 == __FILE__) then
  TestWWWsrv::constants.map{ |name|
    TestWWWsrv::const_get(name)
  }.select{ |const|
    (const.kind_of? Class) && (const < RUNIT::TestCase)
  }.each do |test_case|
    RUNIT::CUI::TestRunner::run(test_case::suite)
  end
end
