#!/usr/local/bin/ruby
# $Id: test_multipart.rb,v 1.5 2002/01/06 15:45:52 toki Exp $
# -
# Test of WWWsrv::Multipart module

require 'pseudo_io'
require 'wwwsrv/htmlconv'
require 'wwwsrv/form'
require 'runit/testcase'
require 'runit/cui/testrunner'

module TestWWWsrv
  class TestMultipart < RUNIT::TestCase
    def make_multipart_input(src)
      input = PseudoIO::new
      src.each do |line|
 	line.chomp!("\n")
 	line.chomp!("\r")
 	input.write(line + "\r\n")
      end

      input
    end
    private :make_multipart_input

    def test_parse_multipart
      content_type = 'multipart/form-data; boundary=AaB03x'
      content_reader = make_multipart_input(<<-EOF)
--AaB03x
Content-Disposition: form-data; name="submit"

OK
--AaB03x
Content-Disposition: form-data; name="date"

1975-11-19
--AaB03x
Content-Disposition: form-data; name="files"
Content-Type: multipart/mixed; boundary="Bb:C04y"

--Bb:C04y
Content-Disposition: file; filename="hello.html"
Content-Type: text/html

<html>
<body><p>Hello world.</p></body>
</html>

--Bb:C04y
Content-Disposition: file; filename="hello.css"
Content-Type: text/css

p { font-weight: bold; font-size: large }

--Bb:C04y--
--AaB03x--
      EOF

      assert_not_nil(multipart = WWWsrv::Multipart::parse_multipart(content_reader, content_type))

      header, body = multipart.shift
      assert_equal('form-data; name="submit"', header.content_disposition)
      assert_equal('OK', body)

      header, body = multipart.shift
      assert_equal('form-data; name="date"', header.content_disposition)
      assert_equal('1975-11-19', body)

      header, submultipart = multipart.shift
      assert_equal('form-data; name="files"', header.content_disposition)
      assert_equal('multipart/mixed; boundary="Bb:C04y"', header.content_type)

      header, body = submultipart.shift
      assert_equal('file; filename="hello.html"', header.content_disposition)
      assert_equal('text/html', header.content_type)
      assert_equal("<html>\r\n<body><p>Hello world.</p></body>\r\n</html>\r\n", body)

      header, body = submultipart.shift
      assert_equal('file; filename="hello.css"', header.content_disposition)
      assert_equal('text/css', header.content_type)
      assert_equal("p { font-weight: bold; font-size: large }\r\n", body)

      assert(submultipart.empty?)
      assert(multipart.empty?)
    end

    def test_parse_multipart_ln
      content_type = 'multipart/form-data; boundary=AaB03x'
      content_reader = make_multipart_input(<<-EOF)
--AaB03x
Content-Disposition: form-data; name="submit"

OK
--AaB03x
Content-Disposition: form-data; name="date"

1975-11-19
--AaB03x
Content-Disposition: form-data; name="files"
Content-Type: multipart/mixed; boundary="Bb:C04y"

--Bb:C04y
Content-Disposition: file; filename="hello.html"
Content-Type: text/html

<html>
<body><p>Hello world.</p></body>
</html>

--Bb:C04y
Content-Disposition: file; filename="hello.css"
Content-Type: text/css

p { font-weight: bold; font-size: large }

--Bb:C04y--
--AaB03x--
      EOF
      content_reader.data.gsub!(/\r\n/, "\n")

      assert_not_nil(multipart = WWWsrv::Multipart::parse_multipart(content_reader, content_type))

      header, body = multipart.shift
      assert_equal('form-data; name="submit"', header.content_disposition)
      assert_equal('OK', body)

      header, body = multipart.shift
      assert_equal('form-data; name="date"', header.content_disposition)
      assert_equal('1975-11-19', body)

      header, submultipart = multipart.shift
      assert_equal('form-data; name="files"', header.content_disposition)
      assert_equal('multipart/mixed; boundary="Bb:C04y"', header.content_type)

      header, body = submultipart.shift
      assert_equal('file; filename="hello.html"', header.content_disposition)
      assert_equal('text/html', header.content_type)
      assert_equal("<html>\n<body><p>Hello world.</p></body>\n</html>\n", body)

      header, body = submultipart.shift
      assert_equal('file; filename="hello.css"', header.content_disposition)
      assert_equal('text/css', header.content_type)
      assert_equal("p { font-weight: bold; font-size: large }\n", body)

      assert(submultipart.empty?)
      assert(multipart.empty?)
    end

    def test_parse_multipart_with_preamble_and_epilogue
      content_type = 'multipart/form-data; boundary=AaB03x'
      content_reader = make_multipart_input(<<-EOF)
This is preamble.
--AaB03x
Content-Disposition: form-data; name="submit"

OK
--AaB03x
Content-Disposition: form-data; name="date"

1975-11-19
--AaB03x
Content-Disposition: form-data; name="files"
Content-Type: multipart/mixed; boundary="Bb:C04y"

This is inner preamble.
--Bb:C04y
Content-Disposition: file; filename="hello.html"
Content-Type: text/html

<html>
<body><p>Hello world.</p></body>
</html>

--Bb:C04y
Content-Disposition: file; filename="hello.css"
Content-Type: text/css

p { font-weight: bold; font-size: large }

--Bb:C04y--
This is inner epilogue.
--AaB03x--
This is epilogue.
      EOF

      assert_not_nil(multipart = WWWsrv::Multipart::parse_multipart(content_reader, content_type))

      header, body = multipart.shift
      assert_equal('form-data; name="submit"', header.content_disposition)
      assert_equal('OK', body)

      header, body = multipart.shift
      assert_equal('form-data; name="date"', header.content_disposition)
      assert_equal('1975-11-19', body)

      header, submultipart = multipart.shift
      assert_equal('form-data; name="files"', header.content_disposition)
      assert_equal('multipart/mixed; boundary="Bb:C04y"', header.content_type)

      header, body = submultipart.shift
      assert_equal('file; filename="hello.html"', header.content_disposition)
      assert_equal('text/html', header.content_type)
      assert_equal("<html>\r\n<body><p>Hello world.</p></body>\r\n</html>\r\n", body)

      header, body = submultipart.shift
      assert_equal('file; filename="hello.css"', header.content_disposition)
      assert_equal('text/css', header.content_type)
      assert_equal("p { font-weight: bold; font-size: large }\r\n", body)

      assert(submultipart.empty?)
      assert(multipart.empty?)
      assert_equal("This is epilogue.\r\n", content_reader.read)
    end

    def test_parse_multipart_incomplete
      content_type = 'multipart/form-data; boundary=AaB03x'
      content_reader = make_multipart_input(<<-EOF)
--AaB03x
Content-Disposition: form-data; name="submit"

OK
--AaB03x
Content-Disposition: form-data; name="date"

1975-11-19
--AaB03x
Content-Disposition: form-data; name="files"
Content-Type: multipart/mixed; boundary="Bb:C04y"

--Bb:C04y
Content-Disposition: file; filename="hello.html"
Content-Type: text/html

<html>
<body><p>Hello world.</p></body>
</html>

--Bb:C04y
Content-Disposition: file; filename="hello.css"
Content-Type: text/css

p { font-weight: bold; font-size: large }

      EOF

      assert_exception(WWWsrv::MultipartParseError) { WWWsrv::Multipart::parse_multipart(content_reader, content_type) }
    end

    def test_parse_multipart_without_boundary
      content_type = 'multipart/form-data; boundary=AaB03x'
      content_reader = make_multipart_input(<<-EOF)
--AaB03x
Content-Disposition: form-data; name="submit"

OK
--AaB03x
Content-Disposition: form-data; name="date"

1975-11-19
--AaB03x
Content-Disposition: form-data; name="files"
Content-Type: multipart/mixed

--Bb:C04y
Content-Disposition: file; filename="hello.html"
Content-Type: text/html

<html>
<body><p>Hello world.</p></body>
</html>

--Bb:C04y
Content-Disposition: file; filename="hello.css"
Content-Type: text/css

p { font-weight: bold; font-size: large }

--Bb:C04y--
--AaB03x--
      EOF

      assert_exception(WWWsrv::MultipartParseError) { WWWsrv::Multipart::parse_multipart(content_reader, content_type) }
    end
  end
end

if ($0 == __FILE__) then
  TestWWWsrv::constants.map{ |name|
    TestWWWsrv::const_get(name)
  }.select{ |const|
    (const.kind_of? Class) && (const < RUNIT::TestCase)
  }.each do |test_case|
    RUNIT::CUI::TestRunner::run(test_case::suite)
  end
end
