#!/usr/local/bin/ruby
# $Id: test_request.rb,v 1.3 2002/01/01 15:39:37 toki Exp $
# -
# Test of WWWsrv::Host class
# Test of WWWsrv::Client class
# Test of WWWsrv::Request class

require 'pseudo_io'
require 'wwwsrv/htmlconv'
require 'wwwsrv/request'
require 'runit/testcase'
require 'runit/cui/testrunner'

module TestWWWsrv
  class TestHost < RUNIT::TestCase
    def setup
      @host = WWWsrv::Host::new('localhost')
      @host2 = WWWsrv::Host::new('localhost', 8888)
    end

    def test_new
      assert_instance_of(WWWsrv::Host, @host)
      assert_instance_of(WWWsrv::Host, @host2)
    end

    def test_host
      assert_equal('localhost', @host.name)
      assert_equal('localhost', @host2.name)
    end

    def test_port
      assert_equal(80, @host.port)
      assert_equal(8888, @host2.port)
    end

    def test_to_s
      assert_equal('localhost:80', @host.to_s)
      assert_equal('localhost:8888', @host2.to_s)
    end

    def test_equal
      host = WWWsrv::Host::new('localhost')
      assert(host == @host)
      assert(host != @host2)
    end
  end

  class TestClient < RUNIT::TestCase
    def setup
      @client = WWWsrv::Client::new('localhost', '127.0.0.1')
      @client2 = WWWsrv::Client::new('tester', '192.168.0.1')
    end

    def test_new
      assert_instance_of(WWWsrv::Client, @client)
      assert_instance_of(WWWsrv::Client, @client2)
    end

    def test_name
      assert_equal('localhost', @client.name)
      assert_equal('tester', @client2.name)
    end

    def test_address
      assert_equal('127.0.0.1', @client.address)
      assert_equal('192.168.0.1', @client2.address)
    end

    def test_to_s
      assert_equal('localhost(127.0.0.1)', @client.to_s)
      assert_equal('tester(192.168.0.1)', @client2.to_s)
    end

    def test_to_i
      assert_equal(0b01111111_00000000_00000000_00000001, @client.to_i)
      assert_equal(0b11000000_10101000_00000000_00000001, @client2.to_i)
    end

    def test_equal
      client = WWWsrv::Client::new('localhost', '127.0.0.1')
      assert(client == @client)
      assert(client != @client2)
    end
  end

  class TestRequest < RUNIT::TestCase
    def setup
      @header = WWWsrv::Header::new
      @header.host = 'localhost'
      @header.user_agent = 'Test/0.9 (with RUNIT)'
      @content = PseudoIO::new
      @host = WWWsrv::Host::new('localhost')
      @client = WWWsrv::Client::new('localhost', '127.0.0.1')
      @request = WWWsrv::Request::new('GET', '/index.html', WWWsrv::HTTP_1_0, @header)
      @request.content = @content
      @request.host = @host
      @request.client = @client
      @request_null = WWWsrv::Request::new
    end

    def test_new
      assert_instance_of(WWWsrv::Request, @request)
      assert_instance_of(WWWsrv::Request, @request_null)
    end

    def test_method
      assert_equal('GET', @request.method)
      assert_nil(@request_null.method)
      assert_equal('GET', @request_null.method = 'GET')
      assert_equal('GET', @request_null.method)
    end

    def test_path
      assert_equal('/index.html', @request.path)
      assert_nil(@request_null.path)
      assert_equal('/index.html', @request_null.path = '/index.html')
      assert_equal('/index.html', @request_null.path)
    end

    def test_version
      assert_equal(WWWsrv::HTTP_1_0, @request.version)
      assert_nil(@request_null.version)
      assert_equal(WWWsrv::HTTP_1_0, @request_null.version = WWWsrv::HTTP_1_0)
      assert_equal(WWWsrv::HTTP_1_0, @request_null.version)
    end

    def test_header
      assert_equal(@header, @request.header)
      assert_nil(@request_null.header)
      assert_equal(@header, @request_null.header = @header)
      assert_equal(@header, @request_null.header)
    end

    def test_host
      assert_equal(@host, @request.host)
      assert_nil(@request_null.host)
      assert_equal(@host, @request_null.host = @host)
      assert_equal(@host, @request_null.host)
    end

    def test_content
      assert_equal(@content, @request.content)
      assert_nil(@request_null.content)
      assert_equal(@content, @request_null.content = @content)
      assert_equal(@content, @request_null.content)
    end

    def test_client
      assert_equal(@client, @request.client)
      assert_nil(@request_null.client)
      assert_equal(@client, @request_null.client = @client)
      assert_equal(@client, @request_null.client)
    end

    def test_normalized_path
      assert_equal('/index.html', @request.normalized_path)
      @request.path = '/foo/./bar'
      assert_equal('/foo/bar', @request.normalized_path)
      @request.path = '/foo/../bar'
      assert_equal('/bar', @request.normalized_path)
      @request.path = '/foo/.'
      assert_equal('/foo/', @request.normalized_path)
      @request.path = '/foo/bar/..'
      assert_equal('/foo/', @request.normalized_path)
      @request.path = '/../foo'
      assert_equal('/foo', @request.normalized_path)
      @request.path = '/foo/.'
      assert_equal('/foo/', @request.normalized_path)
      @request.path = '/../foo'
      assert_equal('/foo', @request.normalized_path)
      @request.path = '/..'
      assert_equal('/', @request.normalized_path)
      @request.path = '/foo?bar'
      assert_equal('/foo', @request.normalized_path)
      @request.path = '/foo+bar'
      assert_equal('/foo bar', @request.normalized_path)
      @request.path = '/foo/%7Ebar'
      assert_equal('/foo/~bar', @request.normalized_path)
      assert_nil(@request_null.normalized_path)
    end

    def test_condition
      etag = '"1234-5678"'
      mtime = Time::gm(2001, 11, 19, 0, 0, 0)

      @request.header = WWWsrv::Header::new
      @request.header.if_match = '*'
      assert_equal(200, @request.condition(mtime, etag))

      @request.header = WWWsrv::Header::new
      @request.header.if_match = '"1234-5678"'
      assert_equal(200, @request.condition(mtime, etag))

      @request.header = WWWsrv::Header::new
      @request.header.if_match = '"8765-4321"'
      assert_equal(412, @request.condition(mtime, etag))

      @request.header = WWWsrv::Header::new
      @request.header.if_none_match = '*'
      assert_equal(304, @request.condition(mtime, etag))

      @request.header = WWWsrv::Header::new
      @request.header.if_none_match = '"1234-5678"'
      assert_equal(304, @request.condition(mtime, etag))

      @request.header = WWWsrv::Header::new
      @request.header.if_none_match = '"8765-4321"'
      assert_equal(200, @request.condition(mtime, etag))

      @request.header = WWWsrv::Header::new
      @request.header.if_modified_since = WWWsrv::HTMLconv::http_date(mtime)
      assert_equal(304, @request.condition(mtime, etag))

      @request.header = WWWsrv::Header::new
      @request.header.if_modified_since = WWWsrv::HTMLconv::http_date(mtime)
      assert_equal(200, @request.condition(mtime + 1, etag))

      @request.header = WWWsrv::Header::new
      @request.header.if_unmodified_since = WWWsrv::HTMLconv::http_date(mtime)
      assert_equal(200, @request.condition(mtime, etag))

      @request.header = WWWsrv::Header::new
      @request.header.if_unmodified_since = WWWsrv::HTMLconv::http_date(mtime)
      assert_equal(412, @request.condition(mtime + 1, etag))

      @request.header = WWWsrv::Header::new
      @request.header.if_none_match = '"8765-4321"'
      @request.header.if_modified_since = WWWsrv::HTMLconv::http_date(mtime)
      assert_equal(200, @request.condition(mtime, etag))

      @request.header = WWWsrv::Header::new
      @request.header.if_match = '"1234-5678"'
      @request.header.if_unmodified_since = WWWsrv::HTMLconv::http_date(mtime - 1)
      assert_equal(200, @request.condition(mtime, etag))
    end

    def test_params
      assert(@request.params.empty?)

      param_data = [
	%w[ double first ],
	%w[ double second ],
	%w[ debug true ]
      ].map{ |param_name, param_value|
	WWWsrv::HTMLconv::encodeURL(param_name) + '=' +
	  WWWsrv::HTMLconv::encodeURL(param_value)
      }.join('&')

      request = WWWsrv::Request::new('GET', '/index.html?' + param_data, WWWsrv::HTTP_1_0, @header)
      request.content = @content
      request.host = @host
      request.client = @client
      assert_equal('first', request.params['double'][0])
      assert_equal('first', request.form_data['double'][0][1])
      assert(request.form_data['double'][0][0].empty?)
      assert_equal('second', request.params['double'][1])
      assert_equal('second', request.form_data['double'][1][1])
      assert(request.form_data['double'][1][0].empty?)
      assert_equal('true', request.params['debug'][0])
      assert_equal('true', request.form_data['debug'][0][1])
      assert(request.form_data['debug'][0][0].empty?)

      @header.content_type = 'Application/X-WWW-Form-URLencoded'
      request = WWWsrv::Request::new('POST', '/index.html', WWWsrv::HTTP_1_0, @header)
      request.content = PseudoIO::new
      request.content.write(param_data)
      request.host = @host
      request.client = @client
      assert_equal('first', request.params['double'][0])
      assert_equal('first', request.form_data['double'][0][1])
      assert(request.form_data['double'][0][0].empty?)
      assert_equal('second', request.params['double'][1])
      assert_equal('second', request.form_data['double'][1][1])
      assert(request.form_data['double'][1][0].empty?)
      assert_equal('true', request.params['debug'][0])
      assert_equal('true', request.form_data['debug'][0][1])
      assert(request.form_data['debug'][0][0].empty?)
    end

    def make_multipart_input(src)
      input = PseudoIO::new
      src.each do |line|
 	line.chomp!("\n")
 	line.chomp!("\r")
 	input.write(line + "\r\n")
      end

      input
    end
    private :make_multipart_input

    def test_form_data
      assert(@request.form_data.empty?)

      @header.content_type = 'multipart/form-data; boundary=AaB03x'
      content_reader = make_multipart_input(<<-EOF)
--AaB03x
Content-Disposition: form-data; name="submit"

OK
--AaB03x
Content-Disposition: form-data; name="date"

1975-11-19
--AaB03x
Content-Disposition: form-data; name="files"
Content-Type: multipart/mixed; boundary="Bb:C04y"

--Bb:C04y
Content-Disposition: file; filename="hello.html"
Content-Type: text/html

<html>
<body><p>Hello world.</p></body>
</html>

--Bb:C04y
Content-Disposition: file; filename="hello.css"
Content-Type: text/css

p { font-weight: bold; font-size: large }

--Bb:C04y--
--AaB03x--
This is epilogue.
      EOF

      request = WWWsrv::Request::new('POST', '/index.html', WWWsrv::HTTP_1_1, @header)
      request.content = content_reader
      request.host = @host
      request.client = @client

      header, body = request.form_data['submit'].shift
      assert_equal('form-data; name="submit"', header.content_disposition)
      assert_equal('OK', body)
      assert_equal('OK', request.params['submit'].shift)

      header, body = request.form_data['date'].shift
      assert_equal('form-data; name="date"', header.content_disposition)
      assert_equal('1975-11-19', body)
      assert_equal('1975-11-19', request.params['date'].shift)

      header, submultipart = request.form_data['files'].shift
      subvalues = request.params['files'].shift
      assert_equal('form-data; name="files"', header.content_disposition)
      assert_equal('multipart/mixed; boundary="Bb:C04y"', header.content_type)

      header, body = submultipart.shift
      assert_equal('file; filename="hello.html"', header.content_disposition)
      assert_equal('text/html', header.content_type)
      assert_equal("<html>\r\n<body><p>Hello world.</p></body>\r\n</html>\r\n", body)
      assert_equal("<html>\r\n<body><p>Hello world.</p></body>\r\n</html>\r\n", subvalues.shift)

      header, body = submultipart.shift
      assert_equal('file; filename="hello.css"', header.content_disposition)
      assert_equal('text/css', header.content_type)
      assert_equal("p { font-weight: bold; font-size: large }\r\n", body)
      assert_equal("p { font-weight: bold; font-size: large }\r\n", subvalues.shift)

      assert(submultipart.empty?)
      assert(subvalues.empty?)
      request.form_data.each_key do |name|
	assert(request.form_data[name].empty?)
      end
      request.params.each_key do |name|
	assert(request.params[name].empty?)
      end
      assert_nil(request.content.read)
    end
  end

  class TestRequestParse < RUNIT::TestCase
    def setup
      @input = PseudoIO::new
    end

    def test_parse_request_line
      @input.write("GET / HTTP/1.0\r\n")
      @input.write("User-Agent: Test/0.9 (with RUNIT)\r\n")
      request_method, request_path, http_version = WWWsrv::Request::parse_line(@input)
      assert_equal("GET", request_method)
      assert_equal("/", request_path)
      assert_equal(WWWsrv::HTTP_1_0, http_version)
    end

    def test_parse_request_line_robustness
      @input.write("\r\n")
      @input.write("GET / HTTP/1.0\r\n")
      @input.write("Server: WWWsrv/0.0 (TEST)\r\n")
      request_method, request_path, http_version = WWWsrv::Request::parse_line(@input)
      assert_equal("GET", request_method)
      assert_equal("/", request_path)
      assert_equal(WWWsrv::HTTP_1_0, http_version)
    end

    def test_parse_request_line_new_line_sep
      @input.write("GET / HTTP/1.0\n")
      @input.write("User-Agent: Test/0.9 (with RUNIT)\n")
      request_method, request_path, http_version = WWWsrv::Request::parse_line(@input)
      assert_equal("GET", request_method)
      assert_equal("/", request_path)
      assert_equal(WWWsrv::HTTP_1_0, http_version)
    end

    def test_parse_request_line_old_http
      @input.write("GET /\r\n")
      @input.write("Content body...")
      request_method, request_path, http_version = WWWsrv::Request::parse_line(@input)
      assert_equal("GET", request_method)
      assert_equal("/", request_path)
      assert_equal(WWWsrv::HTTP_0_9, http_version)
    end

    def test_parse_request_line_invalid
      @input.write("DETARAME!\r\n")
      @input.write("User-Agent: Test/0.9 (with RUNIT)\r\n")
      assert_exception(WWWsrv::RequestLineParseError) { WWWsrv::Request::parse_line(@input) }
    end

    def test_parse_request_line_invalid_version
      @input.write("GET / HTTP/7.7.7\r\n")
      @input.write("User-Agent: Test/0.9 (with RUNIT)\r\n")
      assert_exception(WWWsrv::HTTPVersionFormatError) { WWWsrv::Request::parse_line(@input) }
    end

    def test_parse_request_line_empty
      request_method, request_path, http_version = WWWsrv::Request::parse_line(@input)
      assert_nil(request_method)
      assert_nil(request_path)
      assert_nil(http_version)
    end

    def test_parse_request
      @input.write("GET / HTTP/1.0\r\n")
      @input.write("Host: localhost\r\n")
      @input.write("User-Agent: Test/0.9 (with RUNIT)\r\n")
      @input.write("\r\n")
      @input.write("...body...")
      request = WWWsrv::Request::parse(@input)
      assert_instance_of(WWWsrv::Request, request)
      assert_equal('GET', request.method)
      assert_equal('/', request.path)
      assert_equal(WWWsrv::HTTP_1_0, request.version)
      assert_instance_of(WWWsrv::Header, request.header)
      assert_equal(2, request.header.length)
      assert_equal('localhost', request.header.host)
      assert_equal('Test/0.9 (with RUNIT)', request.header.user_agent)
    end
  end
end

if ($0 == __FILE__) then
  TestWWWsrv::constants.map{ |name|
    TestWWWsrv::const_get(name)
  }.select{ |const|
    (const.kind_of? Class) && (const < RUNIT::TestCase)
  }.each do |test_case|
    RUNIT::CUI::TestRunner::run(test_case::suite)
  end
end
