#!/usr/local/bin/ruby
# $Id: test_server.rb,v 1.4 2002/01/06 15:45:53 toki Exp $
# -
# Test of WWWsrv::ResponseWriter class

require 'pseudo_io'
require 'wwwsrv/server'
require 'runit/testcase'
require 'runit/cui/testrunner'

module TestWWWsrv
  class TestResponseWriter < RUNIT::TestCase
    def setup
      @output = PseudoIO::new
      @watcher = WWWsrv::ResponseWatcher::new
      @request = WWWsrv::Request::new('GET', '/index.html')
      @request.header = WWWsrv::Header::new
      @request.host = @host
      @response = WWWsrv::Response::new
    end

    def make_writer(version, *header)
      @request.version = version
      for field_name, field_value in header
	@request.header[field_name] = field_value
      end

      writer = WWWsrv::ResponseWriter::new(@output, @request, @watcher)
      assert_instance_of(WWWsrv::ResponseWriter, writer)
      assert(! writer.used?)

      writer
    end
    private :make_writer

    def test_http_1_0
      sid = '1-1'
      content = "Hello world.\n"
      @response.header.content_type = 'text/plain'
      wout = make_writer(WWWsrv::HTTP_1_0)
      wout.write(sid, @response)
      wout.write(sid, content)
      wout.terminate(sid)

      version, status, reason = WWWsrv::Parser::parse_status_line(@output)
      header = WWWsrv::Parser::parse_header(@output)
      assert_equal(WWWsrv::HTTP_1_1, version)
      assert_equal(200, status)
      assert_equal('OK', reason)
      assert_equal('text/plain', header.content_type)
      assert_equal(content.length.to_s, header.content_length)
      assert_equal('close', header.connection)
      assert_equal(content, @output.data)
    end

    def test_http_1_0_with_content_length
      sid = '1-1'
      content = "Hello world.\n"
      @response.header.content_type = 'text/plain'
      @response.header.content_length = content.length.to_s
      wout = make_writer(WWWsrv::HTTP_1_0)
      wout.write(sid, @response)
      wout.write(sid, content)
      wout.terminate(sid)

      version, status, reason = WWWsrv::Parser::parse_status_line(@output)
      header = WWWsrv::Parser::parse_header(@output)
      assert_equal(WWWsrv::HTTP_1_1, version)
      assert_equal(200, status)
      assert_equal('OK', reason)
      assert_equal('text/plain', header.content_type)
      assert_equal(content.length.to_s, header.content_length)
      assert_equal('close', header.connection)
      assert_equal(content, @output.data)
    end

    def test_http_1_0_connection_keep_alive_by_request
      sid = '1-1'
      content = "Hello world.\n"
      @response.header.content_type = 'text/plain'
      wout = make_writer(WWWsrv::HTTP_1_0, %w[ Connection Keep-Alive ])
      wout.write(sid, @response)
      wout.write(sid, content)
      wout.terminate(sid)

      version, status, reason = WWWsrv::Parser::parse_status_line(@output)
      header = WWWsrv::Parser::parse_header(@output)
      assert_equal(WWWsrv::HTTP_1_1, version)
      assert_equal(200, status)
      assert_equal('OK', reason)
      assert_equal('text/plain', header.content_type)
      assert_equal(content.length.to_s, header.content_length)
      assert_equal('Keep-Alive', header.connection)
      assert_equal(content, @output.data)
    end

    def test_http_1_0_connection_close_by_response
      sid = '1-1'
      content = "Hello world.\n"
      @response.header.content_type = 'text/plain'
      @response.header.connection = 'close'
      wout = make_writer(WWWsrv::HTTP_1_0, %w[ Connection Keep-Alive ])
      wout.write(sid, @response)
      wout.write(sid, content)
      wout.terminate(sid)

      version, status, reason = WWWsrv::Parser::parse_status_line(@output)
      header = WWWsrv::Parser::parse_header(@output)
      assert_equal(WWWsrv::HTTP_1_1, version)
      assert_equal(200, status)
      assert_equal('OK', reason)
      assert_equal('text/plain', header.content_type)
      assert_equal(content.length.to_s, header.content_length)
      assert_equal('close', header.connection)
      assert_equal(content, @output.data)
    end

    def test_http_1_1
      sid = '1-1'
      content = "Hello world.\n"
      @response.header.content_type = 'text/plain'
      wout = make_writer(WWWsrv::HTTP_1_1)
      wout.write(sid, @response)
      wout.write(sid, content)
      wout.terminate(sid)

      version, status, reason = WWWsrv::Parser::parse_status_line(@output)
      header = WWWsrv::Parser::parse_header(@output)
      assert_equal(WWWsrv::HTTP_1_1, version)
      assert_equal(200, status)
      assert_equal('OK', reason)
      assert_equal('text/plain', header.content_type)
      assert_equal('chunked', header.transfer_encoding)
      assert_nil(header.content_length)
      assert_nil(header.connection)

      chunked_content = []
      chunked_content.push([ content, [] ])
      chunked_content.push([ nil, [] ])
      WWWsrv::Parser::each_chunk(@output) do |result_data, result_extension|
	assert(! chunked_content.empty?)
	chunk_data, chunk_extension = chunked_content.shift
	assert_equal(chunk_data, result_data)
	assert_equal(chunk_extension, result_extension)
      end
      trailer = WWWsrv::Parser::parse_header(@output)
      assert(trailer.empty?)
      assert_nil(@output.data)
    end

    def test_http_1_1_with_chunk_extension
      sid = '1-1'
      content = "Hello world.\n"
      @response.header.content_type = 'text/plain'
      wout = make_writer(WWWsrv::HTTP_1_1)
      wout.write(sid, @response)
      wout.write(sid, content, 'debug')
      wout.write(sid, nil, [ 'test', 'true' ])
      wout.terminate(sid)

      version, status, reason = WWWsrv::Parser::parse_status_line(@output)
      header = WWWsrv::Parser::parse_header(@output)
      assert_equal(WWWsrv::HTTP_1_1, version)
      assert_equal(200, status)
      assert_equal('OK', reason)
      assert_equal('text/plain', header.content_type)
      assert_equal('chunked', header.transfer_encoding)
      assert_nil(header.content_length)
      assert_nil(header.connection)

      chunked_content = []
      chunked_content.push([ content, [ [ 'debug', nil ] ] ])
      chunked_content.push([ nil, [ [ 'test', 'true' ] ] ])
      WWWsrv::Parser::each_chunk(@output) do |result_data, result_extension|
	assert(! chunked_content.empty?)
	chunk_data, chunk_extension = chunked_content.shift
	assert_equal(chunk_data, result_data)
	assert_equal(chunk_extension, result_extension)
      end
      trailer = WWWsrv::Parser::parse_header(@output)
      assert(trailer.empty?)
      assert_nil(@output.data)
    end

    def test_http_1_1_with_trailer
      sid = '1-1'
      content = "Hello world.\n"
      @response.header.content_type = 'text/plain'
      @response.trailer.x_debug = 'true'
      wout = make_writer(WWWsrv::HTTP_1_1)
      wout.write(sid, @response)
      wout.write(sid, content)
      wout.terminate(sid)

      version, status, reason = WWWsrv::Parser::parse_status_line(@output)
      header = WWWsrv::Parser::parse_header(@output)
      assert_equal(WWWsrv::HTTP_1_1, version)
      assert_equal(200, status)
      assert_equal('OK', reason)
      assert_equal('text/plain', header.content_type)
      assert_equal('chunked', header.transfer_encoding)
      assert_nil(header.content_length)
      assert_nil(header.connection)

      chunked_content = []
      chunked_content.push([ content, [] ])
      chunked_content.push([ nil, [] ])
      WWWsrv::Parser::each_chunk(@output) do |result_data, result_extension|
	assert(! chunked_content.empty?)
	chunk_data, chunk_extension = chunked_content.shift
	assert_equal(chunk_data, result_data)
	assert_equal(chunk_extension, result_extension)
      end
      trailer = WWWsrv::Parser::parse_header(@output)
      assert_equal('true', trailer.x_debug)
      assert_nil(@output.data)
    end

    def test_http_1_1_connection_close_by_request
      sid = '1-1'
      content = "Hello world.\n"
      @response.header.content_type = 'text/plain'
      wout = make_writer(WWWsrv::HTTP_1_1, %w[ Connection close ])
      wout.write(sid, @response)
      wout.write(sid, content)
      wout.terminate(sid)

      version, status, reason = WWWsrv::Parser::parse_status_line(@output)
      header = WWWsrv::Parser::parse_header(@output)
      assert_equal(WWWsrv::HTTP_1_1, version)
      assert_equal(200, status)
      assert_equal('OK', reason)
      assert_equal('text/plain', header.content_type)
      assert_equal('chunked', header.transfer_encoding)
      assert_nil(header.content_length)
      assert_equal('close', header.connection)

      chunked_content = []
      chunked_content.push([ content, [] ])
      chunked_content.push([ nil, [] ])
      WWWsrv::Parser::each_chunk(@output) do |result_data, result_extension|
	assert(! chunked_content.empty?)
	chunk_data, chunk_extension = chunked_content.shift
	assert_equal(chunk_data, result_data)
	assert_equal(chunk_extension, result_extension)
      end
      trailer = WWWsrv::Parser::parse_header(@output)
      assert(trailer.empty?)
      assert_nil(@output.data)
    end

    def test_http_1_1_connection_close_by_response
      sid = '1-1'
      content = "Hello world.\n"
      @response.header.content_type = 'text/plain'
      @response.header.connection = 'close'
      wout = make_writer(WWWsrv::HTTP_1_1)
      wout.write(sid, @response)
      wout.write(sid, content)
      wout.terminate(sid)

      version, status, reason = WWWsrv::Parser::parse_status_line(@output)
      header = WWWsrv::Parser::parse_header(@output)
      assert_equal(WWWsrv::HTTP_1_1, version)
      assert_equal(200, status)
      assert_equal('OK', reason)
      assert_equal('text/plain', header.content_type)
      assert_equal('chunked', header.transfer_encoding)
      assert_nil(header.content_length)
      assert_equal('close', header.connection)

      chunked_content = []
      chunked_content.push([ content, [] ])
      chunked_content.push([ nil, [] ])
      WWWsrv::Parser::each_chunk(@output) do |result_data, result_extension|
	assert(! chunked_content.empty?)
	chunk_data, chunk_extension = chunked_content.shift
	assert_equal(chunk_data, result_data)
	assert_equal(chunk_extension, result_extension)
      end
      trailer = WWWsrv::Parser::parse_header(@output)
      assert(trailer.empty?)
      assert_nil(@output.data)
    end

    def test_location
      sid = '1-1'
      content = "Hello world.\n"
      @response.header.content_type = 'text/plain'
      @response.header.location = 'http://localhost:8888/hello.html'
      wout = make_writer(WWWsrv::HTTP_1_0)
      wout.write(sid, @response)
      wout.write(sid, content)
      wout.terminate(sid)

      version, status, reason = WWWsrv::Parser::parse_status_line(@output)
      header = WWWsrv::Parser::parse_header(@output)
      assert_equal(WWWsrv::HTTP_1_1, version)
      assert_equal(200, status)
      assert_equal('OK', reason)
      assert_equal('text/plain', header.content_type)
      assert_equal(content.length.to_s, header.content_length)
      assert_equal('close', header.connection)
      assert_equal('http://localhost:8888/hello.html', header.location)
      assert_equal(content, @output.data)
    end
  end
end

if ($0 == __FILE__) then
  TestWWWsrv::constants.map{ |name|
    TestWWWsrv::const_get(name)
  }.select{ |const|
    (const.kind_of? Class) && (const < RUNIT::TestCase)
  }.each do |test_case|
    RUNIT::CUI::TestRunner::run(test_case::suite)
  end
end
