#!/usr/local/bin/ruby
# $Id: test_ssi.rb,v 1.2 2002/01/01 13:21:13 toki Exp $
# -
# Test of WWWsrv::SSIdocument class

require 'comm'
require 'pseudo_io'
require 'pseudo_doc'
require 'wwwsrv/logger'
require 'wwwsrv/request'
require 'wwwsrv/response'
require 'wwwsrv/local'
require 'wwwsrv/ssi'
require 'runit/testcase'
require 'runit/cui/testrunner'

module TestWWWsrv
  class TestSSIdocument < RUNIT::TestCase
    include Communicate

    def setup
      @root = WWWsrv::DocumentFolder::new
      @local_file = WWWsrv::LocalFileDocument::new(Dir::getwd)
      @root.mount(@local_file, '/')
      @ssi = WWWsrv::SSIdocument::new(@local_file)
      @ssi_path = 'test.shtml'
      @root.mount(@ssi, '/', @ssi_path)

      @log_iobuf = PseudoIO::new
      @srv_opt = WWWsrv::ServerOption::new
      @srv_opt.root = @root
      @srv_opt.log = WWWsrv::Logger::new(@log_iobuf, WWWsrv::Logger::DEBUG)

      @request = WWWsrv::Request::new('GET', '/' + @ssi_path, WWWsrv::HTTP_1_1)
      @request.header = WWWsrv::Header::new
      @request.header.host = 'localhost'
      @request.header.user_agent = 'Test/0.9 (with RUNIT)'
      @request.content = PseudoIO::new
      @request.host = WWWsrv::Host::new('localhost')
      @request.client = WWWsrv::Client::new('localhost', '127.0.0.1')
    end

    def create_file(path, content)
      File::open(path, 'w') do |output|
	output.binmode
	output.write(content)
      end
    end
    private :create_file

    def test_content_include_file
      create_file(@ssi_path, <<-'EOF')
<!-- #include file="hello.txt" -->
      EOF
      create_file('hello.txt', "Hello world.\n")
      begin
	response = WWWsrv::Response::new
	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@root, @sid, @srv_opt, '', @request, response)
	pattern = Regexp::new(Regexp::quote("Hello world.\n"), 'n')
	assert_match(pattern, doc_content)
      ensure
	File::unlink(@ssi_path)
	File::unlink('hello.txt')
      end
    end

    def test_content_include_virtual
      create_file(@ssi_path, <<-'EOF')
<!-- #include virtual="hello.txt" -->
      EOF
      create_file('hello.txt', "Hello world.\n")
      begin
	response = WWWsrv::Response::new
	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@root, @sid, @srv_opt, '', @request, response)
	pattern = Regexp::new(Regexp::quote("Hello world.\n"), 'n')
	assert_match(pattern, doc_content)
      ensure
	File::unlink(@ssi_path)
	File::unlink('hello.txt')
      end
    end

    def test_content_echo_var
      create_file(@ssi_path, <<-'EOF')
<!-- #echo var="PATH" -->
      EOF
      begin
	response = WWWsrv::Response::new
	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@root, @sid, @srv_opt, '', @request, response)
	pattern = Regexp::new(Regexp::quote(ENV['PATH']), 'n')
	assert_match(pattern, doc_content)
      ensure
	File::unlink(@ssi_path)
      end

      create_file(@ssi_path, <<-'EOF')
<!-- #echo var="DOCUMENT_NAME" -->
      EOF
      begin
	response = WWWsrv::Response::new
	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@root, @sid, @srv_opt, '', @request, response)
	pattern = Regexp::new(Regexp::quote(@ssi_path), 'n')
	assert_match(pattern, doc_content)
      ensure
	File::unlink(@ssi_path)
      end

      create_file(@ssi_path, <<-'EOF')
<!-- #echo var="DOCUMENT_URI" -->
      EOF
      begin
	response = WWWsrv::Response::new
	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@root, @sid, @srv_opt, '', @request, response)
	pattern = Regexp::new(Regexp::quote('/' + @ssi_path), 'n')
	assert_match(pattern, doc_content)
      ensure
	File::unlink(@ssi_path)
      end

      create_file(@ssi_path, <<-'EOF')
<!-- #echo var="LAST_MODIFIED" -->
      EOF
      begin
	response = WWWsrv::Response::new
	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@root, @sid, @srv_opt, '', @request, response)
	mtime = File::stat(@ssi_path).mtime.gmtime
	pattern = Regexp::new(Regexp::quote(mtime.to_s), 'n')
	assert_match(pattern, doc_content)
      ensure
	File::unlink(@ssi_path)
      end

      create_file(@ssi_path, <<-'EOF')
<!-- #config timefmt="%y-%m-%d %H:%M:%S" -->
<!-- #echo var="LAST_MODIFIED" -->
      EOF
      begin
	response = WWWsrv::Response::new
	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@root, @sid, @srv_opt, '', @request, response)
	mtime = File::stat(@ssi_path).mtime.gmtime
	pattern = Regexp::new(Regexp::quote(mtime.strftime("%y-%m-%d %H:%M:%S")), 'n')
	assert_match(pattern, doc_content)
      ensure
	File::unlink(@ssi_path)
      end

      create_file(@ssi_path, <<-'EOF')
<!-- #echo var="SERVER_SOFTWARE" -->
      EOF
      begin
	response = WWWsrv::Response::new
	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@root, @sid, @srv_opt, '', @request, response)
	pattern = Regexp::new(Regexp::quote(WWWsrv::PRODUCT), 'n')
	assert_match(pattern, doc_content)
      ensure
	File::unlink(@ssi_path)
      end
    end

    def test_content_fsize_file
      create_file(@ssi_path, <<-'EOF')
<!-- #fsize file="test.dat" -->
      EOF
      File::open('test.dat', 'w') do |output|
	output.binmode
	output.seek(1023, IO::SEEK_SET)
	output.write('z')
      end
      begin
	response = WWWsrv::Response::new
	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@root, @sid, @srv_opt, '', @request, response)
	pattern = Regexp::new(Regexp::quote('1024'), 'n')
	assert_match(pattern, doc_content)
      ensure
	File::unlink(@ssi_path)
	File::unlink('test.dat')
      end

      create_file(@ssi_path, <<-'EOF')
<!-- #config sizefmt="bytes" -->
<!-- #fsize file="test.dat" -->
      EOF
      File::open('test.dat', 'w') do |output|
	output.binmode
	output.seek(1023, IO::SEEK_SET)
	output.write('z')
      end
      begin
	response = WWWsrv::Response::new
	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@root, @sid, @srv_opt, '', @request, response)
	pattern = Regexp::new(Regexp::quote('1024'), 'n')
	assert_match(pattern, doc_content)
      ensure
	File::unlink(@ssi_path)
	File::unlink('test.dat')
      end

      create_file(@ssi_path, <<-'EOF')
<!-- #config sizefmt="abbrev" -->
<!-- #fsize file="test.dat" -->
      EOF
      File::open('test.dat', 'w') do |output|
	output.binmode
	output.seek(127, IO::SEEK_SET)
	output.write('z')
      end
      begin
	response = WWWsrv::Response::new
	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@root, @sid, @srv_opt, '', @request, response)
	pattern = Regexp::new(Regexp::quote('128'), 'n')
	assert_match(pattern, doc_content)
      ensure
	File::unlink(@ssi_path)
	File::unlink('test.dat')
      end

      create_file(@ssi_path, <<-'EOF')
<!-- #config sizefmt="abbrev" -->
<!-- #fsize file="test.dat" -->
      EOF
      File::open('test.dat', 'w') do |output|
	output.binmode
	output.seek(1024, IO::SEEK_SET)
	output.write('z')
      end
      begin
	response = WWWsrv::Response::new
	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@root, @sid, @srv_opt, '', @request, response)
	pattern = Regexp::new(Regexp::quote('1'), 'n')
	assert_match(pattern, doc_content)
      ensure
	File::unlink(@ssi_path)
	File::unlink('test.dat')
      end

      create_file(@ssi_path, <<-'EOF')
<!-- #config sizefmt="abbrev" -->
<!-- #fsize file="test.dat" -->
      EOF
      File::open('test.dat', 'w') do |output|
	output.binmode
	output.seek(1024**2, IO::SEEK_SET)
	output.write('z')
      end
      begin
	response = WWWsrv::Response::new
	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@root, @sid, @srv_opt, '', @request, response)
	pattern = Regexp::new(Regexp::quote('1'), 'n')
	assert_match(pattern, doc_content)
      ensure
	File::unlink(@ssi_path)
	File::unlink('test.dat')
      end
    end

    def test_content_fsize_virtual
      create_file(@ssi_path, <<-'EOF')
<!-- #fsize virtual="test.dat" -->
      EOF
      File::open('test.dat', 'w') do |output|
	output.binmode
	output.seek(1023, IO::SEEK_SET)
	output.write('z')
      end
      begin
	response = WWWsrv::Response::new
	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@root, @sid, @srv_opt, '', @request, response)
	pattern = Regexp::new(Regexp::quote('1024'), 'n')
	assert_match(pattern, doc_content)
      ensure
	File::unlink(@ssi_path)
	File::unlink('test.dat')
      end

      create_file(@ssi_path, <<-'EOF')
<!-- #config sizefmt="bytes" -->
<!-- #fsize virtual="test.dat" -->
      EOF
      File::open('test.dat', 'w') do |output|
	output.binmode
	output.seek(1023, IO::SEEK_SET)
	output.write('z')
      end
      begin
	response = WWWsrv::Response::new
	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@root, @sid, @srv_opt, '', @request, response)
	pattern = Regexp::new(Regexp::quote('1024'), 'n')
	assert_match(pattern, doc_content)
      ensure
	File::unlink(@ssi_path)
	File::unlink('test.dat')
      end

      create_file(@ssi_path, <<-'EOF')
<!-- #config sizefmt="abbrev" -->
<!-- #fsize virtual="test.dat" -->
      EOF
      File::open('test.dat', 'w') do |output|
	output.binmode
	output.seek(127, IO::SEEK_SET)
	output.write('z')
      end
      begin
	response = WWWsrv::Response::new
	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@root, @sid, @srv_opt, '', @request, response)
	pattern = Regexp::new(Regexp::quote('128'), 'n')
	assert_match(pattern, doc_content)
      ensure
	File::unlink(@ssi_path)
	File::unlink('test.dat')
      end

      create_file(@ssi_path, <<-'EOF')
<!-- #config sizefmt="abbrev" -->
<!-- #fsize virtual="test.dat" -->
      EOF
      File::open('test.dat', 'w') do |output|
	output.binmode
	output.seek(1024, IO::SEEK_SET)
	output.write('z')
      end
      begin
	response = WWWsrv::Response::new
	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@root, @sid, @srv_opt, '', @request, response)
	pattern = Regexp::new(Regexp::quote('1'), 'n')
	assert_match(pattern, doc_content)
      ensure
	File::unlink(@ssi_path)
	File::unlink('test.dat')
      end

      create_file(@ssi_path, <<-'EOF')
<!-- #config sizefmt="abbrev" -->
<!-- #fsize virtual="test.dat" -->
      EOF
      File::open('test.dat', 'w') do |output|
	output.binmode
	output.seek(1024**2, IO::SEEK_SET)
	output.write('z')
      end
      begin
	response = WWWsrv::Response::new
	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@root, @sid, @srv_opt, '', @request, response)
	pattern = Regexp::new(Regexp::quote('1'), 'n')
	assert_match(pattern, doc_content)
      ensure
	File::unlink(@ssi_path)
	File::unlink('test.dat')
      end
    end

    def test_content_flastmod_file
      create_file(@ssi_path, <<-'EOF')
<!-- #flastmod file="hello.txt" -->
      EOF
      create_file('hello.txt', "Hello world.\n")
      begin
	response = WWWsrv::Response::new
	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@root, @sid, @srv_opt, '', @request, response)
	mtime = File::stat('hello.txt').mtime.gmtime
	pattern = Regexp::new(Regexp::quote(mtime.to_s), 'n')
	assert_match(pattern, doc_content)
      ensure
	File::unlink(@ssi_path)
	File::unlink('hello.txt')
      end

      create_file(@ssi_path, <<-'EOF')
<!-- #config timefmt="%y-%m-%d %H:%M:%S" -->
<!-- #flastmod file="hello.txt" -->
      EOF
      create_file('hello.txt', "Hello world.\n")
      begin
	response = WWWsrv::Response::new
	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@root, @sid, @srv_opt, '', @request, response)
	mtime = File::stat('hello.txt').mtime.gmtime
	pattern = Regexp::new(Regexp::quote(mtime.strftime("%y-%m-%d %H:%M:%S")), 'n')
	assert_match(pattern, doc_content)
      ensure
	File::unlink(@ssi_path)
	File::unlink('hello.txt')
      end
    end

    def test_content_flastmod_virtual
      create_file(@ssi_path, <<-'EOF')
<!-- #flastmod virtual="hello.txt" -->
      EOF
      create_file('hello.txt', "Hello world.\n")
      begin
	response = WWWsrv::Response::new
	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@root, @sid, @srv_opt, '', @request, response)
	mtime = File::stat('hello.txt').mtime.gmtime
	pattern = Regexp::new(Regexp::quote(mtime.to_s), 'n')
	assert_match(pattern, doc_content)
      ensure
	File::unlink(@ssi_path)
	File::unlink('hello.txt')
      end

      create_file(@ssi_path, <<-'EOF')
<!-- #config timefmt="%y-%m-%d %H:%M:%S" -->
<!-- #flastmod virtual="hello.txt" -->
      EOF
      create_file('hello.txt', "Hello world.\n")
      begin
	response = WWWsrv::Response::new
	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@root, @sid, @srv_opt, '', @request, response)
	mtime = File::stat('hello.txt').mtime.gmtime
	pattern = Regexp::new(Regexp::quote(mtime.strftime("%y-%m-%d %H:%M:%S")), 'n')
	assert_match(pattern, doc_content)
      ensure
	File::unlink(@ssi_path)
	File::unlink('hello.txt')
      end
    end

    def test_content_exec_cmd
      create_file(@ssi_path, <<-'EOF')
<!-- #exec cmd="ruby -e 'puts %q{Hello world.}'" -->
      EOF
      begin
	response = WWWsrv::Response::new
	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@root, @sid, @srv_opt, '', @request, response)
	pattern = Regexp::new(Regexp::quote(`ruby -e 'puts %q{Hello world.}'`), 'n')
	assert_match(pattern, doc_content)
      ensure
	File::unlink(@ssi_path)
      end
    end

    def test_content_exec_cgi
      create_file(@ssi_path, <<-'EOF')
<!-- #exec cgi="hello.cgi" -->
      EOF
      create_file('hello.cgi', <<-'EOF')
#!/usr/local/bin/ruby
msg = "Hello world.\n"
print "Content-Type: text/plain\r\n"
print "Content-Length: #{msg.length}\r\n"
print "\r\n"
print msg
      EOF
      cgi = WWWsrv::CGIdocument::new(Dir::getwd, false, 'ruby')
      @root.mount(cgi, '/', 'hello.cgi')
      begin
	response = WWWsrv::Response::new
	state, err_args, doc_response, doc_content, doc_content_alist =
	  communicate(@root, @sid, @srv_opt, '', @request, response)
	pattern = Regexp::new(Regexp::quote("Hello world.\n"), 'n')
	assert_match(pattern, doc_content)
      ensure
	File::unlink(@ssi_path)
	File::unlink('hello.cgi')
      end
    end
  end
end

if ($0 == __FILE__) then
  TestWWWsrv::constants.map{ |name|
    TestWWWsrv::const_get(name)
  }.select{ |const|
    (const.kind_of? Class) && (const < RUNIT::TestCase)
  }.each do |test_case|
    RUNIT::CUI::TestRunner::run(test_case::suite)
  end
end
