#!/usr/local/bin/ruby
# $Id: wwwsrv.rb,v 1.468.2.2 2002/02/27 14:16:39 toki Exp $
# -
# World Wide Web Server

=begin
Copyright (c) 2001-2002
	TOKI Yoshinori <toki@freedom.ne.jp>. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1.  Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

  2.  Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

  3.  The name of the author may not be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.
=end

require 'socket'
require 'wwwsrv/server'

autoload :Etc, 'etc'

module WWWsrv
  class Daemon
    def self::take_off
      catch(:RUN_DAEMON) do
	unless (fork) then
	  Process::setsid
	  unless (fork) then
	    Dir::chdir('/')
	    File::umask(077)
	    STDIN.reopen('/dev/null', 'r')
	    STDOUT.reopen('/dev/null', 'w')
	    STDERR.reopen('/dev/null', 'w')
	    throw :RUN_DAEMON
	  end
	end
	exit!
      end
    end

    def initialize
      @option = {}
      @server = nil
      @srv_sock = nil
      @access_control_list = []
      @alias_list = []
      @document_list = []
      @config_path = nil
      @base_dir = nil
      @prefork = false
      @host_conf = ''
      @serv_conf = ''
      @init_conf = ''
      @final_conf = ''
      @warm_up_conf = ''
      @cool_down_conf = ''
      @host_mod = nil
      @serv_mod = nil
      @init_mod = nil
      @final_mod = nil
      @warm_up_mod = nil
      @cool_down_mod = nil
      @run_sem = Semaphore::new
    end

    attr_accessor :config_path
    attr_accessor :base_dir

    DEF_OPT = {
      :hostname      => 'localhost',
      :port          => 8888,
      :strict        => false,
      :log_file      => nil,
      :logging_level => Logger::MESSAGE,
      :display_level => Logger::MESSAGE,
      :request_line_logging_priority     => nil,
      :request_header_logging_priority   => nil,
      :request_content_logging_priority  => nil,
      :request_trailer_logging_priority  => nil,
      :status_line_logging_priority      => nil,
      :response_header_logging_priority  => nil,
      :response_content_logging_priority => nil,
      :response_trailer_logging_priority => nil,
      :effective_user_id  => 'nobody',
      :effective_group_id => 'nobody',
      :timeout       => Server::TIMEOUT,
      :keep_alive    => Server::KEEP_ALIVE,
      :connections   => Server::MAX_CONNECTIONS,
      :write_buffer  => Server::BUFFER_SIZE,
      :prefork_processes => 0,
      :quiet         => false,
      :debug         => false
    }

    CONST_OPT = {}
    DEF_OPT.each_key do |opt_id|
      opt_name = opt_id.id2name
      CONST_OPT[opt_id] = opt_name.upcase.intern

      eval <<-EOF
        def #{opt_name}
	  @option[:#{opt_name}]
	end

	def #{opt_name}=(opt_value)
	  @option[:#{opt_name}] = opt_value
	end
      EOF
    end

    def getopt(opt_id, opt_mod=nil)
      opt_val = DEF_OPT[opt_id]
      if (opt_mod && (opt_mod.const_defined? CONST_OPT[opt_id])) then
	opt_val = opt_mod.const_get(CONST_OPT[opt_id])
      end
      if (@option.include? opt_id) then
	opt_val = @option[opt_id]
      end

      opt_val
    end
    private :getopt

    def add_access_control(action, filter)
      @access_control_list.push([ action, filter ])
    end

    def add_alias(dst_path, src_path)
      @alias_list.push([ dst_path, src_path ])
    end

    def add_document(&doc_maker)
      @document_list.push(doc_maker)
      self
    end

    def load_conf
      @host_conf = ''
      @serv_conf = ''
      @init_conf = ''
      @final_conf = ''
      @warm_up_conf = ''
      @cool_down_conf = ''
      state = :head
      File::open(@config_path) do |input|
	for line in input
	  case (line)
	  when /^\s*__HOST__\s*$/n
	    state = :host
	  when /^\s*__SERVER__\s*$/n
	    state = :server
	  when /^\s*__INIT__\s*/n
	    state = :init
	  when /^\s*__FINAL__\s*/n
	    state = :final
	  when /^\s*__WARM_UP__\s*$/n
	    state = :warm_up
	  when /^\s*__COOL_DOWN__\s*$/n
	    state = :cool_down
	  else
	    case (state)
	    when :host
	      @host_conf.concat(line)
	    when :server
	      @serv_conf.concat(line)
	    when :init
	      @init_conf.concat(line)
	    when :final
	      @final_conf.concat(line)
	    when :warm_up
	      @warm_up_conf.concat(line)
	    when :cool_down
	      @cool_down_conf.concat(line)
	    end
	  end
	end
      end
    end
    private :load_conf

    def eval_conf(sect, mod, conf)
      begin
	LOCK.synchronize do
	  mod.module_eval(conf)
	end
      rescue StandardError, ScriptError
	unless (STDERR.closed?) then
	  STDERR.print("#{sect} section error: #{$!}\n")
	end
      end
    end
    private :eval_conf

    def copy_const(mod_dst, mod_src)
      mod_dst.module_eval{ include mod_src } if mod_src
    end
    private :copy_const

    def init
      load_conf if @config_path

      unless (@init_mod) then
	@init_mod = Module::new
	@init_mod.const_set(:BASE_DIR, @base_dir)
	eval_conf('__INIT__', @init_mod, @init_conf)
      end

      @warm_up_mod = Module::new
      copy_const(@warm_up_mod, @init_mod)
      @init_mod = nil
      @warm_up_mod.const_set(:BASE_DIR, @base_dir)
      eval_conf('__WARM_UP__', @warm_up_mod, @warm_up_conf)

      @host_mod = Module::new
      copy_const(@host_mod, @warm_up_mod)
      @warm_up_mod = nil
      eval_conf('__HOST__', @host_mod, @host_conf)

      if (getopt(:debug, @host_mod)) then
	alert = DebugAlertDocument::new
      elsif (@host_mod.const_defined? :ALERT) then
	alert = @host_mod.const_get(:ALERT)
      elsif (@host_mod.const_defined? :ALARM) then
	alert = @host_mod.const_get(:ALARM) # for backward compatibility
      else
	alert = AlertDocument::new
      end

      unless (@srv_sock) then
	if (getopt(:strict, @host_mod)) then
	  @srv_sock = TCPserver::open(getopt(:hostname, @host_mod), getopt(:port, @host_mod))
	else
	  @srv_sock = TCPserver::open(getopt(:port, @host_mod))
	end
      end
      @server = Server::new(@srv_sock, alert)

      if (Process::euid == 0) then
	begin
	  if (euid = getopt(:effective_user_id, @host_mod)) then
	    unless (euid.kind_of? Numeric) then
	      if (euid =~ /^\d+$/n) then
		euid = euid.to_i
	      else
		euid = Etc::getpwnam(euid).uid
	      end
	    end
	  end

	  if (egid = getopt(:effective_group_id, @host_mod)) then
	    unless (egid.kind_of? Numeric) then
	      if (egid =~ /^\d+$/n) then
		egid = egid.to_i
	      else
		egid = Etc::getgrnam(egid).gid
	      end
	    end
	  end

	  Process.egid = egid if (Process.egid != egid)
	  Process.euid = euid if (Process.euid != euid)
	rescue LoadError
	  print "warning: #{$!}\n"
	end
      end

      if (@host_mod.const_defined? :LOGGER) then
	@server.add_logger(@host_mod.const_get(:LOGGER))
      end
      if (log_file = getopt(:log_file, @host_mod)) then
	output = File::open(log_file, 'a')
	output.sync = true
	@server.add_logger(Logger::new(output, getopt(:logging_level, @host_mod)))
      end
      if (! getopt(:quiet, @host_mod) && ! STDOUT.closed?) then
	STDOUT.sync = true
	@server.add_logger(PreservedLogger::new(STDOUT, getopt(:display_level, @host_mod)))
      end

      for action, filter in @access_control_list
	@server.add_access_control(action, true, &filter)
      end

      @serv_mod = Module::new
      copy_const(@serv_mod, @host_mod)
      @serv_mod.const_set(:SERVER, @server)
      eval_conf('__SERVER__', @serv_mod, @serv_conf)

      [ [ @server, :timeout=, :timeout ],
	[ @server, :keep_alive=, :keep_alive ],
	[ @server, :connections=, :connections ],
	[ @server, :buffer_size=, :write_buffer ],
	[ @server.request_logger, :request_line_logging_priority=, :request_line_logging_priority ],
	[ @server.request_logger, :header_logging_priority=, :request_header_logging_priority ],
	[ @server.request_logger, :content_logging_priority=, :request_content_logging_priority ],
	[ @server.request_logger, :trailer_logging_priority=, :request_trailer_logging_priority ],
	[ @server.response_logger, :status_line_logging_priority=, :status_line_logging_priority ],
	[ @server.response_logger, :header_logging_priority=, :response_header_logging_priority ],
	[ @server.response_logger, :content_logging_priority=, :response_content_logging_priority ],
	[ @server.response_logger, :trailer_logging_priority=, :response_trailer_logging_priority ]
      ].each do |receiver, opt_set, option|
	if (getopt(option)) then
	  receiver.send(opt_set, getopt(option))
	end
      end

      for dst_path, src_path in @alias_list
	@server.alias(dst_path, src_path)
      end

      for doc_maker in @document_list
	doc_maker.call(@server)
      end

      self
    end
    private :init

    def cool_down
      @cool_down_mod = Module::new
      copy_const(@cool_down_mod, @serv_mod)
      @serv_mod = nil
      eval_conf('__COOL_DOWN__', @cool_down_mod, @cool_down_conf)

      self
    end
    private :cool_down

    def final
      @final_mod = Module::new
      copy_const(@final_mod, @cool_down_mod)
      @cool_down_mod = nil
      eval_conf('__FINAL__', @final_mod, @final_conf)

      self
    end
    private :final

    def run
      @run_sem.synchronize do
	begin
	  loop do
	    begin
	      init
	      unless (@prefork) then
		@prefork = true
		getopt(:prefork_processes).times do
		  fork or break
		end
	      end
	      begin
		@run_sem.post
		@server.accept
	      ensure
		@run_sem.wait
	      end
	    ensure
	      cool_down
	    end
	  end
	ensure
	  final
	end
      end

      self
    end

    def close
      @run_sem.synchronize do
	if (@server) then
	  @server.close
	  return self
	else
	  return nil
	end
      end
    end
  end
end

if ($0 == __FILE__) then
  require 'getoptlong'

  path_sep = File::PATH_SEPARATOR
  path_sep_pattern = Regexp::compile(Regexp::quote(path_sep), false, 'n')

  option = WWWsrv::Accessor::wrap(Hash::new)
  option.daemon = false
  option.umask = nil
  option.version = false
  option.help = false

  logging_priority = {
    'internal' => WWWsrv::Logger::INTERNAL,
    'debug'    => WWWsrv::Logger::DEBUG,
    'system'   => WWWsrv::Logger::SYSTEM,
    'message'  => WWWsrv::Logger::MESSAGE,
    'warning'  => WWWsrv::Logger::WARNING,
    'alert'    => WWWsrv::Logger::ALERT,
    'error'    => WWWsrv::Logger::ERROR,
    # for backward compatibility
    'raw'      => WWWsrv::Logger::INTERNAL,
    'discreet' => WWWsrv::Logger::SYSTEM
  }
  logging_priority_description = '[internal, debug, system, message, warning, alert, error]'

  parse_logging_priority = proc{ |priority|
    if (logging_priority.include? priority) then
      logging_priority[priority]
    else
      raise "unknown logging priority: #{priority}"
    end
  }

  parse_optional_logging_priority = proc{ |priority|
    if (priority && ! priority.empty?) then
      parse_logging_priority.call(priority)
    else
      WWWsrv::Logger::MESSAGE
    end
  }

  add_access_control = proc{ |daemon, action, attr, src|
    (reverse = src =~ /^!/n) and src.gsub!(/^!/n, '')
    pattern = Regexp::compile(src, false, 'n')
    unless (reverse) then
      daemon.add_access_control action, proc{ |request|
	request.client.send(attr) =~ pattern
      }
    else
      daemon.add_access_control action, proc{ |request|
	! (request.client.send(attr) =~ pattern)
      }
    end
  }

  split_page_path = proc{ |path|
    path = path.sub(%r![^/]*$!n, '')
    page = $&
    (path == '/') or path.sub!(%r!/$!n, '')
    [ path, page ]
  }

  opt_action = []

  opt_action.push [
    [ '-m', '--hostname' ],
    proc{ |daemon, opt_arg|
      daemon.hostname = opt_arg
    },
    GetoptLong::REQUIRED_ARGUMENT,
    '[HOSTNAME]'
  ]

  opt_action.push [
    [ '-n', '--port' ],
    proc{ |daemon, opt_arg|
      daemon.port = opt_arg.to_i
    },
    GetoptLong::REQUIRED_ARGUMENT,
    '[port]'
  ]

  opt_action.push [
    [ '-t', '--strict' ],
    proc{ |daemon, opt_arg|
      daemon.strict = true
    },
    GetoptLong::NO_ARGUMENT,
  ]

  opt_action.push [
    [ '--not-strict' ],
    proc{ |daemon, opt_arg|
      daemon.strict = false
    },
    GetoptLong::NO_ARGUMENT
  ]

  opt_action.push [
    [ '-l', '--log-file'],
    proc{ |daemon, opt_arg|
      daemon.log_file = File::expand_path(opt_arg)
    },
    GetoptLong::REQUIRED_ARGUMENT,
    '[path]'
  ]

  opt_action.push [
    [ '--no-log-file'],
    proc{ |daemon, opt_arg|
      daemon.log_file = nil
    },
    GetoptLong::NO_ARGUMENT
  ]

  opt_action.push [
    [ '-v', '--logging-level' ],
    proc{ |daemon, opt_arg|
      daemon.logging_level = parse_logging_priority.call(opt_arg)
    },
    GetoptLong::REQUIRED_ARGUMENT,
    logging_priority_description
  ]

  opt_action.push [
    [ '-p', '--display-level' ],
    proc{ |daemon, opt_arg|
      daemon.display_level = parse_logging_priority.call(opt_arg)
    },
    GetoptLong::REQUIRED_ARGUMENT,
    logging_priority_description
  ]

  opt_action.push [
    [ '--request-line-logging' ],
    proc{ |daemon, opt_arg|
      daemon.request_line_logging_priority = parse_optional_logging_priority.call(opt_arg)
    },
    GetoptLong::OPTIONAL_ARGUMENT,
    logging_priority_description
  ]

  opt_action.push [
    [ '--request-header-logging' ],
    proc{ |daemon, opt_arg|
      daemon.request_header_logging_priority = parse_optional_logging_priority.call(opt_arg)
    },
    GetoptLong::OPTIONAL_ARGUMENT,
    logging_priority_description
  ]

  opt_action.push [
    [ '--request-content-logging' ],
    proc{ |daemon, opt_arg|
      daemon.request_content_logging_priority = parse_optional_logging_priority.call(opt_arg)
    },
    GetoptLong::OPTIONAL_ARGUMENT,
    logging_priority_description
  ]

  opt_action.push [
    [ '--request-trailer-logging' ],
    proc{ |daemon, opt_arg|
      daemon.request_trailer_logging_priority = parse_optional_logging_priority.call(opt_arg)
    },
    GetoptLong::OPTIONAL_ARGUMENT,
    logging_priority_description
  ]

  opt_action.push [
    [ '--status-line-logging' ],
    proc{ |daemon, opt_arg|
      daemon.status_line_logging_priority = parse_optional_logging_priority.call(opt_arg)
    },
    GetoptLong::OPTIONAL_ARGUMENT,
    logging_priority_description
  ]

  opt_action.push [
    [ '--response-header-logging' ],
    proc{ |daemon, opt_arg|
      daemon.response_header_logging_priority = parse_optional_logging_priority.call(opt_arg)
    },
    GetoptLong::OPTIONAL_ARGUMENT,
    logging_priority_description
  ]

  opt_action.push [
    [ '--response-content-logging' ],
    proc{ |daemon, opt_arg|
      daemon.response_content_logging_priority = parse_optional_logging_priority.call(opt_arg)
    },
    GetoptLong::OPTIONAL_ARGUMENT,
    logging_priority_description
  ]

  opt_action.push [
    [ '--response-trailer-logging' ],
    proc{ |daemon, opt_arg|
      daemon.response_trailer_logging_priority = parse_optional_logging_priority.call(opt_arg)
    },
    GetoptLong::OPTIONAL_ARGUMENT,
    logging_priority_description
  ]

  opt_action.push [
    [ '-e', '--effective-user-id' ],
    proc{ |daemon, opt_arg|
      if (opt_arg =~ /^\d+$/n) then
	daemon.effective_user_id = opt_arg.to_i
      else
	daemon.effective_user_id = opt_arg
      end
    },
    GetoptLong::REQUIRED_ARGUMENT,
    '[euid]'
  ]

  opt_action.push [
    [ '-g', '--effective-group-id' ],
    proc{ |daemon, opt_arg|
      if (opt_arg =~ /^\d+$/n) then
	daemon.effective_group_id = opt_arg.to_i
      else
	daemon.effective_group_id = opt_arg
      end
    },
    GetoptLong::REQUIRED_ARGUMENT,
    '[egid]'
  ]

  opt_action.push [
    [ '--timeout' ],
    proc{ |daemon, opt_arg|
      daemon.timeout = opt_arg.to_i
    },
    GetoptLong::REQUIRED_ARGUMENT,
    '[timeout]'
  ]

  opt_action.push [
    [ '--keep-alive' ],
    proc{ |daemon, opt_arg|
      daemon.keep_alive = opt_arg.to_i
    },
    GetoptLong::REQUIRED_ARGUMENT,
    '[timeout]'
  ]

  opt_action.push [
    [ '-q', '--quiet' ],
    proc{ |daemon, opt_arg|
      daemon.quiet = true
    },
    GetoptLong::NO_ARGUMENT
  ]

  opt_action.push [
    [ '--not-quiet' ],
    proc{ |daemon, opt_arg|
      daemon.quiet = false
    },
    GetoptLong::NO_ARGUMENT
  ]

  opt_action.push [
    [ '-d', '--debug' ],
    proc{ |daemon, opt_arg|
      daemon.debug = true
    },
    GetoptLong::NO_ARGUMENT
  ]

  opt_action.push [
    [ '--no-debug' ],
    proc{ |daemon, opt_arg|
      daemon.debug = false
    },
    GetoptLong::NO_ARGUMENT
  ]

  opt_action.push [
    ['-f', '--config'  ],
    proc{ |daemon, opt_arg|
      daemon.config_path = File::expand_path(opt_arg)
    },
    GetoptLong::REQUIRED_ARGUMENT,
    '[path]'
  ]

  opt_action.push [
    [ '-b', '--base-dir' ],
    proc{ |daemon, opt_arg|
      daemon.base_dir = opt_arg
    },
    GetoptLong::REQUIRED_ARGUMENT,
    '[directory]'
  ]

  opt_action.push [
    [ '-s', '--daemon' ],
    proc{ |daemon, opt_arg|
      option.daemon = true
    },
    GetoptLong::NO_ARGUMENT
  ]

  opt_action.push [
    [ '--not-daemon'  ],
    proc{ |daemon, opt_arg|
      option.daemon = false
    },
    GetoptLong::NO_ARGUMENT
  ]

  opt_action.push [
    [ '-u', '--umask' ],
    proc{ |daemon, opt_arg|
      if (opt_arg =~ /^[0-9]+$/n) then
	option.umask = opt_arg.oct
      else
	raise "invalid umask format: #{opt_arg}"
      end
    },
    GetoptLong::REQUIRED_ARGUMENT,
    '[umask]'
  ]

  opt_action.push [
    [ '--connections' ],
    proc{ |daemon, opt_arg|
      daemon.connections = opt_arg.to_i
    },
    GetoptLong::REQUIRED_ARGUMENT,
    '[max connections]'
  ]

  opt_action.push [
    [ '--write-buffer' ],
    proc{ |daemon, opt_arg|
      daemon.write_buffer = opt_arg.to_i
    },
    GetoptLong::REQUIRED_ARGUMENT,
    '[buffer size]',
    :hidden
  ]

  opt_action.push [
    [ '--prefork-processes' ],
    proc{ |daemon, opt_arg|
      daemon.prefork_processes = opt_arg.to_i
    },
    GetoptLong::REQUIRED_ARGUMENT,
    '[processes]',
    :hidden
  ]

  opt_action.push [
    [ '--actl-allow-name' ],
    proc{ |daemon, opt_arg|
      add_access_control.call(daemon, :allow, :name, opt_arg)
    },
    GetoptLong::REQUIRED_ARGUMENT,
    '(!)[regexp]'
  ]

  opt_action.push [
    [ '--actl-deny-name' ],
    proc{ |daemon, opt_arg|
      add_access_control.call(daemon, :deny, :name, opt_arg)
    },
    GetoptLong::REQUIRED_ARGUMENT,
    '(!)[regexp]'
  ]

  opt_action.push [
    [ '--actl-allow-address' ],
    proc{ |daemon, opt_arg|
      add_access_control.call(daemon, :allow, :address, opt_arg)
    },
    GetoptLong::REQUIRED_ARGUMENT,
    '(!)[regexp]'
  ]

  opt_action.push [
    [ '--actl-deny-address' ],
    proc{ |daemon, opt_arg|
      add_access_control.call(daemon, :deny, :address, opt_arg)
    },
    GetoptLong::REQUIRED_ARGUMENT,
    '(!)[regexp]'
  ]

  opt_action.push [
    [ '-a', '--alias' ],
    proc{ |daemon, opt_arg|
      dst_path, src_path = opt_arg.split(path_sep_pattern, 2)
      if ((! dst_path || dst_path.empty?) || (! src_path || src_path.empty?)) then
	raise "invalid alias format: #{opt_arg}"
      end
      daemon.add_alias(dst_path, src_path)
    },
    GetoptLong::REQUIRED_ARGUMENT,
    "[dst. path]#{path_sep}[src. path]"
  ]

  opt_action.push [
    [ '--mount-echo' ],
    proc{ |daemon, opt_arg|
      require 'wwwsrv/echo'
      mount_path = opt_arg
      daemon.add_document{ |server|
	server.mount(WWWsrv::Echo::new, mount_path)
      }
    },
    GetoptLong::REQUIRED_ARGUMENT,
    '[HTTP path]'
  ]

  opt_action.push [
    [ '--mount-local' ],
    proc{ |daemon, opt_arg|
      require 'wwwsrv/local'
      mount_path, local_path, local_regexp = opt_arg.split(path_sep_pattern, 3)
      (mount_path && local_path) or
	raise "invalid local mount format: #{opt_arg}"
      local_path = File::expand_path(local_path)
      if (local_regexp && ! local_regexp.empty?) then
	local_regexp = Regex::compile(local_regexp, false, 'n')
      else
	local_regexp = nil
      end
      if (File::directory? local_path) then
	daemon.add_document{ |server|
	  server.mount(WWWsrv::LocalFileDocument::new(local_path), mount_path, local_regexp)
	}
      elsif (File::file? local_path) then
	local_dir, local_name = File::split(local_path)
	daemon.add_document{ |server|
	  server.mount(WWWsrv::LocalFileDocument::new(local_dir),
		       mount_path, local_name)
	}
      else
	raise "not found a local file or directory: #{local_path}"
      end
    },
    GetoptLong::REQUIRED_ARGUMENT,
    "[HTTP path]#{path_sep}[local path](#{path_sep}[regexp])"
  ]

  opt_action.push [
    [ '--mount-cgi' ],
    proc{ |daemon, opt_arg|
      require 'wwwsrv/cgi'
      mount_path, cgi_path, cgi_regexp, cgi_exec = opt_arg.split(path_sep_pattern, 4)
      (mount_path && cgi_path) or
	raise "invalid cgi mount format: #{opt_arg}"
      cgi_path = File::expand_path(cgi_path)
      if (cgi_regexp && ! cgi_regexp.empty?) then
	cgi_regexp = Regexp::compile(cgi_regexp, false, 'n')
      else
	cgi_regexp = nil
      end
      if (cgi_exec && cgi_exec.empty?) then
	cgi_exec = nil
      end
      if (File::directory? cgi_path) then
	daemon.add_document{ |server|
	  server.mount(WWWsrv::CGIdocument::new(cgi_path, false, cgi_exec),
		       mount_path, cgi_regexp)
	}
      elsif (File::file? cgi_path) then
	cgi_dir, cgi_name = File::split(cgi_path)
	daemon.add_document{ |server|
	  server.mount(WWWsrv::CGIdocument::new(cgi_dir, false, cgi_exec),
		       mount_path, cgi_name)
	}
      else
	raise "not found a CGI file or directory: #{cgi_path}"
      end
    },
    GetoptLong::REQUIRED_ARGUMENT,
    "[HTTP path]#{path_sep}[CGI path](#{path_sep}[regexp](#{path_sep}[interpreter]))"
  ]

  opt_action.push [
    [ '--mount-nph-cgi' ],
    proc{ |daemon, opt_arg|
      require 'wwwsrv/cgi'
      mount_path, cgi_path, cgi_regexp, cgi_exec = opt_arg.split(path_sep_pattern, 4)
      (mount_path && cgi_path) or
	raise "invalid cgi mount format: #{opt_arg}"
      cgi_path = File::expand_path(cgi_path)
      if (cgi_regexp && ! cgi_regexp.empty?) then
	cgi_regexp = Regexp::compile(cgi_regexp, false, 'n')
      else
	cgi_regexp = nil
      end
      if (cgi_exec && cgi_exec.empty?) then
	cgi_exec = nil
      end
      if (File::directory? cgi_path) then
	daemon.add_document{ |server|
	  server.mount(WWWsrv::CGIdocument::new(cgi_path, true, cgi_exec),
		       mount_path, cgi_regexp)
	}
      elsif (File::file? cgi_path) then
	cgi_dir, cgi_name = File::split(cgi_path)
	daemon.add_document{ |server|
	  server.mount(WWWsrv::CGIdocument::new(cgi_dir, true, cgi_exec),
		       mount_path, cgi_name)
	}
      else
	raise "not found a CGI file or directory: #{cgi_path}"
      end
    },
    GetoptLong::REQUIRED_ARGUMENT,
    "[HTTP path]#{path_sep}[CGI path](#{path_sep}[regexp](#{path_sep}[interpreter]))"
  ]

  opt_action.push [
    [ '--attach-ssi' ],
    proc{ |daemon, opt_arg|
      require 'wwwsrv/ssi'
      mount_path, ssi_regexp, ssi_tags = opt_arg.split(path_sep_pattern, 3)
      if (ssi_regexp && ! ssi_regexp.empty?) then
	ssi_page = Regexp::compile(ssi_regexp, false, 'n')
      else
	mount_path, ssi_page = split_page_path.call(mount_path)
      end
      if (ssi_tags && ! ssi_tags.empty?) then
	ssi_tags = ssi_tags.split(/,/n)
      else
	ssi_tags = nil
      end
      if (ssi_tags) then
	daemon.add_document{ |server|
	  server.attach(mount_path, ssi_page) { |document|
	    WWWsrv::SSIdocument::new(document, ssi_tags)
	  }
	}
      else
	daemon.add_document{ |server|
	  server.attach(mount_path, ssi_page) { |document|
	    WWWsrv::SSIdocument::new(document)
	  }
	}
      end
    },
    GetoptLong::REQUIRED_ARGUMENT,
    "[HTTP path](#{path_sep}[regexp](#{path_sep}[SSI tags]))"
  ]

  opt_action.push [		# for backward compatibility
    [ '--mount-ssi' ],
    opt_action.last[1],
    opt_action.last[2],
    opt_action.last[3],
    :obsolete
  ]

  opt_action.push [
    [ '-h', '--help' ],
    proc{ |daemon, opt_arg|
      option.help = true
    },
    GetoptLong::NO_ARGUMENT
  ]

  opt_action.push [
    [ '--version' ],
    proc{ |daemon, opt_arg|
      option.version = true
    },
    GetoptLong::NO_ARGUMENT
  ]

  opt_proc_map = {}
  opt_obs_map = {}
  for opt_list, opt_proc, opt_type, opt_arg_desc, category in opt_action
    for opt_name in opt_list
      opt_proc_map[opt_name] = opt_proc
      opt_obs_map[opt_name] = category == :obsolete
    end
  end

  getoptlong = GetoptLong::new
  getoptlong.set_options *opt_action.map{ |opt_list, opt_proc, opt_type, opt_arg_desc, category|
    opt_list + [ opt_type ]
  }

  daemon = WWWsrv::Daemon::new
  for opt_name, opt_arg in getoptlong
    if (opt_proc = opt_proc_map[opt_name]) then
      if (opt_obs_map[opt_name]) then
	STDERR.print "warning: option `#{opt_name}' is obsolete!\n"
      end
      opt_proc.call(daemon, opt_arg)
    end
  end
  daemon.base_dir = Dir::getwd unless daemon.base_dir

  if (! ARGV.empty? || option.version || option.help) then
    if (! ARGV.empty?) then
      print 'error: invalid arguments: ', ARGV.map{ |arg|
	arg.inspect
      }.join(', '), "\n"
    end
    if (option.version) then
      print WWWsrv::SRVNAME, ' version ', WWWsrv::VERSION
      print ' (', WWWsrv::CODENAME, '; ', WWWsrv::RCS_DATE, ")\n"
    end
    if (option.help) then
      print "Usage: #{File::basename($0)} [options]\n"
      print "Options:\n"
      for opt_list, opt_proc, opt_type, opt_arg_desc, category in opt_action
	(category == :obsolete) and next
	(category == :hidden) and next
	print '  '
	print '    ' if (opt_list.first !~ /^-[^-]$/n)
	print opt_list.join(', ')
	case (opt_type)
	when GetoptLong::REQUIRED_ARGUMENT
	  print '=', opt_arg_desc
	when GetoptLong::OPTIONAL_ARGUMENT
	  print '(=', opt_arg_desc, ')'
	end
	print "\n"
      end
    end
    exit
  end

  if (option.daemon) then
    daemon.quiet = true
    WWWsrv::Daemon::take_off

    reload = Queue::new
    trap('HUP') {
      reload.push(true)
    }
    Thread::new{
      loop do
	reload.pop
	daemon.close
      end
    }.priority = WWWsrv::SYSTEM_PRIORITY
  end

  # for handling final hooks of mounted documents
  trap('TERM') {
    raise SignalException, 'SIGTERM'
  }

  File::umask(option.umask) if option.umask

  if (daemon.debug) then
    Thread.abort_on_exception = true
  end

  daemon.run
end
