# $Id: content.rb,v 1.8.2.2 2002/02/27 14:16:56 toki Exp $
# -
# HTTP content reader

=begin
Copyright (c) 2001-2002
	TOKI Yoshinori <toki@freedom.ne.jp>. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1.  Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

  2.  Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

  3.  The name of the author may not be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.
=end

require 'singleton'
require 'wwwsrv/const'
require 'wwwsrv/thread'
require 'wwwsrv/ioutil'
require 'wwwsrv/chunk'
require 'wwwsrv/header'

module WWWsrv
  class ContentReader
    include Timeout
    include ConnectionProtect

    def initialize(sid, input, watcher, timeout)
      @sid = sid
      @input = input
      @watcher = watcher
      @timeout = timeout
      @trailer = Header::new
    end

    attr_reader :trailer

    def sysread(length)
      fragment = wait(@timeout) {
	protect_connection{ @input.read(length) }
      }
      @watcher.watch_request_content(@sid, fragment)
      fragment
    end

    def read(length=nil)
      sysread(length)
    end

    def to_io
      @input && @input.to_io
    end
  end

  class LengthLimitedContentReader < ContentReader
    def initialize(sid, input, watcher, timeout, content_length=nil)
      super(sid, input, watcher, timeout)
      @read_length = content_length
    end

    def read(length=nil)
      (length && length < 0) and
	raise ArgumentError, "negative length #{length} given"

      if (! @read_length) then
	return sysread(length)
      elsif (length && length <= @read_length) then
	fragment = sysread(length)
	if (fragment) then
	  @read_length -= fragment.length
	else
	  @read_length = 0
	end
	return fragment
      elsif (@read_length > 0) then
	fragment = sysread(@read_length)
	@read_length = 0
	return fragment
      else
	return nil
      end
    end
  end

  class ChunkedContentReader < ContentReader
    include ChunkedTransferCoding

    def initialize(sid, input, watcher, timeout)
      super(sid, input, watcher, timeout)
      @state = :chunk
      @chunk_spool = ''
    end

    def read_chunk
      case (@state)
      when :chunk
	chunk_data, chunk_extension = wait(@timeout) {
	  protect_connection{ parse_chunk(@input) }
	}
	@watcher.watch_request_content(@sid, chunk_data, *chunk_extension)
	unless (chunk_data) then
	  wait(@timeout) do
	    protect_connection{ Header::parse(@input, self.trailer) }
	  end
	  @watcher.watch_request_trailer(@sid, @trailer)
	  @state = :final
	end
	return chunk_data, chunk_extension
      when :final
	return nil
      else
	raise "invalid internal state: #{@state}"
      end
    end

    def each_chunk
      while (chunk = read_chunk)
	yield(chunk)
      end

      self
    end

    def read(length=nil)
      (length && length < 0) and
	raise ArgumentError, "negative length #{length} given"

      case (@state)
      when :chunk
	while (! length || @chunk_spool.length < length)
	  chunk_data, chunk_extension = read_chunk
	  chunk_data or break
	  @chunk_spool.concat(chunk_data)
	end

	if (length) then
	  fragment = @chunk_spool[0, length]
	  @chunk_spool[0, length] = ''
	else
	  fragment = @chunk_spool
	  @chunk_spool = nil
	end

	return fragment
      when :final
	return nil
      end
    end
  end

  class EmptyContentReader < ContentReader
    include Singleton

    def initialize
      super(nil, nil, nil, nil)
    end

    def sysread(length)
      nil
    end

    def each_fragment
      self
    end
  end

  class LineReader
    include Enumerable

    def initialize(reader)
      @reader = reader
      @fragment = nil
    end

    def sysread(length)
      @reader.sysread(length)
    end

    def read(length=nil)
      if (@fragment) then
	if (length) then
	  if (length < @fragment.length) then
	    read_fragment = @fragment[0, length]
	    @fragment[0, length] = ''
	    return read_fragment
	  else
	    if (@fragment.length > length) then
	      next_fragment = @reader.read(length - @fragment.length)
	      @fragment.concat(next_fragment) if next_fragment
	    end
	    read_fragment, @fragment = @fragment, nil
	    return read_fragment
	  end
	else
	  @fragment.concat(@reader.read)
	  read_fragment, @fragment = @fragment, nil
	  return read_fragment
	end	  
      else
	return @reader.read(length)
      end
    end

    def gets(eol=$/)
      unless (@fragment) then
	unless (@fragment = @reader.read(IO_UNIT_SIZE)) then
	  return nil
	end
      end

      if (eol == LF) then
	line_scan = /^.*?\n/n
      else
	line_scan = Regexp::compile('^.*?' + Regexp::quote(eol, 'n'), false, 'n')
      end

      until (@fragment.sub!(line_scan, ''))
	unless (next_fragment = @reader.read(IO_UNIT_SIZE)) then
	  read_fragment, @fragment = @fragment, nil
	  return read_fragment
	end
	@fragment.concat(next_fragment)
      end
      @fragment = nil if @fragment.empty?

      $&
    end

    def each_line(eol=$/)
      while (line = gets(eol))
	yield(line)
      end

      self
    end

    alias each each_line

    # for chunked transfer encoding

    def trailer
      @reader.trailer
    end

    def read_chunk
      @reader.read_chunk
    end

    def each_chunk
      @reader.each_chunk do |*args|
	yield(*args)
      end

      self
    end

    def to_io
      @reader && @reader.to_io
    end
  end
end
