# $Id: document.rb,v 1.10.2.2 2002/02/27 14:16:59 toki Exp $
# -
# server document

=begin
Copyright (c) 2001-2002
	TOKI Yoshinori <toki@freedom.ne.jp>. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1.  Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

  2.  Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

  3.  The name of the author may not be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.
=end

require 'md5'
require 'thread'
require 'parsedate'
require 'wwwsrv/const'
require 'wwwsrv/path'
require 'wwwsrv/htmlconv'
require 'wwwsrv/header'
require 'wwwsrv/document'

module WWWsrv
  module HTTPerror
    def http_error(status, message=nil, header={}, version=HTTP_1_1)
      throw :WWWSRV_HTTP_ERROR, [ :error, [ status, message, header, version ] ]
    end
    module_function :http_error

    # for backward compatibility
    def make_error_response(response, status)
      error_response = Response::new(response.version, status)
      error_response.header.server = response.header.server
      error_response
    end
    module_function :make_error_response
  end

  class Document
    include HTMLconv
    include HTTPerror

    class << self
      # for backward compatibility
      def http_date(time)
	HTMLconv::http_date(time)
      end

      # for backward compatibility
      alias format_date http_date

      # for backward compatibility
      def quoteHTML(data)
	HTMLconv::quoteHTML(data)
      end

      # for backward compatibility
      def unquoteHTML(data)
	HTMLconv::unquoteHTML(data)
      end

      # for backward compatibility
      def quoteSHELL(data)
	HTMLconv::quoteSHELL(data)
      end

      # for backward compatibility
      def encodeURL(data)
	HTMLconv::encodeURL(data)
      end

      # for backward compatibility
      def decodeURL(data)
	HTMLconv::decodeURL(data)
      end

      def catch_state
	state, err_args = catch(:WWWSRV_HTTP_ERROR) {
	  yield
	  [ :success, [] ]
	}

	return state, err_args
      end
    end

    INIT_LOCK = Mutex::new

    def setup(log, mark={})
      unless (mark.include? self.id) then
	mark[self.id] = true

	count = INIT_LOCK.synchronize{
	  @init_count = 0 unless @init_count
	  @init_count += 1
	  @init_count
	}

	if (count == 1) then
	  warm_up(log)
	end
      end

      self
    end

    def cleanup(log, mark={})
      unless (mark.include? self.id) then
	mark[self.id] = true

	count = INIT_LOCK.synchronize{
	  @init_count = 0 unless @init_count
	  @init_count -= 1
	  @init_count
	}

	if (count == 0) then
	  cool_down(log)
	end
      end

      self
    end

    def warm_up(log)
      if (respond_to? :init) then
	log.warning("warning: #{type}\##init method is obsolete!")
	init(log)
      end

      self
    end

    def cool_down(log)
      if (respond_to? :final) then
	log.warning("warning: #{tpe}\##final method is obsolete!")
	final(log)
      end

      self
    end

    def subdocument(path)
      DocumentPart::new(self, path)
    end

    def content(sid, srv_opt, prefix, request, response)
      srv_opt.log.warning("sid #{sid}: warning: #{type}\#session method is obsolete!")
      session(sid, srv_opt, prefix, request, response) # for backward compatibility
    end

    def session(sid, srv_opt, prefix, request, response)

      #
      # WWWsrv::Document#session method is obsolete.
      #
      # Usually, the word of `session' means a series of related pairs
      # of HTTP request and response. But obsolete session method
      # works on only one pair of HTTP response and request. Therefore
      # `session' is not suitable for a name of this method and I
      # renamed.
      #
      # Sorry.
      #

      srv_opt.log.debug("sid #{sid}: debug #{type}")
      http_error(404, "Empty document: #{request.normalized_path}.")
      self
    end
  end

  class DocumentPart < Document
    include Path

    def initialize(document, path)
      super()
      @document = document
      @path = path.sub(%r!/+$!n, '')
      @path_pattern = Regexp::compile(Regexp::quote(@path) + '$', false, 'n')
    end

    attr_reader :document

    def content(sid, srv_opt, prefix, request, response)
      srv_opt.log.debug("sid #{sid}: debug #{type}")
      if (prefix =~ @path_pattern) then
	modified_prefix = prefix.sub(@path_pattern, '')
	srv_opt.log.debug{
	  "sid #{sid}: debug #{type}: modify request path: " +
	  request.subpath(prefix) + ' -> ' + request.subpath(modified_prefix)
	}
	prefix = modified_prefix
      else
	unless (subpath = request.subpath(prefix)) then
	  http_error(404, "Invalid path: #{request.normalized_path}.")
	end
	srv_opt.log.debug("sid #{sid}: debug #{type}: ",
			  "modify request path: ",
			  subpath, '  -> ', @path, subpath)
	modified_path = prefix + @path + subpath
	modified_path += '?' + request.query if request.query
	request.path = modified_path
      end

      @document.content(sid, srv_opt, prefix, request, response) do |arg, *opts|
	yield(arg, *opts)
      end

      self
    end

    def setup(log, mark={})
      unless (mark.include? self.id) then
	super
	@document.setup(log, mark)
      end

      self
    end

    def cleanup(log, mark={})
      unless (mark.include? self.id) then
	super
	@document.cleanup(log, mark)
      end

      self
    end
  end

  PartialDocument = DocumentPart # for backward compatibility

  class DocumentNode
    def initialize
      @page = {}
      @filter = []
      @document = nil
      @subnode = {}
    end

    def mount(document, path_fragments, page)
      if (path_fragments.empty?) then
	if (page) then
	  if (page.kind_of? Regexp) then
	    regexp_pattern = page
	    if (filter_pair = @filter.assoc(regexp_pattern)) then
	      filter_pair[1] = document
	    else
	      @filter.push([ regexp_pattern, document ])
	    end
	  else
	    (@subnode.include? page) and
	      raise "already mounted as directory: #{page}"
	    @page[page] = document
	  end
	else
	  @document = document
	end
      else
	path_segment = path_fragments.shift
	unless (child_node = @subnode[path_segment]) then
	  (@page.include? path_segment) and
	    raise "already mounted as page: #{path_segment}"
	  @subnode[path_segment] = child_node = DocumentNode::new
	end
	child_node.mount(document, path_fragments, page)
      end

      document
    end

    def umount(path_fragments, page)
      if (path_fragments.empty?) then
	if (page) then
	  if (page.kind_of? Regexp) then
	    document = nil
	    regexp_pattern = page
	    @filter.delete_if{ |regexp_pattern2, document2|
	      if (regexp_pattern2 == regexp_pattern) then
		document = document2
		true
	      end
	    }
	    return document
	  else
	    return @page.delete(page)
	  end
	else
	  document = @document
	  @document = nil
	  return document
	end
      else
	path_segment = path_fragments.shift
	if (child_node = @subnode[path_segment]) then
	  return child_node.umount(path_fragments, page)
	end
      end

      nil
    end

    def find(path_fragments, mount_path)
      if (path_fragments.empty?) then
	if (document = @page['']) then
	  # page of directory
	  return document
	end
      else
	path_segment = path_fragments.shift
	if (child_node = @subnode[path_segment]) then
	  if (document = child_node.find(path_fragments, mount_path)) then
	    mount_path.unshift(path_segment)
	    return document
	  end
	end

	if (path_fragments.empty?) then
	  if (document = @page[path_segment]) then
	    return document
	  end
	end

	path_fragments.unshift(path_segment)
	page_path = path_fragments.join('/')
	for regexp_pattern, document in @filter
	  if (page_path =~ regexp_pattern) then
	    return document
	  end
	end
      end

      @document
    end

    def each(mount_path)
      yield(@document) if @document

      for regexp_pattern, document in @filter
	yield(document, regexp_pattern)
      end

      for page, document in @page
	yield(document, page)
      end

      for path_segment, child_node in @subnode
	mount_path.push(path_segment)
	child_node.each(mount_path) do |document|
	  yield(document)
	end
	mount_path.pop
      end

      self
    end
  end

  class DocumentTree
    include Path
    include Enumerable

    def initialize
      @root = DocumentNode::new
    end

    def mount(document, path, page=nil)
      path_fragments = split_path(path)
      @root.mount(document, path_fragments, page)
    end

    def umount(path, page=nil)
      path_fragments = split_path(path)
      @root.umount(path_fragments, page)
    end

    def find(path)
      mount_path = []
      path_fragments = split_path(path)
      if (document = @root.find(path_fragments, mount_path)) then
	return document, join_path(mount_path)
      end

      nil
    end

    def each
      mount_path = []
      @root.each(mount_path) do |document, page|
	yield(document, join_path(mount_path), page)
      end

      self
    end
  end

  class DocumentFolder < Document
    include Path

    def initialize
      super()
      @folder = {}
    end

    def mount(document, path, page=nil, host=:default)
      @folder[host] = DocumentTree::new unless (@folder.include? host)
      @folder[host].mount(document, path, page)
      document
    end

    def umount(path, page=nil, host=:default)
      if (@folder.include? host) then
	return @folder[host].umount(path, page)
      end

      nil
    end

    def attach(path, page=nil, target=nil, host=:default)
      target_path = path
      target_path += '/' + target if target
      mount_pair = find(target_path, host) or
	raise "not found a mounted document: #{target_path}"

      document, mount_path = mount_pair
      if (path == mount_path) then
	modified_document = yield(document)
      else
	subpath = WWWsrv::Path::subpath(path, mount_path)
	subdocument = document.subdocument(subpath)
	modified_document = yield(subdocument)
      end

      mount(modified_document, path, page, host)
    end

    def find(path, host=:default)
      if (@folder.include? host) then
	return @folder[host].find(path)
      end

      nil
    end

    def each(host=:default)
      if (@folder.include? host) then
	for document, path, page in @folder[host]
	  case (page)
	  when String
	    yield(document, path + '/' + page)
	  when Regexp
	    yield(document, path + '/' + page.source)
	  when NilClass
	    yield(document, path)
	  else
	    raise "unknown page type: #{page.type}"
	  end
	end
      end

      self
    end

    def virtual_host
      @folder.keys.reject{ |hostname| hostname != :default }
    end

    def virtual_host_defined?(hostname)
      @folder.include? hostname
    end

    def setup(log, mark={})
      unless (mark.include? self.id) then
	super
	for hostname, doc_tree in @folder
	  for document, mount_path in doc_tree
	    document.setup(log, mark)
	  end
	end
      end

      self
    end

    def cleanup(log, mark={})
      unless (mark.include? self.id) then
	super
	for hostname, doc_tree in @folder
	  for document, mount_path in doc_tree
	    document.cleanup(log, mark)
	  end
	end
      end

      self
    end

    def warm_up(log)
      log.debug("debug #{type}: object ID #{format('0x%x', self.id)}")

      if (doc_tree = @folder[:default]) then
	for document, path, page in doc_tree
	  if (page) then
	    log.debug("debug #{type}: " +
		      "mount #{document.type}(#{format('0x%x', document.id)}) => " +
		      "#{path.inspect} + #{page.inspect}")
	  else
	    log.debug("debug #{type}: " +
		      "mount #{document.type}(#{format('0x%x', document.id)}) => " +
		      "#{path.inspect}")
	  end
	end
      end

      hosts = @folder.keys - [ :default ]
      for host in hosts
	if (doc_tree = @folder[:default]) then
	  for document, path, page in doc_tree
	    if (page) then
	      log.debug("debug #{type}: host #{host}: " +
			"mount #{document.type}(#{format('0x%x', document.id)}) => " +
			"#{path} + #{page.inspect}")
	    else
	      log.debug("debug #{type}: host #{host}: " +
			"mount #{document.type}(#{format('0x%x', document.id)}) => " +
			"#{path}")
	    end
	  end
	end
      end

      self
    end

    def content(sid, srv_opt, prefix, request, response)
      srv_opt.log.debug("sid #{sid}: debug #{type}")
      if (request) then
	path = request.subpath(prefix)
	if (virtual_host_defined? request.host.to_s) then
	  mount_pair = find(path, request.host.to_s)
	elsif (virtual_host_defined? request.host.name) then
	  mount_pair = find(path, request.host.name)
	else
	  mount_pair = find(path)
	end
      else
	# for fatal alarm document
	path = '/'
	mount_pair = find(path)
      end

      if (mount_pair) then
	document, mount_path = mount_pair
	if (request.normalized_path !~ %r!/$!n && prefix + mount_path == request.normalized_path) then
	  srv_opt.log.debug("sid #{sid}: debug #{type}: ",
			    "redirect from #{request.normalized_path} ",
			    "to #{request.normalized_path}/")
	  response.status = 301
	  response.header.location =
	    srv_opt.scheme + '://' + request.host.name + ':' + request.host.port.to_s +
	    request.normalized_path + '/'
	  response.header.content_type = 'text/plain'
	  location = response.header.location
	  yield(response)
	  yield("Moved to #{location}\n")
	else
	  srv_opt.log.debug("sid #{sid}: debug #{type}: ",
			    "request to subdocument: ",
			    prefix + mount_path)
	  prefix += mount_path if (mount_path != '/')
	  document.content(sid, srv_opt, prefix, request, response) do |arg, *opts|
	    yield(arg, *opts)
	  end
	end
      else
	http_error(404, "Not found a document of #{request.path}.")
      end

      self
    end
  end

  class DocumentDispatcher < Document
    def initialize(default_document=nil)
      @default_document = default_document
      @document_alist = []
      @vary_field_names = {}
    end

    attr_accessor :default_document

    def setup(log, mark={})
      unless (mark.include? log) then
	super
	@default_document.setup(log, mark)
	for cond, document in @document_alist
	  document.setup(log, mark)
	end
      end

      self
    end

    def cleanup(log, mark={})
      unless (mark.include? self.id) then
	super
	@default_document.setup(log, mark)
	for cond, document in @document_alist
	  document.cleanup(log, mark)
	end
      end

      self
    end

    def add_cond(document, *vary_field_names, &cond)
      @document_alist.push([ cond, document ])
      for field_name in vary_field_names
	@vary_field_names[Header::normalize(field_name)] = true
      end

      self
    end

    def cond_call(cond, srv_opt, request)
      if (cond.arity.abs > 1) then
	return cond.call(srv_opt, request)
      else
	return cond.call(request) # for backward compatibility
      end
    end
    private :cond_call

    def content(sid, srv_opt, prefix, request, response)
      srv_opt.log.debug("sid #{sid}: debug #{type}")

      @vary_field_names.each_key do |field_name|
	response.header.push('Vary', field_name)
      end

      for cond, document in @document_alist
	if (cond_call(cond, srv_opt, request)) then
	  srv_opt.log.debug("sid #{sid}: debug #{type}: ",
			    "dispatch document: #{document.type}")
	  document.content(sid, srv_opt, prefix, request, response) do |arg, *opts|
	    yield(arg, *opts)
	  end
	  return self
	end
      end

      if (@default_document) then
	srv_opt.log.debug("sid #{sid}: debug #{type}: ",
			  "default document: #{@default_document.type}")
	@default_document.content(sid, srv_opt, prefix, request, response) do |arg, *opts|
	  yield(arg, *opts)
	end
	return self
      end

      http_error(404, "Not found a document at #{request.normalized_path}.\n")
    end
  end

  class Page < Document
    def initialize(content, type='text/html', mtime=nil)
      super()
      @content = content
      @type = type
      @mtime = mtime
      md5 = MD5::new
      md5.update(@content)
      @content_md5 = [ md5.digest ].pack('m').strip
      md5.update(@type)
      md5.update(@mtime.to_i.to_s) if @mtime
      @entity_tag = '"' + md5.hexdigest + '"'
    end

    def content(sid, srv_opt, prefix, request, response)
      srv_opt.log.debug("sid #{sid}: debug #{type}")
      case (request.method)
      when 'GET', 'HEAD'
	response.status = request.condition(@mtime, @hexdigest)
	response.status = 200 if (response.status == 206)
	response.header.etag = @entity_tag
	response.header.content_md5 = @content_md5
	response.header.last_modified = http_date(@mtime) if @mtime
	if (response.status == 200) then
	  response.header.content_type = @type
	  response.header.content_length = @content.length.to_s
	end
	status = response.status
	yield(response)
	yield(@content.dup) if (request.method == 'GET' && status == 200)
      else
	http_error(405, "Not supported method: #{request.method}.", { 'Allow' => 'GET, HEAD' })
      end
    end
  end

  class Redirect < Document
    def initialize(status=301, &redirect)
      @status = status
      @redirect = redirect
    end

    def content(sid, srv_opt, prefix, request, response)
      srv_opt.log.debug("sid #{sid}: debug #{type}")

      if (@redirect.arity.abs > 1) then
	location = @redirect.call(srv_opt, request)
      else
	location = @redirect.call(request) # for backward compatibility
      end

      if (location !~ %r!^[A-Za-z][A-Za-z0-9.+-]*://!n) then
	if (location !~ %r!^/!n) then
	  location = request.normalized_path.sub(%r![^/]*$!, '') + location
	end
	location = srv_opt.scheme + '://' + request.host.name + ':' + request.host.port.to_s + location
      end

      response.status = @status
      response.header.location = location
      response.header.content_type = 'text/plain'

      srv_opt.log.debug("sid #{sid}: debug #{type}: ",
			"redirect from #{request.normalized_path} ",
			"to #{response.header.location}")

      yield(response)
      yield("Moved to #{location}.\n")

      self
    end
  end

  module DocumentCascade
    def method_missing(id, *args, &block)
      ret_val = document.send(id, *args, &block)
      if (ret_val.equal? document) then
	return self
      else
	return ret_val
      end
    end

    def setup(log, mark={})
      unless (mark.include? self.id) then
	super
	document.setup(log, mark)
      end

      self
    end

    def cleanup(log, mark={})
      unless (mark.include? self.id) then
	super
	document.cleanup(log, mark)
      end

      self
    end
  end

  class DocumentModifier < Document
    include DocumentCascade

    def initialize(document)
      super()
      @document = document
    end

    attr_reader :document

    def terminate
      throw(:wwwsrv_document_modifier_terminal)
    end

    alias prune terminate

    def content(sid, srv_opt, prefix, request, response, &block)
      srv_opt.log.debug("sid #{sid}: debug #{type}")
      context = {}
      catch(:wwwsrv_document_modifier_terminal) do
	modify_request(sid, srv_opt, context, prefix, request, response, &block)
	doc_response = nil
	@document.content(sid, srv_opt, prefix, request, response) do |arg, *opts|
	  unless (doc_response) then
	    doc_response = arg
	    modify_response(sid, srv_opt, context, prefix, request, doc_response, &block)
	  else
	    modify_content(sid, srv_opt, context, arg, *opts, &block)
	  end
	end
      end
      modify_terminal(sid, srv_opt, context, &block)
    end

    def modify_request(sid, srv_opt, context, prefix, request, response)
      self
    end

    def modify_response(sid, srv_opt, context, prefix, request, response)
      yield(response)
      self
    end

    def modify_content(sid, srv_opt, context, content_fragment, *opts)
      yield(content_fragment, *opts)
      self
    end

    def modify_terminal(sid, srv_opt, context)
      self
    end
  end

  class DocumentAlias < DocumentModifier
    include Path

    def initialize(document)
      super(document)
      @map = {}
    end

    def alias(dst_path, src_path)
      @map[dst_path] = src_path
      self
    end

    def modify_request(sid, srv_opt, context, prefix, request, response)
      if (alias_path = @map[request.subpath(prefix)]) then
	count = 1
	expanded_path = alias_path
	while (alias_path = @map[expanded_path])
	  if (count > 10) then
	    http_error(404, "Failed to expand alias path.")
	  end
	  count += 1
	  expanded_path = alias_path
	end

	srv_opt.log.debug("sid #{sid}: debug #{type}: ",
			  "expand alias path: ",
			  request.normalized_path, ' -> ', expanded_path)

	expanded_path += '?' + request.query if request.query
	request.path = expanded_path
      else
	srv_opt.log.debug("sid #{sid}: debug #{type}: ",
			  "not modified path: ",
			  request.normalized_path)
      end

      self
    end
  end

  class AccessControl < DocumentModifier
    ALLOW = :allow
    DENY  = :deny

    def initialize(document)
      super(document)
      @access_list = []
    end

    def add_access_control(action, logging=false, &filter)
      @access_list.push([ action, filter, logging ])
      self
    end

    def filter_call(filter, srv_opt, request)
      if (filter.arity.abs > 1) then
	return filter.call(srv_opt, request)
      else
	return filter.call(request) # for backward compatibility
      end
    end
    private :filter_call

    def modify_request(sid, srv_opt, context, prefix, request, response)
      catch(:ACCESS_ALLOW) do
	allow_filter_used = false
	for action, filter, logging in @access_list
	  case (action)
	  when ALLOW
	    allow_filter_used = true
	    if (filter_call(filter, srv_opt, request)) then
	      if (logging) then
		srv_opt.log.system("sid #{sid}: access allow: #{request.client}")
	      end
	      throw(:ACCESS_ALLOW)
	    end
	  when DENY
	    if (filter_call(filter, srv_opt, request)) then
	      if (logging) then
		srv_opt.log.system("sid #{sid}: access deny: #{request.client}")
	      end
	      http_error(403, 'Access denied.')
	    end
	  else
	    raise "unknown action: #{action}"
	  end
	end

	if (allow_filter_used) then
	  if (logging) then
	    srv_opt.log.system("sid #{sid}: access deny: #{request.client}")
	  end
	  http_error(403, 'Access denied.')
	end
      end

      self
    end
  end

  class PartialContent < DocumentModifier
    def parse_range(field_value)
      case (field_value)
      when /^bytes=(\d+)-(\d+)?$/n
	return [ $1.to_i, $2 && $2.to_i ]
      when /^bytes=-(\d+)$/n
	return [ nil, $1.to_i ]
      else
	# not support multipart/byteranges.
	return nil
      end
    end
    private :parse_range

    def normalize_range(first_pos, last_pos, content_length)
      (content_length == 0) and http_error(416)
      if (first_pos) then
	(first_pos >= content_length) and http_error(416)
	if (! last_pos || last_pos >= content_length) then
	  last_pos = content_length - 1
	end
      else
	if (last_pos < content_length) then
	  first_pos = content_length - last_pos
	  last_pos = content_length - 1
	else
	  first_pos = 0
	  last_pos = content_length - 1
	end
      end
      (first_pos >= last_pos) and http_error(416)

      [ first_pos, last_pos ]
    end
    private :normalize_range

    def cut_range(content_fragment, curr_pos, first_pos, last_pos)
      end_pos = curr_pos + content_fragment.length - 1
      if (first_pos <= curr_pos && curr_pos <= last_pos) then
	if (end_pos <= last_pos) then
	  yield(content_fragment)
	else
	  yield(content_fragment[0, last_pos - curr_pos + 1])
	end
      elsif (first_pos <= end_pos && end_pos <= last_pos) then
	yield(content_fragment[first_pos - curr_pos, end_pos - first_pos + 1])
      end
    end
    private :cut_range

    def modify_request(sid, srv_opt, context, prefix, request, response)
      context[:request] = request
      self
    end

    def modify_response(sid, srv_opt, context, prefix, request, response)
      request = context[:request]
      if (response.status == 200 && request.header.range) then
	if (request.header.if_range) then
	  if (mtime = response.header.last_modified) then
	    mtime = Time::gm(*ParseDate::parsedate(mtime)[0, 6])
	  end
	  if (request.condition(mtime, response.header.etag) == 206) then
	    response.status = 206
	  end
	else
	  response.status = 206
	end
	if (response.status == 206) then
	  unless (range = parse_range(request.header.range)) then
	    response.status = 200
	  end
	end
      else
	range = nil
      end

      if (range) then
	response.header.content_md5 = nil
	if (content_length = response.header.content_length) then
	  range = normalize_range(range[0], range[1], content_length.to_i)
	  response.header.content_range = "bytes #{range[0]}-#{range[1]}/#{content_length}"
	  response.header.content_length = (range[1] - range[0] + 1).to_s
	  yield(response)

	  context[:state] = :through
	  context[:range] = range
	  context[:position] = 0
	else
	  context[:state] = :spool
	  context[:range] = range
	  context[:response] = response
	  context[:content_length] = 0
	  context[:content_spool] = []
	end
      else
	yield(response)
      end

      self
    end

    def modify_content(sid, srv_opt, context, content_fragment, *opts)
      case (context[:state])
      when :through
	cut_range(content_fragment, context[:position], *context[:range]) do |range_fragment|
	  yield(range_fragment)
	end
	context[:position] += content_fragment.length
      when :spool
	context[:content_length] += content_fragment.length
	context[:content_spool].push(content_fragment)
      else
	yield(content_fragment, *opts)
      end

      self
    end

    def modify_terminal(sid, srv_opt, context)
      case (context[:state])
      when :through
	# nothing to do.
      when :spool
	curr_pos = 0
	first_pos, last_pos =
	  normalize_range(context[:range][0], context[:range][1], context[:content_length])
	response = context[:response]
	response.header.content_range = "bytes #{first_pos}-#{last_pos}/#{context[:content_length]}"
	response.header.content_length = (last_pos - first_pos + 1).to_s
	yield(response)

	for content_fragment in context[:content_spool]
	  cut_range(content_fragment, curr_pos, first_pos, last_pos) do |range_fragment|
	    yield(range_fragment)
	  end
	  curr_pos += content_fragment.length
	end
      end

      self
    end
  end
end
