# $Id: header.rb,v 1.6.2.2 2002/02/27 14:17:01 toki Exp $
# -
# HTTP header

=begin
Copyright (c) 2001-2002
	TOKI Yoshinori <toki@freedom.ne.jp>. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1.  Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

  2.  Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

  3.  The name of the author may not be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.
=end

require 'thread'
require 'wwwsrv/const'

module WWWsrv
  class HeaderParseError < ParseError; end

  class Header
    include Enumerable

    IRREG_WORD = {
      'WWW'  => 'WWW',
      'ETAG' => 'ETag',
      'MD5'  => 'MD5',
      'TE'   => 'TE'
    }

    FIELD_NAME_CACHE = {}
    FIELD_NAME_MAX_ENTRIES = 1000
    FIELD_NAME_LOCK = Mutex::new

    ATTR_CACHE = {}
    ATTR_MAX_ENTRIES = 1000
    ATTR_LOCK = Mutex::new

    def self::normalize(field_name)
      unless (normalized_field_name = FIELD_NAME_CACHE[field_name]) then
	FIELD_NAME_LOCK.synchronize do
	  unless (normalized_field_name = FIELD_NAME_CACHE[field_name]) then
	    field_name_segments = field_name.split(/-/n, -1)
	    field_name_segments.map!{ |word| IRREG_WORD[word.upcase] || word.capitalize }
	    normalized_field_name = field_name_segments.join('-')
	    FIELD_NAME_CACHE.clear if (FIELD_NAME_CACHE.size > FIELD_NAME_MAX_ENTRIES) # defence of cache overflow
	    FIELD_NAME_CACHE[field_name] = normalized_field_name
	  end
	end
      end

      normalized_field_name
    end

    def initialize
      @field_alist = []
      @multi_field = {
	'Set-Cookie'  => true,
	'Set-Cookie2' => true
      }
      @attr_reader = {}
      @attr_writer = {}
    end

    def size
      @field_alist.size
    end

    def length
      @field_alist.length
    end

    def empty?
      @field_alist.empty?
    end

    def set_multi_field(field_name)
      @multi_field[Header::normalize(field_name)] = true
      self
    end

    def unset_multi_field(field_name)
      @multi_field.delete(Header::normalize(field_name))
      self
    end

    def regist_attr(field_name)
      unless (attr_pair = ATTR_CACHE[field_name]) then
	ATTR_LOCK.synchronize do
	  unless (attr_pair = ATTR_CACHE[field_name]) then
	    # attribute reader
	    attr_name = field_name.downcase
	    attr_name.gsub!(/-/n, '_')
	    attr_reader = attr_name.intern

	    # attribute writer
	    attr_name.concat('=')
	    attr_writer = attr_name.intern

	    attr_pair = [ attr_reader, attr_writer ]
	    ATTR_CACHE.clear if (ATTR_CACHE.size > ATTR_MAX_ENTRIES) # defence of cache overflow
	    ATTR_CACHE[field_name] = attr_pair
	  end
	end
      end

      attr_reader, attr_writer = attr_pair
      @attr_reader[attr_reader] = field_name
      @attr_writer[attr_writer] = field_name
    end
    private :regist_attr

    def push(field_name, field_value)
      field_name = Header::normalize(field_name)
      if (field_pair = @field_alist.assoc(field_name)) then
	field_pair[1].push(field_value) if field_value
      else
	regist_attr(field_name)
	@field_alist.push([ field_name, [ field_value ] ]) if field_value
      end

      field_value
    end

    def pop(field_name)
      field_name = Header::normalize(field_name)
      i = @field_alist.length - 1
      @field_alist.reverse_each do |name, values|
	if (name == field_name) then
	  field_value = values.pop
	  @field_alist.delete_at(i) if (values.empty?)
	  return field_value
	end
	i -= 1
      end

      nil
    end

    def set_first(field_name)
      field_name = Header::normalize(field_name)
      field_values = nil
      @field_alist.delete_if{ |name, values|
	if (name == field_name) then
	  field_values = values
	  true
	end
      }
      @field_alist.unshift([ field_name, field_values ]) if field_values

      self
    end

    def set_last(field_name)
      field_name = Header::normalize(field_name)
      field_values = nil
      @field_alist.delete_if{ |name, values|
	if (name == field_name) then
	  field_values = values
	  true
	end
      }
      @field_alist.push([ field_name, field_values ]) if field_values

      self
    end

    def fields(field_name)
      field_name = Header::normalize(field_name)
      if (field_pair = @field_alist.assoc(field_name)) then
	return field_pair[1]
      else
	return []
      end
    end

    def [](field_name)
      field_name = Header::normalize(field_name)
      if (field_pair = @field_alist.assoc(field_name)) then
	return field_pair[1].join(', ')
      end

      nil
    end

    def []=(field_name, field_value)
      field_name = Header::normalize(field_name)
      if (field_pair = @field_alist.assoc(field_name)) then
	if (field_value) then
	  field_pair[1] = [ field_value ]
	else
	  @field_alist.delete_if{ |name, values| name == field_name }
	end
      else
	push(field_name, field_value) if field_value
      end

      field_value
    end

    def split_token(field_name, parse_params=false, qvalue_order=false)
      field_values = fields(field_name)
      unless (field_values.empty?) then
	tokens = field_values.map{ |value| value.split(/\s*,\s*/n) }
	tokens.flatten!
	return tokens unless parse_params

	tokens.map!{ |segment|
	  token, *params = segment.split(/\s*;\s*/n)
	  params.map!{ |param|
	    param_name, param_value = param.split(/=/n, 2)
	    if (param_value && param_value =~ /^".*"$/n) then
	      param_value.sub!(/^"/n, '')
	      param_value.sub!(/"$/n, '')
	    end
	    [ param_name, param_value ]
	  }
	  [ token ] + params
	}
	return tokens unless qvalue_order

	i = 0
	tokens.map!{ |token, *params|
	  i += 1
	  qvalue = 1
	  if (qpair = params.assoc('q')) then
	    qvalue = qpair[1].to_f
	  end
	  [ qvalue, i, token , params ]
	}
	tokens.sort!{ |(qvalue1, order1, token1, params1), (qvalue2, order2, token2, params2)|
	  cmp = qvalue2 <=> qvalue1
	  cmp = order1 <=> order2 if (cmp == 0)
	  cmp
	}
	tokens.map!{ |qvalue, order, token, params| [ token ] + params }

	return tokens
      end

      []
    end

    # for backward compatibility
    alias token_list split_token

    # attribute accessor
    def method_missing(attr_id, *args)
      case (args.length)
      when 0
	if (field_name = @attr_reader[attr_id]) then
	  return self[field_name]
	else
	  return nil
	end
      when 1
	field_name = @attr_writer[attr_id]  or
	  field_name = attr_id.id2name.sub(/=$/n, '').gsub(/_/n, '-')
	field_value = args.shift
	return self[field_name] = field_value
      else
	raise "invalid arguments of #{attr_id}: " + args.map{ |argument|
	  argument.inspect
	}.join(', ')
      end
    end

    def each
      for field_name, field_values in @field_alist
	if (@multi_field.include? field_name) then
	  for field_value in field_values
	    yield(field_name, field_value)
	  end
	else
	  unless (field_values.empty?) then
	    yield(field_name, field_values.join(', '))
	  end
	end
      end

      self
    end

    def parse(input)
      field_name = nil
      field_value = nil
      while (line = input.gets(LF))
	line.chomp!(LF)
	line.chomp!(CR)
	line.empty? and break
	if (line !~ /^\s/n) then
	  field_value.strip! if field_value # strip previous field value
	  field_name, field_value = line.split(/:/n, 2)
	  $& or raise HeaderParseError,
	    "failed to parse HTTP header field: #{line.inspect}"
	  field_name = Header::normalize(field_name)
	  regist_attr(field_name)
	  if (field_pair = @field_alist.assoc(field_name)) then
	    field_pair[1].push(field_value)
	  else
	    field_pair = [ field_name, [ field_value ] ]
	    @field_alist.push(field_pair)
	  end
	else
	  field_name or raise HeaderParseError,
	    "failed to parse HTTP header field: #{line.inspect}"
	  field_value.concat(line)
	end
      end
      field_value.strip! if field_value # strip last field value

      self
    end

    def self::parse(input, header=self::new)
      header.parse(input)
    end
  end
end
