# $Id: logger.rb,v 1.2.2.1 2002/01/14 13:13:21 toki Exp $
# -
# logging utilities

=begin
Copyright (c) 2001-2002
	TOKI Yoshinori <toki@freedom.ne.jp>. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1.  Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

  2.  Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

  3.  The name of the author may not be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.
=end

require 'wwwsrv/const'

module WWWsrv
  class Logger
    # logging level
    INTERNAL = 0
    DEBUG    = 1
    SYSTEM   = 2
    MESSAGE  = 3
    WARNING  = 4
    ALERT    = 5
    ERROR    = 6

    # for backward compatibility
    RAW      = INTERNAL
    DISCREET = SYSTEM

    def initialize(output=STDOUT, level=MESSAGE)
      @output = output
      @level = level
    end

    def print(priority, *messages)
      if (priority >= @level) then
	message = "#{Time::now}: pid #{$$}: "
	if (messages.empty?) then
	  message.concat(yield)
	else
	  for message_fragment in messages
	    message.concat(message_fragment)
	  end
	end
	message.concat("\n")
	@output.write(message)
      end
      self
    end

    def internal(*msgs, &bock)
      print(INTERNAL, *msgs, &block)
    end

    def debug(*msgs, &block)
      print(DEBUG, *msgs, &block)
    end

    def system(*msgs, &block)
      print(SYSTEM, *msgs, &block)
    end

    def message(*msgs, &block)
      print(MESSAGE, *msgs, &block)
    end

    def warning(*msgs, &block)
      print(WARNING, *msgs, &block)
    end

    def alert(*msgs, &block)
      print(ALERT, *msgs, &block)
    end

    def error(*msgs, &block)
      print(ERROR, *msgs, &block)
    end

    def close
      @output.close
      self
    end

    # for backward compatibility
    alias discreet system
  end

  class PreservedLogger < Logger
    def close
      self
    end
  end

  class MultiLogger < Logger
    def initialize
      @log_vec = []
    end

    def add_logger(log)
      @log_vec.push(log)
    end

    def print(priority, *messages, &block)
      for log in @log_vec
	log.print(priority, *messages, &block)
      end

      self
    end

    def close
      for log in @log_vec
	log.close
      end

      self
    end
  end

  class RequestWatcher
    def watch_request(sid, request)
      self
    end

    def watch_request_content(sid, fragment, *opts)
      self
    end

    def watch_request_trailer(sid, trailer)
      self
    end
  end

  class ResponseWatcher
    def watch_response(sid, response)
      self
    end

    def watch_response_content(sid, fragment, *opts)
      self
    end

    def watch_response_trailer(sid, trailer)
      self
    end
  end

  class MultiRequestWatcher < RequestWatcher
    def initialize
      super()
      @watcher_list = []
    end

    def add_request_watcher(watcher)
      @watcher_list.push(watcher)
    end

    def watch_request(sid, request)
      for watcher in @watcher_list
	watcher.watch_request(sid, request)
      end

      self
    end

    def watch_request_content(sid, fragment, *opts)
      for watcher in @watcher_list
	watcher.watch_request_content(sid, fragment, *opts)
      end

      self
    end

    def watch_request_trailer(sid, trailer)
      for watcher in @watcher_list
	watcher.watch_request_trailer(sid, trailer)
      end

      self
    end
  end

  class MultiResponseWatcher < ResponseWatcher
    def initialize
      super()
      @watcher_list = []
    end

    def add_response_watcher(watcher)
      @watcher_list.push(watcher)
    end

    def watch_response(sid, response)
      for watcher in @watcher_list
	watcher.watch_response(sid, response)
      end

      self
    end

    def watch_response_content(sid, fragment, *opts)
      for watcher in @watcher_list
	watcher.watch_response_content(sid, fragment, *opts)
      end

      self
    end

    def watch_response_trailer(sid, trailer)
      for watcher in @watcher_list
	watcher.watch_response_trailer(sid, trailer)
      end

      self
    end
  end

  class RequestLogger < RequestWatcher
    def initialize(logger,
		   request_line_logging_priority=Logger::MESSAGE,
		   header_logging_priority=Logger::DEBUG,
		   content_logging_priority=Logger::INTERNAL,
		   trailer_logging_priority=Logger::DEBUG)
      super()
      @log = logger
      @request_line_logging_priority = request_line_logging_priority
      @header_logging_priority = header_logging_priority
      @content_logging_priority = content_logging_priority
      @trailer_logging_priority = trailer_logging_priority
    end

    attr_accessor :request_line_logging_priority
    attr_accessor :header_logging_priority
    attr_accessor :content_logging_priority
    attr_accessor :trailer_logging_priority

    def watch_request(sid, request)
      @log.print(@request_line_logging_priority,
		 "sid #{sid}: request line: ", request.line)

      for field_name, field_value in request.header
	@log.print(@header_logging_priority,
		   "sid #{sid}: request header: ", field_name, ': ', field_value)
      end

      self
    end

    def watch_request_content(sid, content_fragment, *opts)
      @log.print(@content_logging_priority) {
	"sid #{sid}: receiving octets: " +
	  (content_fragment ? content_fragment.length.to_s : '0') +
	  ((! opts.empty?) ? ': option: ' + opts.join(', ') : '') +
	  ': content: ' + content_fragment.inspect
      }

      self
    end

    def watch_request_trailer(sid, trailer)
      for field_name, field_value in trailer
	@log.print(@trailer_logging_priority,
		   "sid #{sid}: request trailer: ", field_name, ': ', field_value)
      end

      self
    end
  end

  class ResponseLogger < ResponseWatcher
    def initialize(logger,
		   status_line_logging_priority=Logger::SYSTEM,
		   header_logging_priority=Logger::DEBUG,
		   content_logging_priority=Logger::INTERNAL,
		   trailer_logging_priority=Logger::DEBUG)
      super()
      @log = logger
      @status_line_logging_priority = status_line_logging_priority
      @header_logging_priority = header_logging_priority
      @content_logging_priority = content_logging_priority
      @trailer_logging_priority = trailer_logging_priority
    end

    attr_accessor :status_line_logging_priority
    attr_accessor :header_logging_priority
    attr_accessor :content_logging_priority
    attr_accessor :trailer_logging_priority

    def watch_response(sid, response)
      @log.print(@status_line_logging_priority,
		 "sid #{sid}: status line: ", response.line)

      for field_name, field_value in response.header
	@log.print(@header_logging_priority,
		   "sid #{sid}: response header: ", field_name, ': ', field_value)
      end

      self
    end

    def watch_response_content(sid, content_fragment, *opts)
      @log.print(@content_logging_priority) {
	"sid #{sid}: sending octets: " +
	  (content_fragment ? content_fragment.length.to_s : '0') +
	  ((! opts.empty?) ? ': option: ' + opts.join(', ') : '') +
	  ': content: ' + content_fragment.inspect
      }

      self
    end

    def watch_response_trailer(sid, trailer)
      for field_name, field_value in trailer
	@log.print(@trailer_logging_priority,
		   "sid #{sid}: response trailer: ", field_name, ': ', field_value)
      end

      self
    end
  end

  # for backward compatibility
  RequestPeeker = RequestLogger
  ResponsePeeker = ResponseLogger
end
