# $Id: request.rb,v 1.4.2.1 2002/01/14 13:13:22 toki Exp $
# -
# HTTP request

=begin
Copyright (c) 2001-2002
	TOKI Yoshinori <toki@freedom.ne.jp>. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1.  Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

  2.  Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

  3.  The name of the author may not be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.
=end

require 'parsedate'
require 'wwwsrv/const'
require 'wwwsrv/path'
require 'wwwsrv/version'
require 'wwwsrv/header'
require 'wwwsrv/srvopt'
require 'wwwsrv/form'

module WWWsrv
  class Host
    HTTP_PORT = 80

    def initialize(name, port=HTTP_PORT)
      @name = name
      @port = port
    end

    attr_reader :name
    attr_reader :port

    def to_s
      "#{@name}:#{@port}"
    end

    def ==(host)
      (host.kind_of? type) && @name = host.name && @port == host.port
    end
  end

  class Client
    def initialize(name, address)
      @name = name
      @address = address
    end

    attr_reader :name
    attr_reader :address

    def to_s
      if (@name == @address) then
	@address
      else
	"#{@name}(#{@address})"
      end
    end

    def to_i
      adrs_vec = @address.split(/\./n, 4)
      n = 0
      for i in adrs_vec
	n <<= 8
	n |= i.to_i
      end

      n
    end

    def ==(client)
      (client.kind_of? type) &&	@name == client.name &&	@address == client.address
    end
  end

  class RequestLineParseError < ParseError; end
  class NotSupportedVersionError < ParseError; end

  class Request
    include Path
    include FormData

    def initialize(method=nil, path=nil, version=nil, header=nil, content=nil)
      @method = method
      self.path = path
      @version = version
      @header = header
      @content = content
      @host = nil
      @client = nil
      @option = Accessor::wrap(Hash::new)
      @form_data = nil
      @params = nil
    end

    alias object_method method

    attr_accessor :method
    attr_accessor :path
    attr_accessor :version
    attr_accessor :header
    attr_accessor :content
    attr_accessor :host
    attr_accessor :client
    attr_reader   :option

    def path=(new_path)
      if (@path = new_path) then
	@normalized_path, @query = normalize_path(@path)
      else
	@normalize_path = nil
	@query = nil
      end

      @path
    end

    def subpath(prefix)
      Path::subpath(prefix, @normalized_path)
    end

    # for backward compatibility
    alias reduce_path subpath

    attr_reader :normalized_path
    attr_reader :query

    def line
      if (@version >= HTTP_1_0) then
	return "#{@method} #{@path} #{@version}"
      else
	return "#{@method} #{@path}"
      end
    end

    def condition(mtime, etag=nil)
      if (@header.if_range) then
	unless (@header.if_match || @header.if_none_match) then
	  unless (@header.if_modified_since || @header.if_unmodified_since) then
	    if (@header.if_range) then
	      time_param = ParseDate::parsedate(@header.if_range)[0, 6]
	      if (time_param.include? nil) then
		# Entity Tag
		if (@header.if_range == etag) then
		  return 206
		end
	      else
		# HTTP date
		timestamp = Time::gm(*time_param)
		if (timestamp == mtime) then
		  return 206
		end
	      end
	    end
	  end
	end
      else
	unless (@header.if_match || @header.if_unmodified_since) then
	  if (etag && @header.if_none_match) then
	    unless (@header.if_none_match != '*' && ! (@header.split_token('If-None-Match').include? etag)) then
	      return 304
	    end
	  elsif (mtime && @header.if_modified_since) then
	    timestamp = Time::gm(*ParseDate::parsedate(@header.if_modified_since)[0, 6]) rescue return 200
	    unless (timestamp != mtime) then
	      return 304
	    end
	  end
	end

	unless (@header.if_none_match || @header.if_modified_since) then
	  if (etag && @header.if_match) then
	    unless (@header.if_match == '*' || (@header.split_token('If-Match').include? etag)) then
	      return 412
	    end
	  elsif (mtime && @header.if_unmodified_since) then
	    timestamp = Time::gm(*ParseDate::parsedate(@header.if_unmodified_since)[0, 6]) rescue return 200
	    unless (timestamp == mtime) then
	      return 412
	    end
	  end
	end
      end

      200
    end

    def trailer
      @content.trailer
    end

    def form_data
      unless (@form_data) then
	param_data = nil
	if (@method == 'POST') then
	  case (@header.content_type)
	  when %r!^multipart/form-data!ni
	    @form_data = parse_form_data(@content, @header.content_type)
	    @content.read	# clear multipart epilogue
	    @params = {}
	    for form_name, form_values in @form_data
	      @params[form_name] = form_values.map{ |header, body|
		if (header.content_type && header.content_type =~ %r!^multipart/mixed!ni) then
		  body.map{ |subheader, subbody| subbody }
		else
		  body
		end
	      }
	    end
	    return @form_data
	  when %r!^application/x-www-form-urlencoded!ni
	    param_data = @content.read
	  end
	else
	  param_data = @query
	end

	if (param_data) then
	  @params = parse_query(param_data)
	  @form_data = {}
	  for param_name, param_values in @params
	    @form_data[param_name] = param_values.map{ |value| [ Header::new, value ] }
	  end
	else
	  @params = {}
	  @form_data = {}
	end
      end

      @form_data
    end

    def params
      form_data unless @params
      @params
    end

    alias param params		# for backward compatibility

    def self::parse_line(input)
      while (request_line = input.gets(LF))
	request_line.chomp!(LF)
	request_line.chomp!(CR)
	request_line.empty? and next

	method, path, version = request_line.split(/\s+/n, 3)
	(! method || method.empty?) and
	  raise RequestLineParseError, 'undefined HTTP request method'
	(! path || path.empty?) and
	  raise RequestLineParseError, 'undefined HTTP request path'
	if (version) then
	  http_version = HTTPversion::parse(version)
	else
	  http_version = HTTP_0_9
	end
	if (http_version >= HTTP_2_0) then
	  raise NotSupportedVersionError, "not supported HTTP version: #{http_version}"
	end

	return method, path, http_version
      end

      nil
    end

    def self::parse(input)
      if (request_line = parse_line(input)) then
	request_method, request_path, http_version = request_line
	if (http_version >= HTTP_1_0) then
	  request_header = Header::parse(input)
	else
	  request_header = Header::new
	end
	return Request::new(request_method, request_path, http_version, request_header)
      end

      nil
    end
  end
end
