# $Id: server.rb,v 1.26.2.2 2002/02/27 14:17:06 toki Exp $
# -
# HTTP server

=begin
Copyright (c) 2001-2002
	TOKI Yoshinori <toki@freedom.ne.jp>. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1.  Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

  2.  Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

  3.  The name of the author may not be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.
=end

require 'wwwsrv/const'
require 'wwwsrv/thread'
require 'wwwsrv/process'
require 'wwwsrv/ioutil'
require 'wwwsrv/htmlconv'
require 'wwwsrv/parser'
require 'wwwsrv/logger'
require 'wwwsrv/request'
require 'wwwsrv/response'
require 'wwwsrv/content'
require 'wwwsrv/document'
require 'wwwsrv/alert'

module WWWsrv
  class WriteBuffer
    include Timeout
    include ConnectionProtect

    def initialize(output, timeout, threshold)
      @output = output
      @timeout = timeout
      @threshold = threshold
      @buffer = ''
      @last_flush_time = Time::now
      @input_wait = []
      @output_wait = [ nil ]
      @except_wait = []
      @write_sem = Semaphore::new
    end

    def _write(fragment)
      wait(@timeout) do
	protect_connection do
	  @output.write(fragment)
	end
      end
    end
    private :_write

    def _flush
      wait(@timeout) do
	protect_connection do
	  @output.write(@buffer) unless @buffer.empty?
	  @output.flush
	  @last_flush_time = Time::now
	end
      end
      @buffer = ''
    end
    private :_flush

    def write(fragment)
      @write_sem.synchronize do
	if (fragment.length >= @threshold) then
	  unless (@buffer.empty?) then
	    _write(@buffer)
	    @buffer = ''
	  end
	  _write(fragment)
	else
	  @buffer.concat(fragment)
	  if (@buffer.length >= @threshold) then
	    _write(@buffer)
	    @buffer = ''
	  end
	end
      end

      self
    end

    def flush
      @write_sem.synchronize do
	_flush
      end

      self
    end

    def writable?(output)
      @output_wait[0] = output
      IO::select(@input_wait, @output_wait, @except_wait, 0) ? true : false
    end
    private :writable?

    def try_flush(veto=0)
      lock = false
      begin
	if (lock = @write_sem.try_wait) then
	  curr_time = Time::now
	  if (curr_time - @last_flush_time >= veto) then
	    if (raw_io = @output.to_io) then
	      if (writable? raw_io) then
		_flush
		return self
	      end
	    else
	      _flush
	      return self
	    end
	  end
	end
	return nil
      ensure
	@write_sem.post if lock
      end
    end
  end

  class ResponseWriter
    include HTMLconv
    include Parser
    include ConnectionProtect

    def initialize(output, request, watcher)
      @output = output
      @request = request
      @watcher = watcher
      @response = nil
      @send_method = nil
      @send_length = nil
      @content_spool = nil
      @content_length = nil
      @last_chunk = false
    end

    attr_reader :response

    def used?
      @send_method != nil
    end

    def write(sid, arg, *opts)
      unless (@response) then
	@response = arg
	if (@response.status < 500) then
	  unless (@response.header.date) then
	    @response.header.date = http_date(Time::now)
	    @response.header.set_first('Date')
	  end
	end
	if (@request) then
	  if (@content_length = @response.header.content_length) then
	    @content_length = @content_length.to_i
	    @send_length = @content_length
	  end

	  if (@request.version >= HTTP_1_0) then
	    if (@request.version >= HTTP_1_1 && @response.version >= HTTP_1_1) then
	      # HTTP/1.1
	      if (transfer_encoding = @response.header.transfer_encoding) then
		if (transfer_encoding.downcase != CHUNKED) then
		  raise "not supported transfer encoding: #{transfer_encoding}"
		end
	      else
		@response.header.content_length = nil
		@response.header.transfer_encoding = CHUNKED
	      end
	      @send_method = :chunked
	    else
	      # HTTP/1.0
	      if (@content_length) then
		@send_method = :through
	      else
		@send_method = :spool
		@content_spool = []
		@content_length = 0
	      end
	    end

	    if (@send_method != :spool) then
	      set_connection(@request, @response)
	      send_response(sid, @response)
	    end
	  else
	    # HTTP/0.9
	    @send_method = :through
	  end
	else
	  # HTTP/Unknown
	  if (@response.header.content_length) then
	    @send_method = :through
	    set_connection(@request, @response)
	  else
	    @send_method = :spool
	    @content_spool = []
	    @content_length = 0
	  end
	end
      else
	if (@send_length) then
	  @send_length -= arg.length
	  (@send_length < 0) and raise 'mismatch content length'
	end

	case (@send_method)
	when :chunked
	  send_chunk(sid, arg, *opts)
	when :through
	  send_content(sid, arg)
	when :spool
	  @content_spool.push(arg)
	  @content_length += arg.length
	else
	  raise "unknown send method: #{@send_method}"
	end
      end

      self
    end

    def terminate(sid)
      unless (@response) then
	raise "empty response"
      end

      if (@request && @request.method != 'HEAD') then
	if (@send_length && @send_length != 0) then
	  raise 'mismatch content length'
	end
      else
	if (@send_length && @send_length != @content_length) then
	  raise 'invalid response message body for HEAD request'
	end
      end

      case (@send_method)
      when :chunked
	if (@request && @request.method != 'HEAD') then
	  unless (@last_chunk) then
	    send_chunk(sid, nil)
	  end
	  @watcher.watch_response_trailer(sid, @response.trailer)
	  send_header(sid, @response.trailer)
	end
      when :through
	# nothing to do
      when :spool
	@response.header.content_length = @content_length.to_s
	set_connection(@request, @response)
	send_response(sid, @response)
	if (! @request || @request.method != 'HEAD') then
	  for content_fragment in @content_spool
	    send_content(sid, content_fragment)
	  end
	end
      else
	raise "unknown send method: #{@send_method}"
      end

      self
    end

    def set_connection(request, response)
      req_conn = request && request.header.connection
      res_conn = response.header.connection

      (res_conn && res_conn =~ /close/n) and return
      if (! request || (req_conn && req_conn =~ /close/n)) then
	response.header.push('Connection', 'close')
	return
      end

      if (request.version <= HTTP_1_0 || response.version <= HTTP_1_0) then
	if (req_conn && req_conn =~ /keep-alive/ni) then
	  response.header.push('Connection', 'Keep-Alive')
	else
	  response.header.push('Connection', 'close')
	end
      end
    end
    private :set_connection

    def send_response(sid, response)
      @watcher.watch_response(sid, response)

      res_msg = ''
      res_msg.concat(response.line)
      res_msg.concat(CRLF)
      for field_name, field_value in response.header
	res_msg.concat(field_name)
	res_msg.concat(': ')
	res_msg.concat(field_value)
	res_msg.concat(CRLF)
      end
      res_msg.concat(CRLF)

      protect_connection do
	@output.write(res_msg)
      end
    end
    private :send_response

    def send_header(sid, header)
      head_msg = ''
      for field_name, field_value in header
	head_msg.concat(field_name)
	head_msg.concat(': ')
	head_msg.concat(field_value)
	head_msg.concat(CRLF)
      end
      head_msg.concat(CRLF)

      protect_connection do
	@output.write(head_msg)
      end
    end
    private :send_header

    def send_chunk(sid, chunk_data, *chunk_extension)
      if (! @request || @request.method != 'HEAD') then
	@watcher.watch_response_content(sid, chunk_data, *chunk_extension) if chunk_data

	chunk_msg = ''
	if (! chunk_data || ! chunk_data.empty?) then
	  if (chunk_data) then
	    chunk_msg.concat(format('%x', chunk_data.length))
	  else
	    chunk_msg.concat('0')
	    @last_chunk = true
	  end
	  unless (chunk_extension.empty?) then
	    for chunk_ext_name, chunk_ext_value in chunk_extension
	      chunk_msg.concat('; ')
	      chunk_msg.concat(chunk_ext_name)
	      if (chunk_ext_value) then
		chunk_msg.concat('=')
		chunk_msg.concat(chunk_ext_value)
	      end
	    end
	  end
	  chunk_msg.concat(CRLF)
	  if (chunk_data) then
	    chunk_msg.concat(chunk_data)
	    chunk_msg.concat(CRLF)
	  end
	end

	protect_connection do
	  @output.write(chunk_msg)
	end
      end
    end
    private :send_chunk

    def send_content(sid, content)
      if (! @request || @request.method != 'HEAD') then
	@watcher.watch_response_content(sid, content)
	protect_connection do
	  @output.write(content)
	end
      end
    end
    private :send_content
  end

  class Server
    include Timeout
    include ConnectionProtect
    include DocumentCascade

    undef setup
    undef cleanup

    SID_LOCK = Mutex::new
    MAX_CONNECTIONS = 64
    TIMEOUT = 300
    KEEP_ALIVE = 15
    BUFFER_SIZE = 1024 * 32
    FLUSH_INTERVAL = 2

    @@sid = 'A'

    def initialize(srv_sock, alert=AlertDocument::new, document=PartialContent::new(DocumentFolder::new))
      @sid = SID_LOCK.synchronize{ sid = @@sid; @@sid = @@sid.succ; sid }

      @closed = false
      @srv_sock = srv_sock
      @accept_sem = Semaphore::new
      @accept_thread = nil
      @connections = MAX_CONNECTIONS
      @conn_sem = Semaphore::new(@connections)
      @dispatch_threads = ThreadGroup::new

      @timeout = TIMEOUT
      @keep_alive = KEEP_ALIVE
      @flush_interval = FLUSH_INTERVAL
      @buffer_size = BUFFER_SIZE

      srv_family, srv_port, srv_name, srv_addr = @srv_sock.addr
      @host = Host::new(srv_name, srv_port)
      @alert = alert
      @document = document

      @log = WWWsrv::MultiLogger::new
      @alias = nil
      @access = nil

      @request_watcher = MultiRequestWatcher::new
      @response_watcher = MultiResponseWatcher::new
      @request_logger = RequestLogger::new(@log)
      @response_logger = ResponseLogger::new(@log)
      @request_watcher.add_request_watcher(@request_logger)
      @response_watcher.add_response_watcher(@response_logger)
    end

    attr_accessor :alert
    attr_accessor :timeout
    attr_accessor :keep_alive
    attr_reader   :connections
    attr_accessor :buffer_size
    attr_accessor :flush_interval

    # for backward compatibility
    alias alarm alert
    alias alarm= alert=

    def connections=(new_conns)
      @conn_sem = Semaphore::new(new_conns)
      @connections = new_conns
    end

    def curr_conns
      @connections - @conn_sem.value
    end

    def hostname
      @host.name
    end

    def port
      @host.port
    end

    attr_reader :document

    def document=(new_doc)
      @alias = nil
      @access = nil
      @document = new_doc
    end

    def mount(document, path, *opts)
      @document.mount(document, path, *opts)
    end

    def umount(path)
      @document.umount(path)
    end

    def add_logger(logger)
      @log.add_logger(logger)
    end

    def add_request_watcher(watcher)
      @request_watcher.add_request_watcher(watcher)
    end

    def add_response_watcher(watcher)
      @response_watcher.add_response_watcher(watcher)
    end

    attr_reader :request_logger
    attr_reader :response_logger

    # for backward compatibility
    alias request_peeker request_logger
    alias response_peeker response_logger

    # see WWWsrv::AccessControl#add
    def add_access_control(*args, &block)
      unless (@access) then
	@access = WWWsrv::AccessControl::new(@document)
	@document = @access
      end
      @access.add_access_control(*args, &block)

      self
    end

    def alias(dst_path, src_path)
      unless (@alias) then
	@alias = WWWsrv::DocumentAlias::new(@document)
	@document = @alias
      end
      @alias.alias(dst_path, src_path)

      self
    end

    def logging_backtrace(sid, except)
      i = 0
      for frame in except.backtrace
	if (i < 10) then
	  @log.debug("sid #{sid}: backtrace: #{frame}")
	  i += 1
	else
	  @log.debug("sid #{sid}: backtrace: ...")
	  break
	end
      end
    end
    private :logging_backtrace

    def make_server_option
      ServerOption::new(@document, @log, 'http')
    end
    private :make_server_option

    def make_response(status=200, version=HTTP_1_1)
      response = Response::new(version, status)
      response.header.server = PRODUCT
      response
    end
    private :make_response

    def make_fatal_error_response(status=500)
      response = make_response(status)
      response.header.connection = 'close'
      response
    end
    private :make_fatal_error_response

    def buffering(wbuf, request)
      begin
	send = ResponseWriter::new(wbuf, request, @response_watcher)
	yield(send)
      ensure
	wbuf.flush
      end
    end
    private :buffering

    def receive_request(sock, count)
      begin
	if (count == 1) then
	  wait(@timeout) do
	    protect_connection do
	      return Request::parse(sock)
	    end
	  end
	else
	  wait(@keep_alive) do
	    protect_connection do
	      return Request::parse(sock)
	    end
	  end
	end
      rescue TimeoutError
	return nil
      end
    end
    private :receive_request

    def set_host(request)
      if (host = request.header.host) then
	host, port = host.split(/:/n, 2)
	request.host = Host::new(host, port ? port.to_i : @host.port)
      elsif (request.version <= HTTP_1_0) then
	request.host = @host
      else
	http_error(400, 'Not found a Host header field.', { 'Connection' => 'close' })
      end
    end
    private :set_host

    def make_content_reader(sid, sock, srv_opt, request)
      if (transfer_encoding = request.header.transfer_encoding) then
	if (transfer_encoding.downcase == CHUNKED) then
	  return ChunkedContentReader::new(sid, sock, @request_watcher, @timeout)
	else
	  err_msg = "unknown Transfer-Encoding of request body: #{transfer_encoding}"
	  http_error(400, err_msg, { 'Connection' => 'close' })
	end
      elsif (content_length = request.header.content_length) then
	content_length = content_length.to_i
	if (content_length >= 0) then
	  return LengthLimitedContentReader::new(sid, sock, @request_watcher, @timeout, content_length)
	else
	  err_msg = "negative Content-Length of request: #{content_length}"
	  http_error(400, err_msg, { 'Connection' => 'close' })
	end
      else
	return EmptyContentReader::instance
      end
    end
    private :make_content_reader

    def slight_alert(sid, send, srv_opt, request, err_args)
      status, message, header, version = err_args
      if (status.kind_of? Response) then
	# for backward compatibility
	response = status
      else
	response = make_response(status, version)
	for field_name, field_value in header
	  response.header[field_name] = field_value
	end
      end

      @log.warning("sid #{sid}: document error: #{response.line}")
      unless (send.used?) then
	srv_opt.error = message
	@alert.content(sid, srv_opt, '', request, response) do |arg, *opts|
	  send.write(sid, arg, *opts)
	end
	send.terminate(sid)
      end
    end
    private :slight_alert

    def fatal_alert(sid, wbuf, status, except, backtrace=false, not_send=false)
      response = make_fatal_error_response(status)
      @log.warning("sid #{sid}: server error: #{response.line}: #{$!} (#{$!.type})")
      logging_backtrace(sid, except) if backtrace
      unless (not_send) then
	buffering(wbuf, nil) do |send|
	  srv_opt = make_server_option
	  srv_opt.error = except
	  @alert.content(sid, srv_opt, '', nil, response) do |arg, *opts|
	    send.write(sid, arg, *opts)
	  end
	  send.terminate(sid)
	end
      end
    end
    private :fatal_alert

    def dispatch(sid, sock, wbuf)
      send = nil
      count = 1
      begin
	srv_family, srv_port, srv_name, srv_addr = protect_connection{ sock.addr }
	cli_family, cli_port, cli_name, cli_addr = protect_connection{ sock.peeraddr }
	client = Client::new(cli_name, cli_addr)
	begin
	  @log.message("sid #{sid}: open connection: ",
		       "#{srv_name}(#{srv_addr}):#{srv_port} <-> ",
		       "#{cli_name}(#{cli_addr}):#{cli_port}")
	  loop do
	    ssid = "#{sid}-#{count}"
	    srv_opt = make_server_option
	    request = receive_request(sock, count) or break
	    request.client = client
	    buffering(wbuf, request) do |send|
	      state, err_args = Document::catch_state{
		begin
		  set_host(request)
		  request.content = LineReader::new(make_content_reader(sid, sock, srv_opt, request))
		  @request_watcher.watch_request(ssid, request)
		  @document.content(ssid, srv_opt, '', request, make_response) do |arg, *opts|
		    send.write(ssid, arg, *opts)
		  end
		  send.terminate(ssid)
		rescue TimeoutError
		  @log.warning("sid #{ssid}: request expired: #{$!}")
		  http_error(408, "Request expired.", { 'Connection' => 'close' })
		end
	      }
	      slight_alert(ssid, send, srv_opt, request, err_args) if (state == :error)
	    end
	    (request.version < HTTP_1_0) and break
	    ((connection = request.header.connection) && connection =~  /close/n) and break
	    ((connection = send.response.header.connection) && connection =~ /close/n) and break
	    @log.system("sid #{sid}: keep alive")
	    count += 1
	  end
	rescue ConnectionError
	  raise
	rescue NotSupportedVersionError
	  fatal_alert(sid, wbuf, 505, $!)
	rescue ParseError
	  fatal_alert(sid, wbuf, 400, $!)
	rescue StandardError, ScriptError
	  fatal_alert(sid, wbuf, 500, $!, true, send && send.used?)
	ensure
	  @log.system("sid #{sid}: close connection")
	  protect_connection{ sock.close }
	end
      rescue ConnectionError
	@log.alert("sid #{sid}: connection error: #{$!} (#{$!.type})")
	logging_backtrace(sid, $!)
      end
    end
    private :dispatch

    def accept
      srv_family, srv_port, srv_name, srv_addr = @srv_sock.addr
      @log.alert("start: #{srv_name}(#{srv_addr}):#{srv_port}")

      i = 1
      begin
	accept_latch = Latch::new
	@accept_thread = Thread::new{
	  accept_latch.ready
	  @accept_sem.synchronize do
	    @document.setup(@log)
	    loop do
	      @closed and break
	      sid = "#{@sid}.#{i}"
	      begin
		@accept_sem.post
		@conn_sem.wait
		sock = @srv_sock.accept
	      rescue
		@log.warning("sid #{sid}: accept error: #{$!}")
		logging_backtrace(sid, $!)
		sock.close if sock
		break
	      ensure
		@accept_sem.wait
	      end
	      @log.system("sid #{sid}: connections: #{curr_conns}")

	      sock.binmode
	      dispatch_latch = Latch::new
	      dispatch_thread = Thread::new{
		dispatch_latch.ready
		wbuf = WriteBuffer::new(sock, @timeout, @buffer_size)
		flush_sem = Semaphore::new
		flush_sem.wait
		begin
		  async_flush_latch = Latch::new
		  async_flush_thread = Thread::new{
		    begin
		      async_flush_latch.ready
		      loop do
			flush_sem.post
			wait do
			  sleep(@flush_interval)
			end
			flush_sem.wait
			wbuf.try_flush(@flush_interval)
		      end
		    rescue TimeoutError
		      # nothing to do.
		    ensure
		      flush_sem.post
		    end
		  }
		  async_flush_thread.priority = TIMEOUT_PRIORITY
		  async_flush_latch.release
		  dispatch(sid, sock, wbuf)
		ensure
		  flush_sem.synchronize do
		    async_flush_thread.exit if async_flush_thread.alive?
		  end
		  @conn_sem.post
		end
	      }
	      dispatch_thread.priority = NORMAL_PRIORITY
	      dispatch_thread[:wwwsrv_sid] = sid
	      @dispatch_threads.add(dispatch_thread)
	      dispatch_latch.release

	      i += 1
	    end
	  end
	}
	@accept_thread.priority = SYSTEM_PRIORITY
	accept_latch.release
	@accept_thread.join
      ensure
	wait_latch = Latch::new
	wait_thread = Thread::new{
	  wait_latch.ready
	  for thread in @dispatch_threads.list
	    begin
	      thread.join
	    rescue TimeoutError
	      # nothing to do.
	    end

	    if (thread.key? :wwwsrv_sid) then
	      @log.warning("sid #{thread[:wwwsrv_sid]}: execution aborted.")
	    end
	  end
	}
	wait_thread.priority = SYSTEM_PRIORITY
	wait_latch.release

	intercpet_latch = Latch::new
	intercept_thread = Thread::new{
	  intercpet_latch.ready
	  while (wait_thread.alive?)
	    Timeout::intercept_all(@dispatch_threads)
	    sleep(0.1)
	  end

	  @document.cleanup(@log)
	  @log.close
	}
	intercept_thread.priority = SYSTEM_PRIORITY
	intercpet_latch.release

	intercept_thread.join
	wait_thread.join
      end

      self
    end

    def close
      @accept_sem.synchronize do
	unless (@closed) then
	  if (@accept_thread && @accept_thread.alive?) then
	    @accept_thread.exit
	  end
	  @closed = true
	end
      end

      self
    end

    def closed?
      @closed
    end
  end
end
