# $Id: thread.rb,v 1.6.2.2 2002/02/27 14:17:09 toki Exp $
# -
# thread utilities

=begin
Copyright (c) 2001-2002
	TOKI Yoshinori <toki@freedom.ne.jp>. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1.  Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

  2.  Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

  3.  The name of the author may not be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.
=end

require 'thread'
require 'wwwsrv/const'

module WWWsrv
  NORMAL_PRIORITY  = 0
  TIMEOUT_PRIORITY = 1
  SYSTEM_PRIORITY  = 2

  class Latch
    def initialize
      @latch = true
      @mutex = Mutex::new
      @cond = ConditionVariable::new
    end

    def ready
      @mutex.synchronize do
	while (@latch)
	  @cond.wait(@mutex)
	end
      end

      self
    end

    def release
      @mutex.synchronize do
	@latch = false
	@cond.broadcast
      end

      self
    end
  end

  class Semaphore
    def initialize(count=1)
      @count = count
      @mutex = Mutex::new
      @cond = ConditionVariable::new
    end

    def value
      @count
    end

    def wait
      @mutex.synchronize do
	until (@count > 0)
	  @cond.wait(@mutex)
	end
	@count -= 1
      end
    end

    def try_wait
      @mutex.synchronize do
	if (@count > 0) then
	  @count -= 1
	  return true
	else
	  return false
	end
      end
    end

    def post
      @mutex.synchronize do
	@cond.signal unless (@count > 0)
	@count += 1
      end
    end

    def synchronize
      wait
      begin
	return yield
      ensure
	post
      end
    end
  end

  module Timeout
    LOCK = Mutex::new
    WAITERS = {}

    # timeout == 0 means infinite timeout span.
    def wait(timeout=0)
      begin
	depth = nil
	LOCK.synchronize do
	  WAITERS[Thread::current] = [] unless (WAITERS.include? Thread::current)
	  WAITERS[Thread::current].push([ Time::now, timeout ])
	  depth = WAITERS[Thread::current].length
	end
	return yield
      ensure
	LOCK.synchronize do
	  if (WAITERS.include? Thread::current) then
	    if (WAITERS[Thread::current].length >= depth) then
	      WAITERS[Thread::current].pop
	      WAITERS.delete(Thread::current) if WAITERS[Thread::current].empty?
	    end
	  end
	end
      end
    end
    module_function :wait

    class << self
      def expire_if(thread, except, message)
	if (thread.alive?) then
	  nth = nil
	  WAITERS[thread].each_with_index do |(start, timeout), i|
	    if (yield(start, timeout)) then
	      nth = i
	      thread.raise(except, message) if thread.alive?
	      break
	    end
	  end

	  if (nth) then
	    WAITERS[thread][nth..-1] = []
	    WAITERS.delete(thread) if WAITERS[thread].empty?
	  end
	else
	  WAITERS.delete(thread)
	end
      end
      private :expire_if

      def intercept_all(thread_group=nil)
	unless (thread_group) then
	  LOCK.synchronize do
	    WAITERS.each_key do |thread|
	      expire_if(thread, TimeoutError, 'execution intercepted') { |start, timeout|
		true		# expire all
	      }
	    end
	  end
	else
	  LOCK.synchronize do
	    thread_group.list.each do |thread|
	      if (WAITERS.include? thread) then
		expire_if(thread, TimeoutError, 'execution intercepted') { |start, timeout|
		  true		# expire all
		}
	      end
	    end
	  end
	end

	self
      end
    end

    # daemon thread
    daemon_latch = Latch::new
    Thread::new{
      daemon_latch.ready
      loop do
	sleep(1)
	LOCK.synchronize do
	  t = Time::now
	  WAITERS.each_key do |thread|
	    expire_if(thread, TimeoutError, 'execution expired') { |start, timeout|
	      (timeout > 0) && (t - start >= timeout)
	    }
	  end
	end
      end
    }.priority = TIMEOUT_PRIORITY
    daemon_latch.release
  end
end
