# $Id: xmlrpcdoc.rb,v 1.3.2.1 2002/01/14 13:13:27 toki Exp $
# -
# XML-RPC document

=begin
Copyright (c) 2001-2002
	TOKI Yoshinori <toki@freedom.ne.jp>. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1.  Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.

  2.  Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.

  3.  The name of the author may not be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.
=end

require 'xmlrpc/client'
require 'xmlrpc/base64'
require 'wwwsrv/const'
require 'wwwsrv/path'
require 'wwwsrv/version'
require 'wwwsrv/document'
require 'wwwsrv/xmlrpcsrv'

module WWWsrv
  module XMLRPCDataConversion
    def degenerate_header(sid, log, header)
      header.to_a
    end
    module_function :degenerate_header

    def expand_header(sid, log, header, header_alist)
      header = Header::new
      for field_name, field_value in header_alist
	header[field_name] = field_value
      end

      header
    end
    module_function :expand_header

    def degenerate_option(sid, log, option)
      degenerated_option = {}
      for opt_key, opt_value in option
	if (opt_key.kind_of? Symbol) then
	  case (opt_value)
	  when String, Numeric
	    degenerated_option[opt_key] = opt_value
	  else
	    log.warning("sid #{sid}: #{type}: ",
			"option value of #{opt_key} is complex object: ",
			opt_value.inspect)
	  end
	else
	  log.warning("sid #{sid}: #{type}: option key is not symbol: ", opt_key.inspect)
	end
      end

      degenerated_option
    end
    module_function :degenerate_option

    def expand_option(sid, log, option, degenerated_option)
      for opt_key, opt_value in degenerated_option
	option[opt_key.intern] = opt_value
      end

      option
    end
    module_function :expand_option
  end

  class XMLRPCDispatcherDocument < Document
    include XMLRPCDataConversion

    def initialize(*args)
      @client = XMLRPC::Client::new(*args)
      @host, @path, @port = args
      @host = 'localhost' unless @host
      @path = '/RPC2' unless @path
      @port = 80 unless @port
    end

    def content(sid, srv_opt, prefix, request, response)
      srv_opt.log.debug("sid #{sid}: debug #{type}")

      xmlrpc_request = {
	'request' => {},
	'response' => {}
      }
      xmlrpc_request['sid'] = sid
      xmlrpc_request['option'] = degenerate_option(sid, srv_opt.log, srv_opt)
      xmlrpc_request['prefix'] = prefix
      xmlrpc_request['request']['method'] = request.method
      xmlrpc_request['request']['path'] = request.reduce_path(prefix)
      xmlrpc_request['request']['path'] += '?' + request.query if request.query
      xmlrpc_request['request']['version'] = request.version.to_s
      xmlrpc_request['request']['header'] = degenerate_header(sid, srv_opt.log, request.header)
      xmlrpc_request['request']['content'] = XMLRPC::Base64::new(request.content.read || '')
      xmlrpc_request['request']['trailer'] = request.trailer.to_a
      xmlrpc_request['request']['host_name'] = request.host.name
      xmlrpc_request['request']['host_port'] = request.host.port
      xmlrpc_request['request']['client_name'] = request.client.name
      xmlrpc_request['request']['client_address'] = request.client.address
      xmlrpc_request['request']['option'] = degenerate_option(sid, srv_opt.log, request.option)
      xmlrpc_request['response']['status'] = response.status
      xmlrpc_request['response']['version'] = response.version.to_s
      xmlrpc_request['response']['reason'] = response.reason
      xmlrpc_request['response']['header'] = degenerate_header(sid, srv_opt.log, response.header)

      srv_opt.log.message("sid #{sid}: #{type}: ",
			  "XML-RPC call to http://#{@host}:#{@port}#{@path}")

      for first_key, substruct in xmlrpc_request
	for second_key, value in substruct
	  srv_opt.log.debug{
	    msg = "sid #{sid}: debug #{type}: XML-RPC send: #{first_key}->#{second_key}"
	    if (first_key != 'sid') then
	      msg << '->'
	      msg << ((second_key != 'content') ? value.inspect : value.decoded.length.to_s + ' octets...')
	    end

	    msg
	  }
	end
      end

      xmlrpc_response = @client.call('wwwsrv.repository.content', xmlrpc_request)

      for key, value in xmlrpc_response
	srv_opt.log.debug{
	  "sid #{sid}: debug #{type}: XML-RPC reply: #{key}->" +
	  ((key != 'content') ? value.inspect : value.join.length.to_s + ' octents...')
	}
      end

      response.version = HTTPversion::parse(xmlrpc_response['version'])
      response.status = xmlrpc_response['status']
      response.reason = xmlrpc_response['reason']
      expand_option(sid, srv_opt.log, response.option, xmlrpc_response['option'])

      response.header =
	expand_header(sid, srv_opt.log, response.header, xmlrpc_response['header'])
      response.trailer =
	expand_header(sid, srv_opt.log, response.trailer, xmlrpc_response['trailer'])

      if (xmlrpc_response.include? 'error') then
	srv_opt.log.debug{
	  "sid #{sid}: debug #{type}: error -> " + xmlrpc_response['error'].inspect
	}
	err_msg = xmlrpc_response['err_msg'] || nil
	http_error(response.status, err_msg, response.header, response.version)
      end

      yield(response)
      for content_fragment, *opts in xmlrpc_response['content']
	yield(content_fragment, *opts)
      end

      self
    end
  end

  class XMLRPCRepositoryContentReader
    def initialize(content, trailer)
      @content = content
      @trailer = trailer
    end

    attr_reader :trailer

    def read(length=nil)
      (length && length < 0) and
	raise ArgumentError, "negative length #{length} given"

      if (@content.empty?) then
	return nil
      elsif (! length || @content.length <= length) then
	content_fragment = @content
	@content = ''
	return content_fragment
      else
	content_fragment = @content[0, length]
	@content[0, length] = ''
	return content_fragment
      end
    end

    alias raw_read read

    def each_fragment
      yield(read)
    end
  end

  class XMLRPCPseudoRootDocument < Document
    include Path

    def initialize(document, prefix)
      @document = document
      @prefix = prefix
    end

    attr_reader :document
    attr_reader :prefix

    def content(sid, srv_opt, prefix, request, response)
      srv_opt.log.debug("sid #{sid}: debug #{type}")
      if (Path::reduce_path(@prefix, prefix)) then
	@document.content(sid, srv_opt, prefix, request, response) do |arg, *opts|
	  yield(arg, *opts)
	end
	return self
      end

      http_error(404, "Not found a document:#{request.normalize_path}")
    end
  end

  class XMLRPCRepositoryDocument < Document
    include DocumentCascade
    include XMLRPCDataConversion

    def initialize(document=DocumentFolder::new, server=XMLRPCServerDocument::new)
      @document = document
      @server = server
      @context = {}
      @root = {}

      @server.add_handler('wwwsrv.repository.content') { |xmlrpc_request|
	repository_content(xmlrpc_request)
      }
    end

    attr_accessor :document
    attr_accessor :server

    def context
      @context[Thread::current.id]
    end
    private :context

    def content(sid, srv_opt, prefix, request, response)
      srv_opt.log.debug("sid #{sid}: debug #{type}")
      begin
	@context[Thread::current.id] = {
	  :sid => sid,
	  :srv_opt => srv_opt,
	  :request => request
	}
	@server.content(sid, srv_opt, prefix, request, response) do |arg, *opts|
	  yield(arg, *opts)
	end
      ensure
	@context.delete(Thread::current.id)
      end
    end

    def repository_content(xmlrpc_request)
      sid = context[:sid]
      prefix = xmlrpc_request['prefix']
      unless (@root.include? prefix) then
	@root[prefix] = XMLRPCPseudoRootDocument::new(@document, prefix)
      end

      srv_opt = ServerOption::new
      srv_opt.root = @root[prefix]
      srv_opt.log = context[:srv_opt].log
      srv_opt.log.message("sid #{sid}: #{type}: ",
			  "XML-RPC call from #{context[:request].client}: ",
			  "client sid #{xmlrpc_request['sid']}")
      expand_option(sid, srv_opt.log, srv_opt, xmlrpc_request['option'])

      for first_key, substruct in xmlrpc_request
	for second_key, value in substruct
	  srv_opt.log.debug{
	    msg = "sid #{sid}: debug #{type}: XML-RPC accept: #{first_key}->#{second_key}"
	    if (first_key != 'sid') then
	      msg << '->'
	      msg << ((second_key != 'content') ? value.inspect : value.length.to_s + ' octets...')
	    end

	    msg
	  }
	end
      end

      request = Request::new(xmlrpc_request['request']['method'],
			     xmlrpc_request['request']['path'],
			     HTTPversion::parse(
			       xmlrpc_request['request']['version']),
			     expand_header(sid, srv_opt.log, Header::new,
					   xmlrpc_request['request']['header']))

      request.content = XMLRPCRepositoryContentReader::new(
			  xmlrpc_request['request']['content'],
			  expand_header(sid, srv_opt.log, Header::new,
					xmlrpc_request['request']['trailer']))

      request.host = Host::new(xmlrpc_request['request']['host_name'],
			       xmlrpc_request['request']['host_port'])

      request.client = Client::new(xmlrpc_request['request']['client_name'],
				   xmlrpc_request['request']['client_address'])

      expand_option(sid, srv_opt.log, request.option,
		    xmlrpc_request['request']['option'])

      http_version = HTTPversion::parse(xmlrpc_request['response']['version'])
      response = Response::new(http_version, xmlrpc_request['response']['status'])
      response.reason = xmlrpc_request['response']['reason']

      doc_response = nil
      content_spool = []

      begin
	state, err_args = Document::catch_state{
	  @document.content(sid, srv_opt, prefix, request, response) do |arg, *opts|
	    unless (doc_response) then
	      doc_response = arg
	    else
	      content_spool.push([ XMLRPC::Base64::new(arg) ] + opts)
	    end
	  end
	}
      rescue StandardError, ScriptError
	state = :error
	err_args = [ 500, $! ]
      end

      xmlrpc_response = {}
      case (state)
      when :success
	xmlrpc_response['version'] = doc_response.version.to_s
	xmlrpc_response['status'] = doc_response.status
	xmlrpc_response['reason'] = doc_response.reason
	xmlrpc_response['header'] = degenerate_header(sid, srv_opt.log, doc_response.header)
	xmlrpc_response['content'] = content_spool
	xmlrpc_response['trailer'] = doc_response.trailer.to_a
	xmlrpc_response['option'] = degenerate_option(sid, srv_opt.log, doc_response.option)
      when :error
	status, message, header, version = err_args
	if (status.kind_of? Response) then
	  # for backward compatibility
	  response = status
	  status = response.status
	  header = response.header
	  version = response.version
	end
	xmlrpc_response['error'] = true
	xmlrpc_response['err_msg'] = message ? message.to_s : ''
	xmlrpc_response['version'] = version.to_s
	xmlrpc_response['status'] = status
	xmlrpc_response['reason'] = HTTP_MSG[status]
	xmlrpc_response['header'] = degenerate_header(sid, srv_opt.log, header)
	xmlrpc_response['option'] = degenerate_option(sid, srv_opt.log, err_response.option)
      else
	raise "unknown document state: #{state}"
      end

      for key, value in xmlrpc_response
	srv_opt.log.debug{
	  "sid #{sid}: debug #{type}: XML-RPC return: #{key}->" +
	  ((key != 'content') ? value.inspect : value.map{ |arg, *opts| arg.decoded }.join.length.to_s + ' octets...')
	}
      end

      xmlrpc_response
    end
    private :repository_content
  end
end
